/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "alloc.h"
#include "buffer.h"
#include "main.h"
#include "session.h"
#include "str.h"

/*#define DEBUGGING 1*/

/* Public functions. */

/* Output buffer linked list. */
struct buffer
  {
    struct buffer *next;	/* Next buffer in list. */
    struct line *line;		/* This line. */
  };

/* A single line of output. */
struct line
  {
    int ref_cnt;		/* Reference count. */
    size_t len;			/* Number of bytes. */
    char *buf;			/* Buffer contents. */
  };

static void free_line (struct line *);

/* Initialize SESSION's buffers. */
void
buf_open_session (struct session *session)
{
  session->in_head = session->in_tail = 0;
  session->in_newline = -1;
  session->in_discard = 0;
  session->out_head = session->out_tail = NULL;
}

/* Free and uninitialize SESSION's buffers. */
void
buf_close_session (struct session *session)
{
  struct buffer *cur, *next;
  
  for (cur = session->out_head; cur; cur = next)
    {
      next = cur->next;
      free_line (cur->line);
      free (cur);
    }
}

/* Creates and returns a line structure containing S, which becomes
   owned by the line. */
struct line *
buf_create_line (char *s)
{
  struct line *line;

  line = xmalloc (sizeof (struct line));
  line->ref_cnt = 0;
  line->len = strlen (s);
  line->buf = s;

  return line;
}

/* Queues the given LINE for output in SESSION. */
void
buf_write_line (struct session *s, struct line *line)
{
  struct buffer *new_buf = xmalloc (sizeof (struct buffer));
  new_buf->next = NULL;
  new_buf->line = line;
  if (s->out_tail == NULL)
    {
      s->out_head = s->out_tail = new_buf;
      s->out_sent = 0;
    }
  else
    {
      s->out_tail->next = new_buf;
      s->out_tail = new_buf;
    }

  line->ref_cnt++;

  FD_SET (s->sock, &write_fds);

#if DEBUGGING
  printf ("%s:%s << %.*s", s->hostinfo, s->username ? s->username : "none",
	  (int) line->len, line->buf);
#endif
}

/* Write STRING to SESSION's output buffer. */
void
buf_write_string (struct session *s, const char *string)
{
  buf_write_line (s, buf_create_line (xstrdup (string)));
}

/* Write FORMAT to SESSION's output buffer after running it through
   printf (). */
void
buf_write_printf (struct session *s, const char *format, ...)
{
  va_list args;
  char buf[BUFSIZ];

  va_start (args, format);
  vsnprintf (buf, BUFSIZ, format, args);
  va_end (args);

  buf_write_string (s, buf);
}

/* Writes as much data from the output buffer to S's socket as
   possible. */
void
buf_flush (struct session *s)
{
  while (s->out_head != NULL)
    {
      struct line *line;
      ssize_t amt;
  
      line = s->out_head->line;
      do
	amt = write (s->sock, line->buf + s->out_sent,
		     line->len - s->out_sent);
      while (amt == -1 && errno == EINTR);

      if (amt == -1)
	{
	  if (errno != EAGAIN && errno != EWOULDBLOCK)
	    {
	      perror ("write");
	      transition (s, CLS_CLOSED);
	    }

	  return;
	}
      else 
	{
	  s->out_sent += amt;
	  assert (s->out_sent <= line->len);
	  
	  if (s->out_sent >= line->len)
	    {
	      struct buffer *next = s->out_head->next;
	      free (s->out_head);
	      s->out_head = next;
	      s->out_sent = 0;

	      if (s->out_head == NULL)
		{
		  s->out_tail = NULL;
		  FD_CLR (s->sock, &write_fds);
		}

	      free_line (line);
	    }
	}
    }

  if (s->state == CLS_CLOSING)
    transition (s, CLS_CLOSED);
}

/* Returns the value of INDEX, which is an index into a struct session
   in_buf, advanced by a single character. */
#define ADVANCE(INDEX) \
	((INDEX) < IN_BUF_SIZE - 1 ? (INDEX) + 1 : 0)

static int try_read (struct session *, size_t count);
static void find_newline (struct session *);

/* Strips a trailing carriage return from string S, if present. */
static inline void 
strip_cr (char *s)
{
  size_t len = strlen (s);

  if (len > 0 && s[len - 1] == '\r')
    s[len - 1] = '\0';
}

/* Tries to find a newline inside S's input buffer, and sets S's
   in_newline member as appropriate whether it finds it or not. */
static inline void
find_newline (struct session *s)
{
  const char *nl;
  
  if (s->in_head >= s->in_tail)
    nl = memchr (s->in_buf + s->in_tail, '\n', s->in_head - s->in_tail);
  else
    {
      nl = memchr (s->in_buf + s->in_tail, '\n', IN_BUF_SIZE - s->in_tail);
      if (nl == NULL)
	nl = memchr (s->in_buf, '\n', s->in_head - 1);
    }

  if (nl)
    s->in_newline = nl - s->in_buf;
  else
    s->in_newline = -1;
}

/* Tries to read AMT bytes into S's input buffer at its head,
   advancing the head pointer as needed.
   Returns nonzero only if AMT is nonzero and that many bytes are
   successfully read.
   If bytes are read, checks for a newline and sets S's in_newline
   member as appropriate.
   If an error occurs, marks S for closure. */
static inline int
try_read (struct session *s, size_t amt)
{
  ssize_t got;
  
  if (amt == 0)
    return 0;

  do
    got = read (s->sock, s->in_buf + s->in_head, amt);
  while (got == -1 && errno == EINTR);

  assert (got <= (ssize_t) amt);
  if (got == 0)
    {
      /* EOF on socket. */
      transition (s, CLS_CLOSED);
      return 0;
    }
  else if (got == -1)
    {
      if (errno != EWOULDBLOCK && errno != EAGAIN)
	transition (s, CLS_CLOSED);
      return 0;
    }

  if (s->in_newline == -1)
    {
      const char *const nl = memchr (s->in_buf + s->in_head, '\n', got);
      if (nl)
	s->in_newline = nl - s->in_buf;
    }
  
  s->in_head += got;
  assert (s->in_head <= IN_BUF_SIZE);
  if (s->in_head == IN_BUF_SIZE)
    s->in_head = 0;

  return (ssize_t) amt == got;
}

/* Reads a line from S and returns it in a static buffer.
   If POLL is non-null then tries to extract more data from the socket
   before reading. */
const char *
buf_read_line (struct session *s, int poll)
{
  if (s->state == CLS_CLOSING)
    return NULL;
  
  if (poll)
    {
      size_t amt;

      do
	{
	  if (s->in_head >= s->in_tail)
	    {
	      amt = IN_BUF_SIZE - s->in_head;
	      if (s->in_tail == 0)
		amt--;
	    }
	  else
	    amt = s->in_tail - s->in_head - 1;
	}
      while (try_read (s, amt));
    }

  if (s->in_newline != -1)
    {
      char *r;
      
      if (s->in_newline >= s->in_tail)
	{
	  r = s->in_buf + s->in_tail;
	  s->in_buf[s->in_newline] = '\0';
	}
      else
	{
	  static char temp[IN_BUF_SIZE];

	  memcpy (temp, s->in_buf + s->in_tail, IN_BUF_SIZE - s->in_tail);
	  memcpy (temp + (IN_BUF_SIZE - s->in_tail), s->in_buf, s->in_newline);
	  temp[IN_BUF_SIZE - s->in_tail + s->in_newline] = '\0';
	  r = temp;
	}

      s->in_tail = ADVANCE (s->in_newline);
      find_newline (s);

      if (s->in_discard)
	{
	  s->in_discard = 0;
	  buf_write_string (s, "500 Line too long\r\n");
	  r = NULL;
	}
      else
	strip_cr (r);

#if DEBUGGING
      printf ("%s:%s >> %s\n",
	      s->hostinfo, s->username ? s->username : "none", r);
#endif
      return r;
    }
  else
    {
      if (ADVANCE (s->in_head) == s->in_tail)
	{
	  s->in_head = s->in_tail = 0;
	  s->in_discard = 1;
	}
      
      return NULL;
    }
}
   

/* Private functions. */

/* Frees LINE. */
static void
free_line (struct line *line)
{
  line->ref_cnt--;
  assert (line->ref_cnt >= 0);
  
  if (line->ref_cnt == 0)
    {
      free (line->buf);
      free (line);
    }
}
