/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000 Michigan State University.
   Copyright (C) 1993-1999, 2000 Free Software Foundation.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

/* Some of the code below was derived from libc.info in glibc 2.1,
   modified heavily for this chat server. */

#include <config.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <grp.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include "alloc.h"
#include "buffer.h"
#include "conf.h"
#include "main.h"
#include "session.h"
#include "str.h"

/* Sessions in progress. */
struct session **sessions;
int n_sessions;
int m_sessions;

/* Set on SIGHUP. */
static sig_atomic_t hup;

/* Server socket. */
static int server_sock;

/* File descriptor sets. */
fd_set read_fds;
fd_set write_fds;

/* Command-line options. */
static int do_detach;		/* 1=fork away from foreground. */
static const char *log_fn;	/* Name of log file, or NULL if none. */
static int port_lo, port_hi;	/* Acceptable range of ports. */
static int pidfile_enable;	/* Enable pidfile. */
static char *pidfile_name;	/* Name of pidfile if enabled. */

/* Have we detached and become a daemon? */
static int detached;

/* Log file. */
static FILE *log_file;


/* Public functions. */
     
static void free_session (struct session *);
static int make_socket (void);
static void main_loop (void) attribute ((noreturn));
static void new_client (void);
static void process_input (struct session *);
static void parse_cmdline (int argc, char **argv);
static void usage (void) attribute ((noreturn));
static void make_daemon (void);
static void write_pidfile (void);
static void open_log (void);
static void sighup (int);

int
main (int argc, char **argv)
{
  parse_cmdline (argc, argv);

  open_log ();

  openlog ("scope-chat", 0, LOG_DAEMON);

  /* Ignore SIGPIPE--we prefer to handle it as a function failure
     return. */
  signal (SIGPIPE, SIG_IGN);

  /* Set up SIGHUP handler. */
  signal (SIGHUP, sighup);

  /* Create the socket and set it up to accept connections. */
  server_sock = make_socket ();
  if (listen (server_sock, 1) < 0)
    fail ("listen: %s", strerror (errno));

  if (do_detach)
    make_daemon ();

  if (pidfile_enable)
    write_pidfile ();

  main_loop ();
}

/* Private functions. */

/* Main loop. */   
static void
main_loop (void)
{
  /* Initialize the set of active sockets. */
  FD_ZERO (&read_fds);
  FD_ZERO (&write_fds);
  FD_SET (server_sock, &read_fds);

  log ("Status: server started, pid=%ld", (long) getpid ());
  for (;;)
    {
      fd_set r = read_fds;
      fd_set w = write_fds;
      int i;

      /* Block until input arrives on one or more active sockets. */
      if (select (FD_SETSIZE, &r, &w, NULL, NULL) < 0
	  && errno != EINTR)
	fail ("select: %s", strerror (errno));

      /* Handle SIGHUP by closing and reopening our log file, to allow
         for log rotation. */
      if (hup)
	{
	  hup = 0;
	  open_log ();
	}

      /* Handle new clients. */
      if (FD_ISSET (server_sock, &r))
	new_client ();

      /* For each session, check for data to read and data to
         write. */
      for (i = 0; i < n_sessions; i++)
	{
	  struct session *const s = sessions[i];
	  
	  if (FD_ISSET (s->sock, &r))
	    process_input (s);

	  if (FD_ISSET (s->sock, &w))
	    buf_flush (s);

	  if (s->state == CLS_CLOSED)
	    {
	      n_sessions--;
	      free_session (s);
	      if (i < n_sessions)
		sessions[i] = sessions[n_sessions];
	      i--;
	    }
	}

      if (log_file)
	fflush (log_file);
    }
}

/* Make a server socket in range [port_lo, port_hi] on this host. */
static int
make_socket (void)
{
  int sock;
  struct sockaddr_in name;
  int port;
     
  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    fail ("socket: %s", strerror (errno));

  for (port = port_lo; port <= port_hi; port++)
    {
      name.sin_family = AF_INET;
      name.sin_port = htons (port);
      name.sin_addr.s_addr = htonl (INADDR_ANY);
      if (bind (sock, (struct sockaddr *) &name, sizeof (name)) >= 0)
	{
	  syslog (LOG_NOTICE, "listening on port %d", port);
	  return sock;
	}

      if (errno != EADDRINUSE)
	fail ("bind: %s", strerror (errno));
    }

  fail ("can't listen on any port from %d to %d: ports all in use",
	port_lo, port_hi);
}

/* Accept a new client from the server socket,
   and set up the session structure. */
static void
new_client (void)
{
  struct sockaddr_in clientname;
  size_t size;
  int sock;

  size = sizeof (clientname);
  sock = accept (server_sock, (struct sockaddr *) &clientname, &size);
  if (sock < 0)
    {
      perror ("accept");
      return;
    }

  syslog (LOG_INFO, "connect from host %s, port %hd",
	  inet_ntoa (clientname.sin_addr), ntohs (clientname.sin_port));

  /* Set the socket for non-blocking mode so that we don't have to
     wait around for stuff to happen. */
  if (fcntl (sock, F_SETFL, fcntl (sock, F_GETFL) | O_NONBLOCK) == -1)
    {
      perror ("fcntl");
      return;
    }

  /* Expand the sessions array if necessary. */
  if (n_sessions >= m_sessions)
    {
      m_sessions += 16;
      sessions = xrealloc (sessions, sizeof (struct session *) * m_sessions);
    }
  
  /* Set up the session structure. */
  {
    struct session *s;
  
    s = sessions[n_sessions++] = xmalloc (sizeof (struct session));
    s->sock = sock;
    s->state = CLS_INIT;

    s->workgroup = NULL;
    s->username = NULL;
    s->hostinfo = blp_asprintf ("%s:%d", inet_ntoa (clientname.sin_addr),
				ntohs (clientname.sin_port));
    s->cookie = NULL;
    s->flags = 0;

    buf_open_session (s);
    buf_write_string (s, "220 TeamSCOPE Chat 1.0 Server\r\n");
  }

  FD_SET (sock, &read_fds);
}

/* Process socket input on session S. */
static void
process_input (struct session *s)
{
  const char *line;

  line = buf_read_line (s, 1);
  while (line != NULL)
    {
      input (s, line);

      line = buf_read_line (s, 0);
    }
}

/* Destroy session S. */
static void
free_session (struct session *s)
{
  FD_CLR (s->sock, &read_fds);
  FD_CLR (s->sock, &write_fds);
  TEMP_FAILURE_RETRY (close (s->sock));

  free (s->username);
  free (s->workgroup);
  free (s->hostinfo);
  free (s->cookie);

  buf_close_session (s);
  free (s);
}

/* Send S, formatted as with printf(), to the log file. */
void
log (const char *s, ...)
{
  va_list args;

  if (log_file == NULL)
    return;
  
  {
    time_t t;
    struct tm *tm;
    char buf[32];

    time (&t);
    tm = localtime (&t);
    if (strftime (buf, 32, "[%d %b %Y %H:%M:%S] ", tm) != 0)
      fputs (buf, log_file);
  }

  va_start (args, s);
  vfprintf (log_file, s, args);
  putc ('\n', log_file);
  va_end (args);
}

/* Parse the command line. */
void
parse_cmdline (int argc, char **argv)
{
  static const struct option long_options[] = 
    {
      {"no-detach", no_argument, NULL, 'd'},
      {"log", optional_argument, NULL, 'l'},
      {"port", required_argument, NULL, 'p'},
      {"pidfile", optional_argument, NULL, 'P'},
      {"help", no_argument, NULL, 'h'},
      {"version", no_argument, NULL, 'v'},
      {0, 0, 0, 0},
    };

  do_detach = 1;
  log_fn = CHAT_LOG_FILE;
  port_lo = 6555;
  port_hi = 6560;
  pidfile_enable = 0;
  pidfile_name = PIDFILE_DIR "/scope-chat.pid";
  
  for (;;)
    {
      int c = getopt_long (argc, argv, "dl:p:P::hv", long_options, NULL);
      if (c == -1)
	break;
      
      switch (c)
	{
	case 'd':
	  do_detach = 0;
	  break;

	case 'l':
	  log_fn = optarg;
	  break;

	case 'p':
	  switch (sscanf (optarg, "%d-%d", &port_lo, &port_hi))
	    {
	    case 0:
	    case EOF:
	      usage ();

	    case 1:
	      port_hi = port_lo;
	      break;

	    case 2:
	      break;

	    default:
	      assert (0);
	    }
	  break;

	case 'P':
	  pidfile_enable = 1;
	  if (optarg)
	    pidfile_name = optarg;
	  break;
	  
	case 'h':
	  usage ();

	case 'v':
	  fputs ("TeamSCOPE Chat 1.0\n"
		 "Copyright (C) 1999, 2000 Michigan State University.\n"
		 "This is free software; see the source for copying "
		 "conditions.  There is NO\nWARRANTY; not even for "
		 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n"
		 "Written by Ben Pfaff <pfaffben@msu.edu>.\n", stdout);
	  exit (EXIT_SUCCESS);

	case '?':
	  usage ();

	case 0:
	  break;

	default:
	  assert (0);
	}
    }
}

/* Print a syntax message and quit. */
static void
usage (void)
{
  fputs (
"TeamSCOPE Chat, server counterpart to the TeamSCOPE Java-based chat client.\n"
"\nUsage: scope-chat [OPTION]...\n"
"If a long option shows an argument as mandatory, then it is mandatory\n"
"for the equivalent short option also.  Similarly for optional arguments.\n"
"\nConfiguration:\n"
"  -d, --no-detach         Do not fork and detach from the current session.\n"
"  -l, --log[=LOGFILE]     Log chat to LOGFILE, or disable if no file given.\n"
"                          Default log file is " CHAT_LOG_FILE ".\n"
"  -p, --port=LO[-HI]      Server ports.  Uses first available port.\n"
"  -P, --pidfile[=PIDFILE] Enable and optionally set name of pidfile.\n"
"                          Default pidfile is " PIDFILE_DIR
"/scope-chat.pid.\n"
"\nInformative output:\n"
"  -h, --help              Print this help, then exit.\n"
"  -v, --version           Show version, then exit.\n",
         stdout);

  exit (EXIT_SUCCESS);
}

/* Make this process into a daemon.
   Adapted from W. Richard Stevens, _Advanced Programming in the UNIX
   Environment_. */
static void
make_daemon (void)
{
  /* Fork and kill off the parent process. */
  {
    pid_t pid;

    pid = fork ();
    if (pid < 0)
      fail ("fork(): %s\n", strerror (errno));
    else if (pid != 0)
      exit (EXIT_SUCCESS);
  }

  /* Become session leader. */
  setsid ();

  chdir ("/");

  umask (0);

  /* Should also close unneeded file descriptors, but it's not
     essential. */

  detached = 1;
}

/* Write our pid to the pidfile. */
static void
write_pidfile (void)
{
  FILE *f = fopen (pidfile_name, "w");
  if (f == NULL)
    {
      syslog (LOG_ERR, "error opening pid file %s: %s\n",
	      pidfile_name, strerror (errno));
      return;
    }

  fprintf (f, "%ld\n", (long) getpid ());
  if (ferror (f) || fclose (f) == EOF)
    log ("error on %s: %s\n", pidfile_name, strerror (errno));
}

/* Open the log file, closing it first if necessary. */
static void
open_log (void)
{
  if (log_file)
    fclose (log_file);

  /* Open the log file. */
  if (log_fn) 
    {
      log_file = fopen (log_fn, "a");
      if (!log_file)
	fail ("error opening log file %s: %s\n", log_fn, strerror (errno));
    }
  else
    log_file = NULL;
}

/* SIGHUP handler. */
static void
sighup (int signum unused)
{
  hup = 1;
}

void
fail (const char *format, ...)
{
  va_list args;

  va_start (args, format);
  if (detached)
    vsyslog (LOG_CRIT, format, args);
  else
    {
      vfprintf (stderr, format, args);
      putc ('\n', stderr);
    }
  va_end (args);

  exit (EXIT_FAILURE);
}
