/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#if !activity_h
#define activity_h 1

#include "db.h"
#include "id-list.h"

/* Activity categories. */
enum activity_cat
  {
    /* These are the ones that users are probably interested in. */
    AC_CAT_FILE = 1,		/* Files and file-related messages. */
    AC_CAT_MSG_BOARD,
    AC_CAT_EVENT,
    AC_N_CATEGORIES,
  };

/* Activity types. */
enum activity_type
  {
    /* File activity types.  Don't change. */
    AC_FILE_DOWNLOAD = 1,
    AC_FILE_DOWNLOAD_ARCHIVE,
    AC_FILE_UPLOAD,
    AC_FILE_DELETE,
    AC_FILE_COPY,
    AC_FILE_RENAME,
    AC_FILE_MKDIR,
    AC_FILE_RMDIR,
    AC_FILE_CHMOD,
    AC_FILE_MOVE,

    AC_FILE_POST,
    AC_FILE_POST_DELETE,
    AC_FILE_POST_READ,
    AC_FILE_POST_MAILED,

    AC_FILE_N_ACTIVITIES,

    /* Message board activities.  Don't change. */
    AC_MSG_POST = 1,
    AC_MSG_DELETE,
    AC_MSG_READ,
    AC_MSG_MAILED,

    AC_MSG_N_ACTIVITIES,

    /* Calendar event types.  */
    AC_EVENT_NEW = 1,
    AC_EVENT_DELETE,
    AC_EVENT_CHANGE,
    AC_EVENT_VIEW,

    AC_EVENT_N_ACTIVITIES,

    /* Maximum number of activities in one category.
       This is the minimum number of bits in an unsigned long. */
    AC_MAX_ACTIVITIES = 32
  };

/* Sort orders. */
enum activity_sort
  {
    AC_SORT_USER,
    AC_SORT_ACTIVITY,
    AC_SORT_FILENAME,
    AC_SORT_DATE,

    AC_SORT_COUNT,

    AC_SORT_DEFAULT
  };

/* Per-user types of activities to view. */
struct activity_person
  {
    struct id_list usr;				/* List of applicable users. */
    unsigned long type[AC_N_CATEGORIES];	/* Types to view. */
  };

/* Represents of a database query for activities. */
struct activity_query
  {
    /* Input. */
    enum activity_sort sort;	/* Sort order. */

    struct activity_person *person;
    int n_people;
    
    char *begin_date;		/* Earliest acceptable date string, or NULL. */
    char *end_date;		/* Latest acceptable date string, or NULL. */

    char *filename;		/* Filename substring search string. */
    int filename_exclude;	/* >0: `filename' is files to include.
				   <0: `filename' is files to exclude. */

    int last_read_seq;		/* Last read sequence number. */
    int search;			/* Nonzero if this is a search. */

    int start;			/* 0-based index of first record to read. */
    int limit;			/* Maximum number of records to read. */

    /* Output. */
    db_result *result;		/* Records read. */
    int count;			/* Number of records actually read. */
  };

struct url;
struct request;
struct context;

void scr_activity (struct request *);
void activity_abbreviated (struct request *, int category);

void activity_show_filter (struct request *, struct activity_query *);

struct std *activity_query_to_dict (const struct activity_query *);
void activity_query_init (struct activity_query *);
void activity_query_from_dict (struct activity_query *, const struct std *,
			       const char *usr);
void activity_query_modify_from_cgi (struct activity_query *query,
				     struct std *cgi, const char *usr);
void activity_query_set_last_read (struct context *,
				   struct activity_query *query);
void activity_query_run (struct context *, struct activity_query *);
void activity_query_present (struct context *, struct activity_query *,
			     const struct url *, const struct std *);
void activity_query_destroy (struct activity_query *);

void log_activity (struct context *, int s1, int s2,
		   enum activity_cat, enum activity_type, int ref_seq);
void log_activity_by_name (struct context *, const char *s1, const char *s2,
			   enum activity_cat, enum activity_type, int ref_seq);

char *activity_dup_nonnull (const char *s);

/* Activity bitmaps. */

extern const unsigned long activity_cat_cnt[AC_N_CATEGORIES];

int activity_all_bits_p (unsigned long bits, int cat);
int activity_zero_bits_p (unsigned long bits, int cat);

int activity_bm_all_p (const unsigned long bm[AC_N_CATEGORIES]);
int activity_bm_zero_p (const unsigned long bm[AC_N_CATEGORIES]);
int activity_bm_equal (unsigned long (*bm)[AC_N_CATEGORIES], int a, int b);
void activity_bm_copy (unsigned long dst[AC_N_CATEGORIES],
		       const unsigned long src[AC_N_CATEGORIES]);
void activity_bm_all (unsigned long bm[AC_N_CATEGORIES]);
void activity_bm_zero (unsigned long bm[AC_N_CATEGORIES]);

/* Returns a mask with bits 1...N set. */
#define ACTIVITY_MASK_FOR_WIDTH(N)		\
	(((N) == AC_MAX_ACTIVITIES		\
         ? ULONG_MAX				\
         : (1ul << (N)) - 1) & ~1ul)

#endif /* activity_h */
