/* TeamSCOPE - Synthetic Construct Optimized for Peacekeeping and Exploration.
   Copyright (C) 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

/* FIXME? This code is riddled with race conditions, if multiple users
   try to update things at once.  I don't think any of them are
   actually harmful, though, just annoying and potentially confusing.
   Hopefully administrators will coordinate their efforts a little. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <grp.h>
#include <pwd.h>
#include "admin.h"
#include "alloc.h"
#include "auth.h"
#include "conf.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "id-list.h"
#include "identity.h"
#include "priv.h"
#include "scope.cgi.h"
#include "std.h"

/* Administrator's level of power. */
enum admin_level
  {
    L_USR,	/* Ordinary user: May modify own info. */
    L_GRP,	/* Group administrator: May administer this group only. */
    L_SYS,	/* System administrator: May do anything. */

    L_CNT,
  };

/* Groups required by a command interface. */
enum admin_group
  {
    G_NONE,	/* No groups. */
    G_MINE,	/* One group: the group the user is in. */
    G_GRP,	/* Any one group. */
    G_GRPS,	/* One or more groups. */
  };

/* Users required by a command interface. */
enum admin_user
  {
    U_NONE,		/* No groups. */
    U_ME,		/* One user: the current user. */
    U_USR,		/* Any one user. */
    U_USRS,		/* One or more users. */
    U_ONLY_MY_USRS,	/* One or more users who belong to this group
			   and to no other groups. */
    U_MY_USR,		/* One user who belongs to this group. */
    U_MY_USRS,		/* One or more users in this group. */
  };

/* Groups or users required by a command. */
enum admin_needs_args
  {
    A_ZERO,	/* No groups/users. */
    A_ONE,	/* One group or user. */
    A_MANY	/* At least one group or user. */
  };

struct admin_request;
static void show_request (struct admin_request *, int terminate);
static void success (struct admin_request *, int terminate);

/* A command execution function.
   Used to execute an administration command that needs root.
   Runs as root, so be careful in writing it.
   Returns nonzero if the request was fully handled.
   Returns zero if the request should be passed to the
   interact function too. */
typedef int admin_execute_func (struct admin_request *);

/* A command interaction function.
   Used to interact with the user when the grp and usr
   prompt functionality isn't enough.
   Runs as the user, not root. */
typedef void admin_interact_func (struct admin_request *);

/* An administration command. */
struct admin_command
  {
    const char *name;			/* Command name. */
    enum admin_needs_args grp, usr;	/* Groups and users needed. */

    /* Some commands need only execution functions.
       Some command need only interaction functions.
       Some need both.  All need one or the other. */
    admin_execute_func *execute;	/* Execution function. */
    admin_interact_func *interact;	/* Interaction function. */
  };

/* Unimplemented command. */
#define UNIMP_CMD(NAME, GRP, USR)			\
	static const struct admin_command c_##NAME =	\
		{#NAME, GRP, USR, NULL, NULL}

/* A command that needs both execution and interaction functions.
   Typically these have you fill out a form and then execute it
   based on the fields. */
#define NORMAL_CMD(NAME, GRP, USR)			\
	static admin_execute_func execute_##NAME;	\
	static admin_interact_func interact_##NAME;	\
	static const struct admin_command c_##NAME =	\
		{#NAME, GRP, USR, execute_##NAME, interact_##NAME}

/* A command that needs only an interaction function.
   Typically these are informational only. */
#define NOEXEC_CMD(NAME, GRP, USR)			\
	static admin_interact_func interact_##NAME;	\
	static const struct admin_command c_##NAME =	\
		{#NAME, GRP, USR, NULL, interact_##NAME}

/* A command that needs only an execution function.
   Typically these are commands that need no input or just
   grp/usr lists. */
#define NOINTR_CMD(NAME, GRP, USR)			\
	static admin_execute_func execute_##NAME;	\
	static const struct admin_command c_##NAME =	\
		{#NAME, GRP, USR, execute_##NAME, NULL}

/* All the administration commands. */
NOEXEC_CMD (menu, A_ZERO, A_ZERO);
UNIMP_CMD (separator, A_ZERO, A_ZERO);

NOEXEC_CMD (viewusr, A_ZERO, A_ONE);
NOEXEC_CMD (viewusrs, A_ZERO, A_ZERO);
NORMAL_CMD (createusr, A_MANY, A_ZERO);
NOINTR_CMD (delusr, A_ZERO, A_MANY);
NORMAL_CMD (editusr, A_ZERO, A_ONE);
NOINTR_CMD (lockusr, A_ZERO, A_MANY);
NOINTR_CMD (unlockusr, A_ZERO, A_MANY);
NOINTR_CMD (designateadmin, A_MANY, A_MANY);
NOINTR_CMD (revokeadmin, A_ZERO, A_MANY);
NORMAL_CMD (notifyusr, A_ZERO, A_MANY);
NORMAL_CMD (setpasswd, A_ZERO, A_ONE);

NOEXEC_CMD (viewgrp, A_ONE, A_ZERO);
NORMAL_CMD (creategrp, A_ZERO, A_ZERO);
NOINTR_CMD (delgrp, A_MANY, A_ZERO);
NORMAL_CMD (editgrp, A_ONE, A_ZERO);

NOINTR_CMD (addusrtogrp, A_MANY, A_MANY);
NOINTR_CMD (rmusrfromgrp, A_MANY, A_MANY);

/* A command interface.
   Every command corresponds to at least one interface.
   There may be at most one interface for a particular command at a 
   particular user level (L_SYS, L_GRP, L_USR).
   Typically when there are multiple interfaces for a command it is
   because different user levels have different authority (i.e., a 
   group admin can only modify his own users). */
struct admin_interface
  {
    /* Basic info. */
    const struct admin_command *command;	/* Corresponding command. */
    const char *title;			/* Command title. */

    /* User prompt information.
       The user will be prompted to select users if necessary. */
    enum admin_user usr;		/* Users required. */
    const char *usr_prompt;		/* Description of users required. */

    /* Group prompt information.
       The user will be prompted to select groups if necessary. */
    enum admin_group grp;		/* Groups required. */
    const char *grp_prompt;		/* Description of groups required. */

    /* Confirmation information.
       If these are non-null then the user will be prompted to confirm
       his action before being allowed to continue. */
    const char *warning_prompt;		/* Warning message. */
    const char *confirm_prompt;		/* Confirmation question. */
  };

#define ADMIN_MENU				\
    {&c_menu, N_("Administration Menu..."),	\
     U_NONE, NULL,				\
     G_NONE, NULL,				\
     NULL, NULL}

#define SEPARATOR(TITLE)			\
    {&c_separator, TITLE,			\
     U_NONE, NULL,				\
     G_GRP, NULL,				\
     NULL, NULL}

/* Interfaces for normal users (L_USR). */
static const struct admin_interface usr_interface[] =
  {
    ADMIN_MENU,

    {&c_editusr, N_("Edit Your Profile"),
     U_ME, NULL,
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewusr, N_("View Profile of User in Group `%s'"),
     U_MY_USR, N_("to view"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewgrp, N_("View Group `%s' Membership"),
     U_NONE, NULL,
     G_MINE, NULL,
     NULL, NULL},
  };

/* Interfaces for group administrators (L_GRP). */
static const struct admin_interface grp_interface[] =
  {
    ADMIN_MENU,

    SEPARATOR (N_("View Users and Groups:")),

    {&c_viewusr, N_("View User in Group `%s'"),
     U_MY_USR, N_("to view"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewusr, N_("View Any User or Group"),
     U_USR, N_("to view"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewusrs, N_("View All Users"),
     U_NONE, NULL,
     G_NONE, NULL,
     NULL, NULL},

    SEPARATOR (N_("Group Maintenance for Group `%s'")),

    {&c_editgrp, N_("Edit Group Profile"),
     U_NONE, NULL,
     G_MINE, NULL,
     NULL, NULL},

    SEPARATOR (N_("User Maintenance in Group `%s'")),

    {&c_createusr, N_("Create New User"),
     U_NONE, NULL,
     G_MINE, NULL,
     NULL, NULL},

    {&c_delusr, N_("Delete User Belonging Only to Group `%s'"),
     U_ONLY_MY_USRS, N_("whose accounts are to be deleted, who must be\n"
			"users belonging only to this group"),
     G_NONE, NULL,
     N_("Deleting a account is a permanent action that cannot be easily or\n"
	"reliably undone."),
     N_("delete the selected user accounts")},

    {&c_editusr, N_("Edit User Profile"),
     U_MY_USR, N_("to edit"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_addusrtogrp, N_("Add Users to Group `%s'"),
     U_USRS, N_("who are to be added to the group"),
     G_MINE, NULL,
     NULL, NULL},

    {&c_rmusrfromgrp, N_("Remove Users from Group `%s'"),
     U_MY_USRS, N_("who are to be removed from this group"),
     G_MINE, NULL,
     NULL, NULL},

    {&c_designateadmin, N_("Designate Group Administrator for Group `%s'"),
     U_MY_USRS, N_("who are to be designated as group administrators\n"
		   "within this group"),
     G_MINE, NULL,
     N_("As a group administrator, you have the power to share your\n"
	"group administration privileges with other members of your group.\n"
	"However, once these privileges are granted, you cannot revoke\n"
	"them without help from one of the higher-level TeamSCOPE\n"
	"administrators.  Keep this in mind as you grant group\n"
	"administration privileges."),
     N_("grant group administration privileges to the selected users")}, 

    {&c_notifyusr, N_("Notify User by Email/Reset Password"),
     U_MY_USRS, N_("who should be notified by email of account information\n"
		   "and whose password should possibly be reset"),
     G_NONE, NULL,
     NULL, NULL},
  };

/* Interfaces for system administrators (L_SYS). */
static const struct admin_interface sys_interface[] =
  {
    ADMIN_MENU,
    SEPARATOR (N_("View Users and Groups:")),

    {&c_viewusr, N_("View User in Group `%s'"),
     U_MY_USR, N_("to view"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewusr, N_("View Any User or Group"),
     U_USR, N_("to view"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_viewusrs, N_("View All Users"),
     U_NONE, NULL,
     G_NONE, NULL,
     NULL, NULL},

    SEPARATOR (N_("Group Maintenance:")),

    {&c_creategrp, N_("Create New Group"),
     U_NONE, NULL,
     G_NONE, NULL,
     NULL, NULL},

    {&c_delgrp, N_("Delete Group"),
     U_NONE, NULL,
     G_GRPS, N_("to delete"),
     N_("Deleting a group is a permanent action that cannot be easily or\n"
	"reliably undone."),
     N_("delete the selected groups")},

    {&c_editgrp, N_("Edit Group Profile"),
     U_NONE, NULL,
     G_GRP, N_("to edit"),
     NULL, NULL},

    SEPARATOR (N_("User Maintenance In Group `%s':")),

    {&c_createusr, N_("Create New User"),
     U_NONE, NULL,
     G_MINE, NULL,
     NULL, NULL},

    {&c_delusr, N_("Delete Users"),
     U_MY_USRS, N_("whose accounts are to be deleted"),
     G_NONE, NULL,
     N_("Deleting a account is a permanent action that cannot be easily or\n"
	"reliably undone."),
     N_("delete the selected user accounts")},

    {&c_lockusr, N_("Lock User Accounts"),
     U_MY_USRS, N_("whose accounts are to be locked"),
     G_NONE, NULL,
     N_("Locking user accounts denies users' access to all TeamSCOPE\n"
	"features and freezes their account settings, for all groups\n"
	"to which the users belong.  It is most suitable for use for\n"
	"closing accounts without deleting them and for temporary\n"
	"use while investigating security violations."),
     N_("lock the selected user accounts")},

    {&c_unlockusr, N_("Unlock User Accounts"),
     U_MY_USRS, N_("whose accounts are to be unlocked"),
     G_NONE, NULL,
     N_("Unlocking user accounts re-enable users' access to all\n"
	"TeamSCOPE features, for all groups to which the users belong.\n"
	"Before unlocking accounts, be certain that the users are not\n"
	"being investigated for security violations."),
     N_("unlock the selected user accounts")},

    {&c_setpasswd, N_("Set User Password"),
     U_MY_USR, N_("whose password is to be changed"),
     G_NONE, NULL,
     N_("Setting a user's password on his (or her) behalf is not\n"
	"recommended practice because of the security risk in\n"
	"sharing the password.  If possible, please use the\n"
	"<STRONG>Notify User by Email/Reset Password</STRONG>\n"
	"function instead, which randomly generates a new password."),
     N_("set this user's password")}, 

    {&c_addusrtogrp, N_("Add Users to Group"),
     U_USRS, N_("who are to be added to the group"),
     G_MINE, NULL,
     NULL, NULL},

    {&c_rmusrfromgrp, N_("Remove Users from Group"),
     U_MY_USRS, N_("who are to be removed from this group"),
     G_MINE, NULL,
     NULL, NULL},

    {&c_designateadmin, N_("Designate Group Administrators"),
     U_USRS, N_("who are to be designated as group administrators"),
     G_MINE, NULL,
     NULL, NULL},

    {&c_notifyusr, N_("Notify User by Email/Reset Password"),
     U_MY_USRS, N_("who should be notified by email of account information\n"
		   "and whose password should possibly be reset"),
     G_NONE, NULL,
     NULL, NULL},

    SEPARATOR (N_("User Maintenance in Any Group:")),

    {&c_createusr, N_("Create New User"),
     U_NONE, NULL,
     G_GRPS, N_("to which the new user should belong"),
     NULL, NULL},

    {&c_delusr, N_("Delete User Account"),
     U_USRS, N_("whose accounts are to be deleted"),
     G_NONE, NULL,
     N_("Deleting an account is a permanent action that cannot be easily or\n"
	"reliably undone."),
     N_("delete the selected user accounts")},

    {&c_lockusr, N_("Lock User Accounts"),
     U_USRS, N_("whose accounts are to be locked"),
     G_NONE, NULL,
     N_("Locking user accounts denies users' access to all TeamSCOPE\n"
	"features and freezes their account settings, for all groups\n"
	"to which the users belong.  It is most suitable for use for\n"
	"closing accounts without deleting them and for temporary\n"
	"use while investigating security violations."),
     N_("lock the selected user accounts")},

    {&c_unlockusr, N_("Unlock User Accounts"),
     U_USRS, N_("whose accounts are to be unlocked"),
     G_NONE, NULL,
     N_("Unlocking user accounts re-enable users' access to all\n"
	"TeamSCOPE features, for all groups to which the users belong.\n"
	"Before unlocking accounts, be certain that the users are not\n"
	"being investigated for security violations."),
     N_("unlock the selected user accounts")},

    {&c_editusr, N_("Edit User Profile"),
     U_USR, N_("to edit"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_setpasswd, N_("Set User Password"),
     U_USR, N_("whose password is to be changed"),
     G_NONE, NULL,
     N_("Setting a user's password on his (or her) behalf is not\n"
	"recommended practice because of the security risk in\n"
	"sharing the password.  If possible, please use the\n"
	"<STRONG>Notify User by Email/Reset Password</STRONG>\n"
	"function instead, which randomly generates a new password."),
     N_("set this user's password")}, 

    {&c_addusrtogrp, N_("Add Users to Groups"),
     U_USRS, N_("who are to be added to new groups"),
     G_GRPS, N_("to which the selected users are to be added"),
     NULL, NULL},

    {&c_rmusrfromgrp, N_("Remove Users from Groups"),
     U_USRS, N_("who are to be removed from groups"),
     G_GRPS, N_("from which the selected users are to be removed"),
     NULL, NULL},

    {&c_designateadmin, N_("Designate Group Administrators"),
     U_USRS, N_("who are to be designated as group administrators"),
     G_GRPS, N_("in which the selected users should be designated\n"
		"group administrators"),
     NULL, NULL},

    {&c_revokeadmin, N_("Revoke Group Administrators"),
     U_USRS, N_("whose group administration privileges are to be\n"
		"revoked in all groups"),
     G_NONE, NULL,
     NULL, NULL},

    {&c_notifyusr, N_("Notify User by Email/Reset Password"),
     U_USRS, N_("who should be notified by email of account information\n"
		"and whose password should possibly be reset"),
     G_NONE, NULL,
     NULL, NULL},
  };

static const struct admin_interface *interface[L_CNT] =
  {
    usr_interface,
    grp_interface,
    sys_interface,
  };

static const int interface_cnt[L_CNT] =
  {
    sizeof usr_interface / sizeof *usr_interface,
    sizeof grp_interface / sizeof *grp_interface,
    sizeof sys_interface / sizeof *sys_interface,
  };


/* Encapsulates information about a administration CGI request. */
struct admin_request
  {
    struct request *rq;			/* Request. */
    enum admin_level level;		/* User level of user. */
    const struct admin_interface *i;	/* Selected interface. */
    struct id_list grp;			/* Selected groups. */
    struct id_list usr;			/* Selected users. */
  };

#if GLOBAL_DEBUGGING
static void validate_interfaces (void);
#endif

static void show_menu (struct admin_request *);
static void parse_command (struct admin_request *);
static int select_grp (struct admin_request *);
static int select_usr (struct admin_request *);
static int confirm (struct admin_request *);
static void shell (struct std *, const char *command, ...)
     attribute ((format (printf, 2, 3)));

/* Main administration screen code. */

static void init_admin_request (struct admin_request *, struct request *);
static void destroy_admin_request (struct admin_request *);

void
scr_admin (struct request *rq)
{
  struct admin_request ar;

  /* At this point we are running with root privileges.
     We do as little as possible here, then drop privileges. */

#if GLOBAL_DEBUGGING
  validate_interfaces ();
#endif

  http_header ();
  html_banner (rq);

  init_admin_request (&ar, rq);
  if (std_contains (ar.rq->trusted, "system-admin"))
    ar.level = L_SYS;
  else if (std_contains (ar.rq->trusted, "group-admin"))
    ar.level = L_GRP;

  /* Select the current interface based on the operation selected by
     the user. */
  ar.i = interface[ar.level];
  {
    int op = stoi (rq->op);
    if (op >= 0 && op < interface_cnt[ar.level])
      ar.i += op;
  }
	
  show_menu (&ar);
  parse_command (&ar);
  if (!select_usr (&ar) || !select_grp (&ar) || !confirm (&ar))
    return;

  if (ar.i->command->execute != NULL && ar.i->command->execute (&ar))
    {
      drop_privs_finish_init (ar.rq);
      return;
    }

  drop_privs_finish_init (ar.rq);

  if (ar.i->command->interact)
    ar.i->command->interact (&ar);
  else
    show_request (&ar, 1);

  destroy_admin_request (&ar);
}

/* Initialize AR for RQ. */
static void
init_admin_request (struct admin_request *ar, struct request *rq)
{
  ar->rq = rq;
  ar->level = L_USR;
  ar->i = NULL;
  id_list_null (&ar->grp);
  id_list_null (&ar->usr);
}

/* Destroy storage for AR. */
static void
destroy_admin_request (struct admin_request *ar)
{
  id_list_destroy (&ar->grp);
  id_list_destroy (&ar->usr);
}

/* Generic user interaction. */

/* Parses the CGI parameters and determines what groups and users have
   been selected.  Disallows selection of groups and users prohibited
   by the selected interface. */
static void 
parse_command (struct admin_request *ar)
{
  switch (ar->i->grp)
    {
    case G_NONE:
      id_list_null (&ar->grp);
      break;

    case G_MINE:
      id_list_one (&ar->grp, ar->rq->cx.grp);
      break;

    case G_GRP:
    case G_GRPS:
      id_list_from_std (&ar->grp, ar->rq->cgi, "grp", ar->i->grp == G_GRP);
      break;

    default:
      abort ();
    }
  
  switch (ar->i->usr)
    {
    case U_NONE:
      id_list_null (&ar->usr);
      break;

    case U_ME:
      id_list_one (&ar->usr, ar->rq->cx.usr);
      break;

    case U_USR:
    case U_USRS:
      id_list_from_std (&ar->usr, ar->rq->cgi, "usr", ar->i->grp == U_USR);
      break;

    case U_ONLY_MY_USRS:
      {
	struct id_list t1, t2;
	
	id_list_one_group (&t1);
	id_list_for_group (&t2, ar->rq->cx.grp);
	id_list_intersect (&t1, &t2);
	id_list_from_std (&t2, ar->rq->cgi, "usr", 0);
	id_list_intersect (&t1, &t2);
	ar->usr = t1;
      }
      break;

    case U_MY_USRS:
    case U_MY_USR:
      {
	struct id_list t1, t2;
	
	id_list_for_group (&t1, ar->rq->cx.grp);
	id_list_from_std (&t2, ar->rq->cgi, "usr", ar->i->grp == U_MY_USR);
	id_list_intersect (&t1, &t2);
	ar->usr = t1;
	break;
      }

    default:
      abort ();
    }
}
  
/* Displays a drop-down menu for request AR. */
static void
show_menu (struct admin_request *ar)
{
  const struct admin_interface *i;

  html_start_form ("admin", HT_GET, HT_URL_ENCODED);
  htmls ("<CENTER>\n"
	 "<SELECT NAME=op ONCHANGE=\"form.submit()\">\n");

  for (i = interface[ar->level];
       i < interface[ar->level] + interface_cnt[ar->level]; i++)
    {
      if (i->command != &c_separator)
	html ("  <OPTION VALUE=%d%s>",
	      (int) (i - interface[ar->level]),
	      ar->i == i ? " SELECTED" : "");
      else
	htmls ("  <OPTION DISABLED>--");
      html (gettext (i->title), ar->rq->cx.grp);
      if (i->command == &c_separator)
	htmls ("--"); 
      htmls ("</OPTION>\n");
    }
  
  htmls ("</SELECT>\n"
	 "<INPUT TYPE=submit VALUE=\"Go!\">\n"
	 "</CENTER>\n");
  html_end_form ();
}

/* Displays the groups listed in R as a table.
   If ARGS is A_ZERO then doesn't allow any selection.
   If ARGS is A_ONE then allows one selection as a link,
   using URL as a base url.
   If ARGS is A_MANY then allows multiple selections as check boxes.
*/
static void
show_groups (db_result *r, enum admin_needs_args args,
	     struct url *url)
{
  size_t i;
  int c = dbr_ncols (r);
  
  html ("<TABLE BGCOLOR=\"#FFFFFF\" CELLSPACING=0 CELLPADDING=0 BORDER=1>\n"
	"  <TR>\n"
	"    <TD></TD>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n",
	_("Group"),
	_("Full Name"));
  if (c > 2)
    html ("    <TH>%s</TH>\n"
	  "    <TH>%s</TH>\n",
	  _("Notified<BR>of Account?"),
	  _("Group<BR>Admin?"));
  htmls ("  </TR>\n");

  if (url != NULL)
    url_mark (url);
  for (i = 0; i < dbr_nrows (r); i++)
    {
      const char *grp = dbr_value (r, i, 0);
      
      htmls ("  <TR>\n");

      if (args != A_MANY)
	htmls ("  <TD></TD>\n");
      else
	html ("  <TD><INPUT TYPE=checkbox NAME=grp VALUE=%s></TD>\n", grp);

      if (args != A_ONE)
	html ("    <TD>%s</TD>\n", grp);
      else
	{
	  url_release (url);
	  url_arg (url, "grp", grp);
	  html ("    <TD><A HREF=\"%s\">%s</A></TD>\n", url_string (url), grp);
	}
      
      html ("    <TD>%s</TD>\n", dbr_value (r, i, 1));
      if (c > 2)
	html ("    <TD>%s</TD>\n"
	      "    <TD>%s</TD>\n",
	      dbr_value (r, i, 2)[0] == 't' ? "yes" : "<STRONG>no</STRONG>",
	      dbr_value (r, i, 3)[0] == 't' ? "<STRONG>yes</STRONG>" : "no");
      htmls ("  </TR>\n");
    }
  
  html ("</TABLE>\n");
}

/* Displays the list of N users in INFO.
   Allows up to QTY of them to be selected.
   The list is assumed to be sorted by the SORT_BY'th column,
   and URL is a base URL to reproduce this screen, if URL is non-null. */
static void
show_user_list (struct user_info *info, size_t n,
		enum admin_needs_args qty,
		struct url *url, int sort_by)
{
  size_t i;
  
  html ("<TABLE BGCOLOR=\"#FFFFFF\" CELLSPACING=0 CELLPADDING=0 BORDER=1>\n"
	"  <TR>\n"
	"    <TD></TD>\n");
  if (url != NULL)
    url_mark (url);
  for (i = 0; i < 5; i++)
    {
      static const char *heads[] = 
	{
	  N_("User"),
	  N_("Full Name"),
	  N_("Email<BR>Recordable?"),
	  N_("Account<BR>Locked?"),
	  N_("Last Usage"),
	};

      if (url == NULL || sort_by == (int) i)
	html ("    <TH>%s</TH>\n", heads[i]);
      else
	{
	  url_release (url);
	  url_arg_int (url, "sortby", i);
	  html ("    <TH><A HREF=\"%s\">%s</A></TH>\n",
		url_string (url), heads[i]);
	}
    }
  
  htmls ("  </TR>\n");

  for (i = 0; i < n; i++, info++)
    {
      htmls ("  <TR>\n");

      if (qty != A_MANY)
	htmls ("  <TD></TD>\n");
      else
	html ("  <TD><INPUT TYPE=checkbox NAME=usr VALUE=%s></TD>\n",
	      info->usr);

      if (qty == A_ONE)
	{
	  url_release (url);
	  url_arg (url, "usr", info->usr);
	  html ("  <TD><A HREF=\"%s\">%s</A></TD>\n",
		url_string (url), info->usr);
	}
      else
	html ("    <TD>%s</TD>\n", info->usr);
      
      html ("    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n"
	    "  </TR>\n",
	    info->fullname,
	    info->recordable ? "yes" : "no",
	    info->locked ? "<STRONG>yes</STRONG>" : "no",
	    info->last_usage);
    }
  
  html ("</TABLE>\n");
}

/* Makes sure that appropriate group(s) are selected for the current
   interface, offering a list if not.
   Returns nonzero if everything's kosher, or zero if something needs to
   be selected and isn't. */
static int
select_grp (struct admin_request *ar)
{
  db_result *r;
  enum admin_needs_args qty;
  
  if (ar->i->grp == G_NONE || ar->grp.n > 0)
    return 1;

  drop_privs_finish_init (ar->rq);

  qty = ar->i->grp == G_GRP ? A_ONE : A_MANY;

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P ALIGN=CENTER>%s %s%s.</P>\n",
	(qty == A_ONE ? _("Select Target Group")
	 : _("Select Target Group(s)")),
	(qty == A_ONE ? _("Click on the group")
	 : _("Select the group or groups")),
	gettext (ar->i->grp_prompt),
	(qty == A_ONE ? ""
	 : _(", then click on the <STRONG>Submit</STRONG> button")));

  db_query ("SELECT grp, name FROM groups ORDER BY grp", DB_TUPLES, &r);

  htmls ("<CENTER>\n");
  if (qty == A_MANY)
    {
      html_start_form ("admin", HT_POST, HT_URL_ENCODED);
      html_submit (_("Submit"));
      url_arg_std (NULL, ar->rq->cgi);
      show_groups (r, qty, NULL);
      html_submit (_("Submit"));
      html_end_form ();
    }
  else
    {
      struct url url;

      url_create (&url, "admin");
      url_arg_std (&url, ar->rq->cgi);
      show_groups (r, qty, &url);
      url_destroy (&url);
    }
  htmls ("</CENTER>\n");

  dbr_destroy (r);

  return 0;
}

/* When users can be selected from any group,
   this offers a list of groups to select from.
   It also offers selection from a list of all users. */
static void
offer_by_group_or_all (struct admin_request *ar, enum admin_needs_args qty)
{
  struct url url;
  db_result *r;
  size_t i;

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P ALIGN=CENTER>%s %s %s</P>\n\n"
	"<UL>\n",
	qty == A_ONE ? _("Select Target User") : _("Select Target Users"),
	(qty == A_ONE
	 ? _("You may select the user") : ("You may select the users")),
	ar->i->usr_prompt,
	_("either by group membership or from a list of all users.\n"
	  "Make your choice below."));

  url_create (&url, "admin");
  url_arg (&url, "more", NULL);
  url_arg_std (&url, ar->rq->cgi);
  url_mark (&url);
  url_arg (&url, "allusers", NULL);
  html ("<LI><A HREF=\"%s\">%s</A>\n", url_string (&url), _("All users"));

  db_query ("SELECT grp, name FROM groups ORDER BY grp", DB_TUPLES, &r);

  for (i = 0; i < dbr_nrows (r); i++)
    {
      const char *grp = dbr_value (r, i, 0);

      url_release (&url);
      url_arg (&url, "fromgrp", grp);
      html ("<LI><A HREF=\"%s\">%s: %s</A>\n",
	     url_string (&url), grp, dbr_value (r, i, 1));
    }

  dbr_destroy (r);

  htmls ("</UL>\n");
}

/* Column to sort by.
   Used by compare_user_info() as a gross hack. */
static int static_sort_by;

/* Returns a pointer to the last word in S.
   If there's only one word, then this is S. */
static const char *
last_word (const char *s)
{
  const char *lw = strrchr (s, ' ');
  return lw != NULL ? lw + 1 : s;
}

/* Compares user_info structures at PA and PB
   based on column static_sort_by. */
static int
compare_user_info (const void *pa, const void *pb)
{
  const struct user_info *a = pa;
  const struct user_info *b = pb;
  int cmp;

  switch (static_sort_by)
    {
    case 0:
    default:
      cmp = strcmp (a->usr, b->usr);
      break;

    case 1:
      cmp = strcmp (last_word (a->fullname), last_word (b->fullname));
      if (cmp == 0)
	cmp = strcmp (a->fullname, b->fullname);
      break;

    case 2:
      cmp = b->recordable - a->recordable;
      if (cmp == 0)
	cmp = strcmp (a->usr, b->usr);
      break;

    case 3:
      cmp = b->locked - a->locked;
      if (cmp == 0)
	cmp = strcmp (a->usr, b->usr);
      break;

    case 4:
      {
	time_t ta = datetime_parse (a->last_usage, 0, 0, NULL);
	time_t tb = datetime_parse (b->last_usage, 0, 0, NULL);
	if (tb > ta)
	  cmp = 1;
	else if (tb < ta)
	  cmp = -1;
	else
	  cmp = 0;
      }
      break;
    }

  return cmp;
}

/* Displays a screen allowing QTY users to be selected from within IDS.  
   The users are sorted based on SORT_BY using base url URL.
   If MORE is nonzero then more users may be selected after finishing
   with this screen, and a button is added for that purpose.
   If QTY is A_ZERO then users are displayed but cannot be selected. */
static void
show_users (struct admin_request *ar, struct id_list *ids,
	    enum admin_needs_args qty, int more,
	    struct url *url, int sort_by)
{
  struct user_info *info;

  if (qty != A_ZERO)
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	    "<P ALIGN=CENTER>%s %s %s",
	    qty == A_ONE ? _("Select Target User") : _("Select Target Users"),
	    qty == A_ONE ? _("Click on the user") : ("Select the users"),
	    ar->i->usr_prompt,
	    (qty == A_ONE ? ""
	     : _("from the list below, then click on the\n"
		 "<STRONG>Submit</STRONG> button")));
      if (qty == A_MANY && more)
	htmls (_(", or, if you wish to select more users from other\n"
		 "groups as well, click on the <STRONG>More Users</STRONG>\n"
		 "button.\n"));
      else
	htmls (".\n");
      htmls ("</P>\n\n");
    }
  
  info = xmalloc (sizeof *info * ids->n);
  get_user_info (ids, info, UI_ALL);

  static_sort_by = sort_by;
  qsort (info, ids->n, sizeof *info, compare_user_info);

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);
  htmls ("<CENTER>\n");
  if (qty == A_MANY)
    {
      html_submit (_("Submit"));
      if (more)
	html_submit_alt ("more", _("More Users"));
    }
  url_arg_std (NULL, ar->rq->cgi);
  show_user_list (info, ids->n, qty, url, sort_by);
  if (qty == A_MANY)
    {
      html_submit (_("Submit"));
      if (more)
	html_submit_alt ("more", _("More Users"));
    }
  htmls ("</CENTER>\n");
  html_end_form ();

  free_user_info (info, ids->n);
  free (info);
}

/* Handles user selection.
   Returns nonzero if everything's good,
   or zero if something needs to be selected that isn't. */
static int
select_usr (struct admin_request *ar)
{
  enum admin_needs_args qty;
  struct url url;
  int sort_by;

  if (ar->i->usr == U_NONE || ar->i->usr == U_ME)
    qty = A_ZERO;
  else if (ar->i->usr == U_USR || ar->i->usr == U_MY_USR)
    qty = A_ONE;
  else
    qty = A_MANY;

  if (qty == A_ZERO
      || (qty == A_ONE && ar->usr.n > 0)
      || (qty == A_MANY && ar->usr.n > 0
	  && !std_contains (ar->rq->cgi, "more")))
    return 1;

  /* If we reach this point, we need to ask the user to select one or
     more users. */
  drop_privs_finish_init (ar->rq);

  sort_by = stoi (std_get (ar->rq->cgi, "sortby"));
  if (sort_by < 0 || sort_by > 4)
    sort_by = 0;
  std_delete (ar->rq->cgi, "sortby");
  url_create (&url, "admin");
  url_arg_std (&url, ar->rq->cgi);
  
  std_delete (ar->rq->cgi, "more");

  switch (ar->i->usr)
    {
    case U_ONLY_MY_USRS:
      {
	struct id_list t1, t2;
	
	id_list_one_group (&t1);
	id_list_for_group (&t2, ar->rq->cx.grp);
	id_list_intersect (&t1, &t2);
	show_users (ar, &t1, qty, 0, &url, sort_by);
      }
      break;
      
    case U_MY_USR:
    case U_MY_USRS:
      {
	struct id_list t1;
      
	id_list_for_group (&t1, ar->rq->cx.grp);
	show_users (ar, &t1, qty, 0, &url, sort_by);
      }
      break;

    case U_USR:
    case U_USRS:
      if (std_contains (ar->rq->cgi, "fromgrp"))
	{
	  struct id_list ids;

	  id_list_for_group (&ids, std_get (ar->rq->cgi, "fromgrp"));
	  std_delete (ar->rq->cgi, "fromgrp");
	  show_users (ar, &ids, qty, ar->i->usr == U_USRS, &url, sort_by);
	}
      else if (std_contains (ar->rq->cgi, "allusers"))
	{
	  struct id_list ids;

	  std_delete (ar->rq->cgi, "allusers");
	  id_list_all (&ids);
	  show_users (ar, &ids, qty, 0, &url, sort_by);
	}
      else
	offer_by_group_or_all (ar, qty);
      break;

    default:
      assert (0);
    }

  url_destroy (&url);

  return 0;
}

/* Makes sure that the selected interface is confirmed, if need be.
   Returns zero if confirmation is still needed,
   nonzero otherwise. */
static int
confirm (struct admin_request *ar)
{
  if (ar->i->confirm_prompt == NULL
      || !strcmp (std_get (ar->rq->cgi, "confirmed"), "yes"))
    return 1;

  drop_privs_finish_init (ar->rq);

  html ("<H2 ALIGN=CENTER>%s: ", _("Confirm"));
  html (gettext (ar->i->title), ar->rq->cx.grp);
  htmls ("</H2>\n\n");

  show_request (ar, 1);

  html ("<P>%s</P>\n\n"
	"<P>%s %s?</P>\n\n",
	gettext (ar->i->warning_prompt),
	_("Do you really want to"),
	gettext (ar->i->confirm_prompt));
  std_delete (ar->rq->cgi, "confirmed");

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);
  url_arg_std (NULL, ar->rq->cgi);
  html ("<P>\n"
	"<LABEL><INPUT TYPE=radio NAME=confirmed VALUE=no CHECKED>%s\n"
	"%s.</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=confirmed VALUE=yes>%s\n"
	"%s.</LABEL>\n"
	"</P>\n",
	_("<STRONG>No</STRONG>, it would be a mistake to"),
	ar->i->confirm_prompt,
	_("<STRONG>Yes</STRONG>, I really want to"),
	ar->i->confirm_prompt);
  html_submit (_("Submit"));
  html_end_form ();

  return 0;
}

/* Displays users and groups selected in request AR
   as an HTML description list.
   If TERMINATE is zero then the list is not terminated,
   so that the caller can add to it. */
static void
show_request (struct admin_request *ar, int terminate)
{
  htmls ("<DL>\n");
  
  if (ar->usr.n)
    {
      html ("<DT>%s", ar->i->command->usr > A_ONE ? _("Users") : _("User"));
      if (ar->i->usr_prompt != NULL)
	html (" %s", ar->i->usr_prompt);
      htmls (":\n"
	     "<DD>");

      id_list_print (&ar->usr);
    }
  
  if (ar->grp.n)
    {
      html ("<DT>%s", ar->i->command->grp > A_ONE ? _("Groups") : _("Group"));
      if (ar->i->grp_prompt != NULL)
	html (" %s", ar->i->grp_prompt);
      htmls (":\n"
	     "<DD>");

      id_list_print (&ar->grp);
    }

  if (terminate)
    htmls ("</DL>\n\n");
}

/* Displays a message indicating that the selected command
   was successfully executed.
   TERMINATE is passed to show_request(). */
static void 
success (struct admin_request *ar, int terminate)
{
  html ("<H2 ALIGN=CENTER>");
  html (gettext (ar->i->title), ar->rq->cx.grp);
  html (": %s</H2>\n"
	"<P>%s</P>\n",
	_("Success"),
	_("The requested operation completed successfully.  "
	  "You may select another task from the menu above.  "
	  "Any details of the operation are listed below."));

  show_request (ar, terminate);
}

/* Calls `scope-db update' for use after add/removing users/groups. */
static void
update (void)
{
  html_flush ();
  shell (0, "scope-db update");
}

/* For use by individual commands for interaction. */

/* Presents a form to the user.
   INPUT[][0] is form element names.
   INPUT[][1] is form element labels.
   INPUT[][2] is form element types; typically "text SIZE=##" or "checkbox".
   INPUT[][3] is form element default values.
   INPUT[][4] is form element descriptions; these should begin
   with <BR> except for checkboxes.
   Checkbox fields will have value `yes' when checked.
   When the form is submitted it will have `execute' set as a CGI field.
   The form and the table must be started before this function is called;
   that way, fixed information can be put at the top of the table. */
static void
present_form (struct admin_request *ar, const char *input[][5], size_t n)
{
  size_t i;
  int from_cgi = std_contains (ar->rq->cgi, "execute");

  for (i = 0; i < n; i++)
    {
      const char *name = input[i][0];
      const char *label = input[i][1];
      const char *type = input[i][2];
      const char *def = input[i][3];
      const char *description = input[i][4];
      
      html ("  <TR>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD><INPUT NAME=%s TYPE=%s",
	    gettext (label), name, type);

      if (strncmp (type, "checkbox", 8))
	{
	  const char *value = from_cgi ? std_get (ar->rq->cgi, name) : def;

	  if (*value)
	    {
	      htmls (" VALUE=\"");
	      html_escape (value);
	      htmlc ('"');
	    }
	}
      else
	{
	  int checked = (from_cgi
			 ? !strcmp (std_get (ar->rq->cgi, name), "yes")
			 : *def == 'y' || *def == 't');

	  htmls (" VALUE=yes");
	  if (checked)
	    htmls (" CHECKED");
	}
	    
      html (">\n"
	    "      <SMALL>%s\n"
	    "      </SMALL>\n"
	    "    </TD>\n"
	    "  </TR>\n\n",
	    description);
      std_delete (ar->rq->cgi, name);
    }
  url_arg_std (NULL, ar->rq->cgi);
  html_hidden ("execute", "yes");
  
  htmls ("</TABLE>\n");

  html_submit (_("Submit"));

  html_end_form ();
}

/* Displays validation error MESSAGE and returns zero. */
static int
validate_error (const char *message)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s: %s  %s</P>\n",
	_("Input Validation Error"),
	_("There was an error in the data supplied"),
	message,
	_("Please correct this error and resubmit the form."));

  return 0;
}

/* Displays validation warning MESSAGE. */
static void
validate_warning (const char *message)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s: %s  %s</P>\n",
	_("Input Validation Warning"),
	_("Suspect but acceptable data were supplied"),
	message,
	_("The action was performed anyhow."));
}

/* Returns nonzero if NAME is an acceptable user or group name. */
static int
validate_name (const char *name)
{
  if (strlen (name) < 1 || strlen (name) > 8)
    return 0;
  if (!islower ((unsigned char) *name))
    return 0;
  for (; *name; name++)
    if (!islower ((unsigned char) *name)
	&& !isdigit ((unsigned char) *name)
	&& *name != '-' && *name != '_')
      return 0;
  return 1;
}

static void
interact_menu (struct admin_request *ar)
{
  const struct admin_interface *i;
  struct url url;

  html ("<P ALIGN=CENTER>%s</P>\n"
	"<CENTER>\n",
	_("Select an action from the drop-box box above, "
	  "or click on one of the choices below:"));

  url_create (&url, "admin");
  url_mark (&url);
  for (i = interface[ar->level] + 1;
       i < interface[ar->level] + interface_cnt[ar->level]; i++)
    {
      if (i->command != &c_separator)
	{
	  url_release (&url);
	  url_arg_int (&url, "op", i - interface[ar->level]);
	  html ("<A HREF=\"%s\">", url_string (&url));
	  html (gettext (i->title), ar->rq->cx.grp);
	  htmls ("</A><BR>\n");
	}
      else
	{
	  htmls ("<P><BIG>");
	  html (gettext (i->title), ar->rq->cx.grp);
	  htmls ("</BIG><BR>\n");
	}
    }
  url_destroy (&url);
  
  htmls ("</CENTER>\n");
}

static void
interact_viewusr (struct admin_request *ar)
{
  struct user_info info;
  db_result *r;

  assert (ar->usr.n == 1);
  get_user_info (&ar->usr, &info, UI_ALL);

  html ("<TABLE><TR><TD>%s<TD>%s\n"
	"<TR><TD>%s<TD>%s\n"
	"<TR><TD>%s<TD>%s\n"
	"<TR><TD>%s<TD>%s\n"
	"<TR><TD>%s<TD>%s\n"
	"<TR><TD>%s<TD>%s\n"
	"</TABLE>\n",
	_("User:"), ar->usr.s[0],
	_("Full Name:"), info.fullname,
	_("Upstream Email:"), info.email ? info.email : "(none)",
	_("Email Recordable?"), info.recordable ? "yes" : "no",
	_("Account Locked?"), info.locked ? "<STRONG>yes</STRONG>" : "no",
	_("Last Usage:"), info.last_usage);

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n", _("Group Membership"));

  db_query ("SELECT user_group.grp, groups.name, sent_email, may_admin "
	    "FROM groups, user_group "
	    "WHERE usr = '%sq' AND user_group.grp = groups.grp "
	    "ORDER BY user_group.grp",
	    DB_FORMAT | DB_TUPLES, &r, ar->usr.s[0]);
  show_groups (r, A_ZERO, NULL);

  free_user_info (&info, 1);
}

static void
interact_viewusrs (struct admin_request *ar)
{
  struct id_list usrs;

  id_list_all (&usrs);
  show_users (ar, &usrs, A_ZERO, 0, NULL, -1);
  id_list_destroy (&usrs);
}

static void
interact_createusr (struct admin_request *ar unused)
{
  static const char *input[][5] = 
    {
      {"fullname", N_("Full Name"), "text SIZE=40", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "User's full name, first followed by last.")},

      {"email", N_("Upstream Email"), "text SIZE=40", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "User's upstream email address.\n"
	  "If user will receive mail directly from account on this server,\n"
	  "enter <STRONG>local</STRONG>.\n"
	  "This is the email address that account creation notification\n"
	  "will be sent to.")},

      {"username", N_("Username"), "text SIZE=8", "",
       N_("<BR><STRONG>(Optional)</STRONG>\n"
	  "Preferred username.\n"
	  "If supplied, must be between 3 and 8 lowercase letters\n"
	  "and digits and must begin with a letter.\n"
	  "If not supplied, a suitable username will be\n"
	  "generated automatically.")},

      {"notify", N_("Notify User?"), "checkbox", "t",
       N_("Notify user of new account via email?\n"
	  "<BR>If checked, user will be immediately mailed notification of\n"
	  "account creation and instructions for access.\n"
	  "If not checked, account is disabled until user is explicitly\n"
	  "notified.")},
    };

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Create User"),
	_("To complete the process of creating a new user in the selected\n"
	  "group(s), fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  htmls ("<TABLE>\n\n");

  html ("  <TR><TD>%s</TD><TD>", _("Group(s)"));
  id_list_print (&ar->grp);
  htmls ("</TD></TR>\n\n");

  present_form (ar, input, sizeof input / sizeof *input);
}

static int
execute_createusr (struct admin_request *ar)
{
  const char *fullname = std_get (ar->rq->cgi, "fullname");
  const char *email = std_get (ar->rq->cgi, "email");
  const char *username = std_get (ar->rq->cgi, "username");
  const int notify = !strcmp (std_get (ar->rq->cgi, "notify"), "yes");
  struct std *d;

  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;
  
  if (strlen (fullname) < 5)
    return validate_error (_("Full name too short."));
  if (strlen (email) < 3)
    return validate_error (_("Email address too short."));
  if (strlen (username) > 0)
    {
      if (strlen (username) < 3)
	return validate_error (_("Username too short."));
      if (strlen (username) > 8)
	return validate_error (_("Username too long."));
      if (!validate_name (username))
	return validate_error (_("Improperly constructed username."));
      if (getpwnam (username) != NULL || getgrnam (username))
	return validate_error (_("User or group by this name "
				 "already exists."));
    }

  if (!strcmp (email, "local"))
    email = "";
  else if (strchr (email, '@') == NULL)
    validate_warning (_("Email address should normally contain `@'."));

  if (!notify)
    validate_warning (_("Don't forget to notify the new user!"));

  d = std_create (0);
  shell (d, "scope-db adduser '%su' '%su' '%sq' '%sq'",
	 ar->grp.s[0], username, fullname, email);

  if (std_contains (d, "Username"))
    username = std_get (d, "Username");

  {
    size_t i;
    
    for (i = 1; i < ar->grp.n; i++)
      shell (NULL, "scope-db addusertogroup '%su' '%su'",
	     username, ar->grp.s[i]);
  }

  /* This must precede notification because the email alias needs to
     be set up first. */
  update ();

  if (notify)
    shell (NULL, "scope-db mail --no-reset '%su'", username);
  
  success (ar, 0);

  html ("<DT>%s:<DD>%s\n"
	"<DT>%s:<DD>%s\n"
	"<DT>%s:<DD>%s\n",
	_("Full name"), fullname,
	_("Email"), email,
	_("Username"), username);

  std_destroy (d);

  return 1;
}

static int
execute_delusr (struct admin_request *ar)
{
  size_t i;

  for (i = 0; i < ar->usr.n; i++)
    shell (NULL, "scope-db deluser '%su'", ar->usr.s[i]);

  success (ar, 1);
  update ();
  return 1;
}

static int
execute_editusr (struct admin_request *ar)
{
  const char *fullname = std_get (ar->rq->cgi, "fullname");
  const char *email = std_get (ar->rq->cgi, "email");
  const int recordable = !strcmp (std_get (ar->rq->cgi, "recordable"), "yes");

  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;
  
  if (strlen (fullname) < 5)
    return validate_error (_("Full name too short."));
  if (strlen (email) < 3)
    return validate_error (_("Email address too short."));

  if (!strcmp (email, "local"))
    email = "";
  else if (strchr (email, '@') == NULL)
    validate_warning (_("Email address should normally contain `@'."));

  db_query ("UPDATE users "
	    "SET name = '%sq', email = '%sq', recordable = '%sq' "
	    "WHERE usr = '%sq'",
	    DB_FORMAT, NULL,
	    fullname, email, recordable ? "t" : "f", ar->usr.s[0]);
  
  success (ar, 0);

  html ("<DT>%s:<DD>%s\n"
	"<DT>%s:<DD>%s\n"
	"<DT>%s<DD>%s\n",
	_("Full name"), fullname,
	_("Email"), email,
	_("Email Recordable?"), recordable ? "yes" : "no");

  update ();
  return 1;
}

static void
interact_editusr (struct admin_request *ar)
{
  static const char *input[][5] = 
    {
      {"fullname", N_("Full Name"), "text SIZE=40", NULL,
       N_("<BR>User's full name, first followed by last.")},

      {"email", N_("Upstream Email"), "text SIZE=40", NULL,
       N_("<BR>User's upstream email address.\n"
	  "If user will receive mail directly from account on this server,\n"
	  "enter <STRONG>local</STRONG>.\n"
	  "This is the email address that all TeamSCOPE correspondence will\n"
	  "be sent to.\n")},

      {"recordable", N_("Email Recordable?"), "checkbox", NULL,
       N_("May user email passing through the server be recorded?\n"
	  "<BR>Email recording is used for research purposes.\n"
	  "You may opt out of email recording using this option.\n"
	  "This option has an effect only if your local TeamSCOPE\n"
	  "installation is set up to record email.\n")},
    };

  struct user_info info;

  assert (ar->usr.n == 1);
  get_user_info (&ar->usr, &info, UI_ALL);
  input[0][3] = info.fullname;
  input[1][3] = *info.email ? info.email : "local";
  input[2][3] = info.recordable ? "t" : "";
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Edit User Profile"),
	_("To edit this user profile,\n"
	  "fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  htmls ("<TABLE>\n\n");

  html ("  <TR><TD>%s</TD><TD>%s</TD></TR>\n", _("Username"), ar->usr.s[0]);

  present_form (ar, input, sizeof input / sizeof *input);

  free_user_info (&info, 1);
}

static int
execute_lockusr (struct admin_request *ar)
{
  size_t i;

  for (i = 0; i < ar->usr.n; i++)
    shell (NULL, "scope-db lock '%su'", ar->usr.s[i]);

  success (ar, 1);
  return 1;
}

static int
execute_unlockusr (struct admin_request *ar)
{
  size_t i;

  for (i = 0; i < ar->usr.n; i++)
    shell (NULL, "scope-db unlock '%su'", ar->usr.s[i]);

  success (ar, 1);
  return 1;
}

static int
execute_designateadmin (struct admin_request *ar)
{
  size_t i, j;

  db_begin ();
  
  for (i = 0; i < ar->usr.n; i++)
    for (j = 0; j < ar->grp.n; j++)
      db_query ("UPDATE user_group SET may_admin = 't' "
		"WHERE usr = '%sq' AND grp = '%sq'",
		DB_FORMAT, NULL,
		ar->usr.s[i], ar->grp.s[j]);

  db_end ();

  success (ar, 1);
  return 1;
}

static int
execute_revokeadmin (struct admin_request *ar)
{
  size_t i;

  for (i = 0; i < ar->usr.n; i++)
    db_query ("UPDATE user_group SET may_admin = 'f' WHERE usr = '%sq'",
	      DB_FORMAT, NULL, ar->usr.s[i]);

  success (ar, 1);
  return 1;
}

static int
execute_notifyusr (struct admin_request *ar)
{
  const char *reset;
  size_t i;

  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;

  if (!strcmp (std_get (ar->rq->cgi, "reset"), "yes"))
    reset = "";
  else
    reset = "--no-reset";

  for (i = 0; i < ar->usr.n; i++)
    shell (NULL, "scope-db mail %ss '%su'",
	   reset, ar->usr.s[i]);

  success (ar, 1);
  return 1;
}

static void
interact_notifyusr (struct admin_request *ar)
{
  static const char *input[][5] = 
    {
      {"reset", N_("Reset Passwords?"), "checkbox", "",
       N_("Reset selected users' passwords?\n"
	  "By default, the notified users are given new, random\n"
	  "passwords only if they don't yet have passwords.\n"
	  "Check this box to reset them even if they do have\n"
	  "passwords already.  This is particularly useful if\n"
	  "the user in question has forgotten his or her password.\n")},
    };

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Notify User by Email / Reset Password"),
	_("To notify the selected user(s) and/or reset their\n"
	  "passwords, fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  htmls ("<TABLE>\n\n");

  html ("  <TR><TD>%s</TD><TD>", _("User(s)"));
  id_list_print (&ar->usr);
  htmls ("</TD></TR>\n\n");

  present_form (ar, input, sizeof input / sizeof *input);
}

static int
execute_setpasswd (struct admin_request *ar)
{
  const char *passwd1 = std_get (ar->rq->cgi, "passwd1");
  const char *passwd2 = std_get (ar->rq->cgi, "passwd2");
  
  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;

  if (strcmp (passwd1, passwd2))
    return validate_error (_("Passwords differ, please try again."));
  if (strlen (passwd1) == strspn (passwd1, " \t\r\n\v"))
    return validate_error (_("Password may not be empty."));
  if (strlen (passwd1) < 6)
    return validate_error (_("Password too short."));
  
  auth_set_web_password (ar->usr.s[0], passwd1);
  auth_set_unix_password (ar->usr.s[0], passwd1);

  success (ar, 1);
  return 1;
}

static void
interact_setpasswd (struct admin_request *ar)
{
  static const char *input[][5] = 
    {
      {"passwd1", N_("Password 1"), "password SIZE=8", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "New password.  Must be at least 6 characters.")},

      {"passwd2", N_("Password 2"), "password SIZE=8", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "New password, again, for confirmation.")},
    };

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Set password"),
	_("To set the user's password,\n"
	  "fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  html ("<TABLE>\n\n"
	"  <TR><TD>%s</TD><TD>%s</TD></TR>\n\n",
	_("User"), ar->usr.s[0]);

  present_form (ar, input, sizeof input / sizeof *input);
}

static void
interact_viewgrp (struct admin_request *ar)
{
  struct id_list usrs;

  assert (ar->grp.n == 1);
  html ("<H2 ALIGN=CENTER>%s: %s</H2>\n\n", _("Group"), ar->grp.s[0]);

  id_list_for_group (&usrs, ar->grp.s[0]);
  show_users (ar, &usrs, A_ZERO, 0, NULL, -1);
  id_list_destroy (&usrs);
}

static int
execute_creategrp (struct admin_request *ar)
{
  const char *groupname = std_get (ar->rq->cgi, "groupname");
  const char *fullname = std_get (ar->rq->cgi, "fullname");

  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;
  
  if (strlen (fullname) < 3)
    return validate_error (_("Full name too short."));
  if (strlen (groupname) < 3)
    return validate_error (_("Group name too short."));
  if (strlen (groupname) > 8)
    return validate_error (_("Group name too long."));
  if (!validate_name (groupname))
    return validate_error (_("Improperly constructed group name."));
  if (getgrnam (groupname) != NULL || getpwnam (groupname))
    return validate_error (_("User or group by this name already exists."));

  shell (NULL, "scope-db addgroup '%su' '%sq'", groupname, fullname);

  success (ar, 0);

  html ("<DT>%s:<DD>%s\n"
	"<DT>%s:<DD>%s\n",
	_("Group name"), groupname,
	_("Full Name"), fullname);

  update ();
  return 1;
}

static void
interact_creategrp (struct admin_request *ar)
{
  static const char *input[][5] = 
    {
      {"groupname", N_("Group Name"), "text SIZE=8", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "Group name.\n"
	  "Must be between 3 and 8 lowercase letters and digits\n"
	  "and must begin with a letter.\n")},

      {"fullname", N_("Full Name"), "text SIZE=40", "",
       N_("<BR><STRONG>(Required)</STRONG>\n"
	  "Full name of group.\n"
	  "May contain most common characters.")},
    };

  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Create Group"),
	_("To create a new group, "
	  "fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  htmls ("<TABLE>\n\n");

  html ("  <TR><TD>%s</TD><TD>", _("User(s)"));
  id_list_print (&ar->usr);
  htmls ("</TD></TR>\n\n");

  present_form (ar, input, sizeof input / sizeof *input);
}
		    
static int
execute_delgrp (struct admin_request *ar)
{
  size_t i;

  for (i = 0; i < ar->grp.n; i++)
    shell (NULL, "scope-db delgroup '%su'", ar->grp.s[i]);

  success (ar, 1);
  update ();
  return 1;
}

static int
execute_editgrp (struct admin_request *ar)
{
  const char *fullname = std_get (ar->rq->cgi, "fullname");

  if (!std_contains (ar->rq->cgi, "execute"))
    return 0;
  
  if (strlen (fullname) < 3)
    return validate_error (_("Full name too short."));

  db_query ("UPDATE groups SET name = '%sq' WHERE grp = '%sq'",
	    DB_FORMAT, NULL,
	    fullname, ar->grp.s[0]);
  
  success (ar, 0);

  html ("<DT>%s<DD>%s\n", _("Full name"), fullname);
  update ();
  return 1;
}

static void
interact_editgrp (struct admin_request *ar)
{
  static const char *input[][5] = 
    {
      {"fullname", N_("Full Name"), "text SIZE=40", NULL,
       N_("<BR>Group's full name.")},
    };

  db_result *r;

  assert (ar->grp.n == 1);
  db_query ("SELECT name FROM groups WHERE grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, ar->grp.s[0]);
  input[0][3] = dbr_value (r, 0, 0);
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n\n"
	"<P>%s</P>\n",
	_("Edit Group Profile"),
	_("To edit this group profile,\n"
	  "fill out the form below, then click on the\n"
	  "<STRONG>Submit</STRONG> button.\n"));

  html_start_form ("admin", HT_POST, HT_URL_ENCODED);

  htmls ("<TABLE>\n\n");

  html ("  <TR><TD>%s</TD><TD>%s</TD></TR>\n", _("Group Name"), ar->grp.s[0]);

  present_form (ar, input, sizeof input / sizeof *input);

  dbr_destroy (r);
}

static int
execute_addusrtogrp (struct admin_request *ar)
{
  size_t i, j;
  
  for (i = 0; i < ar->usr.n; i++)
    for (j = 0; j < ar->grp.n; j++)
      shell (NULL, "scope-db addusertogroup '%su' '%su'",
	     ar->usr.s[i], ar->grp.s[j]);

  success (ar, 1);
  update ();
  return 1;
}

static int
execute_rmusrfromgrp (struct admin_request *ar)
{
  size_t i, j;
  
  for (i = 0; i < ar->usr.n; i++)
    for (j = 0; j < ar->grp.n; j++)
      shell (NULL, "scope-db deluserfromgroup '%su' '%su'",
	     ar->usr.s[i], ar->grp.s[j]);

  success (ar, 1);
  update ();
  return 1;
}

/* Reads F into string S.
   If D is non-null, parses lines read from F
   to read any variables declared in form `##key: value'
   and puts them into D. */
static void
read_output (FILE *f, struct st *s, struct std *d)
{
  size_t line_beg = 0;

  st_init (s, 256);
  for (;;)
    {
      int c = getc (f);
      if (c == EOF)
	break;

      st_putchar (s, c);

      if (c == '\n' && d != NULL)
	{
	  const char *line;
	  const char *bp, *ep;
	  char *key, *value;

	  /* Find ## leader. */
	  line = st_value (s) + line_beg;
	  line_beg = st_length (s);
	  bp = strstr (line, "##");
	  if (bp == NULL)
	    continue;

	  /* Extract key. */
	  bp += 2;
	  if (*bp == ' ')
	    bp++;
	  ep = strchr (bp, ':');
	  if (ep == NULL)
	    continue;
	  key = xstrndup (bp, ep - bp);

	  /* Extract value. */
	  bp = ep + 1;
	  if (*bp == ' ')
	    bp++;
	  ep = strchr (bp, '\0');
	  if (ep > bp && ep[-1] == '\n')
	    ep--;
	  value = xstrndup (bp, ep - bp);

	  /* Finish up. */
	  std_put (d, key, value);

	  free (value);
	  free (key);
	}
    }
}

/* Runs the shell command specified by FORMAT,
   which can contain the following escapes:
	%sq: Quoted strings limited to "safe" characters.
	%su: Usernames and groupnames.
	%ss: (discouraged) Arbitrary strings.
   If DICT is non-null then the command's output is parsed
   for variable declarations, which are stored into DICT. */
static void 
shell (struct std *dict, const char *format, ...)
{
  struct st command;
  
  {
    const char *p;
    va_list args;

    va_start (args, format);

    st_init (&command, 64);
    p = format;
    while (*p)
      {
	if (*p != '%')
	  {
	    st_putchar (&command, *p++);
	    continue;
	  }

	p++;
	if (*p != 's')
	  abort ();

	switch (*++p)
	  {
	  case 'q':
	    {
	      static const char okay[] =
		"abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		" 0123456789@%&()[]<>/.-_";
	      
	      const char *s = va_arg (args, const char *);
	      for (; *s; s++)
		if (strchr (okay, *s))
		  st_putchar (&command, *s);
	    }
	    break;

	  case 'u':
	    {
	      const char *s = va_arg (args, const char *);
	      const char *end;

	      end = strchr (s, '\0');
	      if (end > s + 8)
		end = s;
	    
	      for (; s < end; s++)
		if (isalnum ((unsigned char) *s) || *s == '-' || *s == '_')
		  st_putchar (&command, *s);
	    }
	    break;

	  case 's':
	    {
	      const char *s = va_arg (args, const char *);
	      st_concat (&command, s);
	    }
	    break;

	  default:
	    abort ();
	  }

	p++;
      }

    va_end (args);
  }

  if (st_length (&command) > 1024)
    fail ("command too long: \"%s\"", st_value (&command));

#if DEBUGGING
  log ("executing command: %s", st_value (&command));
#endif
  
  {
    FILE *proc = popen (st_value (&command), "r");
    struct st output;
    
    if (proc == NULL)
      {
	html ("<H2>%s</H2>\n"
	      "<P>%s: %s.  %s</P>\n",
	      _("Error executing command"),
	      _("An error occurred while executing your command"),
	      strerror (errno),
	      _("More information may be available in the webserver log."));
	log ("command failed: \"%s\"", st_value (&command));
	exit (EXIT_FAILURE);
      }

    read_output (proc, &output, dict);

    if (WEXITSTATUS (pclose (proc)) != 0)
      {
	html ("<H2>%s</H2>\n"
	      "<P>%s</P>\n",
	      _("Error executing command"),
	      _("An error occurred while executing your command.\n"
		"More information may be available in the webserver log."));
	log ("command failed: \"%s\" with output: \"%s\"",
	     st_value (&command), st_value (&output));
	exit (EXIT_FAILURE);
      }

    st_destroy (&output);
  }

  st_destroy (&command);
}

#if GLOBAL_DEBUGGING
/* Ensures that interfaces[] correspond properly to commands. */
static void 
validate_interfaces (void)
{
  int j;

  for (j = 0; j < L_CNT; j++)
    {
      const struct admin_interface *i;
      
      for (i = interface[j]; i < interface[j] + interface_cnt[j]; i++)
	{
	  assert (i->command != NULL && i->title != NULL);

	  if (i->command == &c_separator)
	    continue;
      
	  if (((i->command->grp == A_ZERO) ^ (i->grp == G_NONE))
	      || (i->command->grp == A_ONE && i->grp == G_GRPS))
	    fail ("interface '%s' mismatches command '%s' in group selection",
		  i->title, i->command->name);

	  if (((i->command->usr == A_ZERO) ^ (i->usr == U_NONE))
	      || (i->command->usr == A_ONE
		  && (i->usr == U_USRS
		      || i->usr == U_ONLY_MY_USRS
		      || i->usr == U_MY_USRS)))
	    fail ("interface '%s' mismatches command '%s' in user selection",
		  i->title, i->command->name);

	  if ((i->warning_prompt != NULL) ^ (i->confirm_prompt != NULL))
	    fail ("confirm/warning inconsistent for interface '%s'", i->title);

	  if ((i->grp == G_GRP || i->grp == G_GRPS) ^ (i->grp_prompt != NULL))
	    fail ("missing/spurious group prompt for interface '%s'", i->title);

	  if ((i->usr == U_USR || i->usr == U_USRS || i->usr == U_ONLY_MY_USRS
	       || i->usr == U_MY_USR || i->usr == U_MY_USRS)
	      ^ (i->usr_prompt != NULL))
	    fail ("missing/spurious user prompt for interface '%s'", i->title);
	}
    }
}
#endif /* GLOBAL_DEBUGGING */
