/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <crypt.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include "alloc.h"
#include "auth.h"
#include "cgi.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "scope.cgi.h"
#include "session.h"
#include "str.h"
#include "std.h"

/* Public functions. */

static void set_user (struct request *, const char *);
static void set_workgroup (struct request *, const char *);
static int user_belongs_to (const char *usr, const char *grp);
static int lock_file (const char *lockfile);
static void unlock_file (const char *lockfile, int fd);
static void rand_salt (char *salt);

/* Authorize the user; i.e., through some form of authentication,
   determine who the user on the other end of the connection really
   is.  After this has been done once, it can persist for the rest of
   the session using the session-data saving features in session.c.

   Also authorizes the user's workgroup; that is, determines which
   workgroup the user works in for this session.  The user can choose
   a workgroup at the beginning of the session.

   Sets `user' and `workgroup' properly if authorization is
   successful; sets them both to a null pointer if not. */
void
auth_initialize (struct request *rq)
{
  strncpy (rq->cx.usr, std_get (rq->trusted, "user"), 8);
  rq->cx.usr[8] = '\0';

  strncpy (rq->cx.grp, std_get (rq->trusted, "workgroup"), 8);
  rq->cx.grp[8] = '\0';
  
  rq->cx.fullgrp = xstrdup (std_get (rq->trusted, "teamname"));
  
  /* Two ways to authorize users: HTTP login/password authentication
     and CGI POST login/password authentication.  Either kind is
     persistent within a session, saved with the session data.

     HTTP login/password authentication cannot override other
     types of authentication.  This is because many web user agents
     don't provide their users a way to turn off such authentication, so
     users could become confused if the two methods were combined. */
  {
    const char *cgi_username = std_get_null (rq->cgi, "username");
    const char *cgi_password = std_get_null (rq->cgi, "password");
    
    if (cgi_username != NULL && cgi_password != NULL)
      {
	if (auth_check_password (cgi_username, cgi_password))
	  set_user (rq, cgi_username);
	else
	  {
	    set_user (rq, NULL);
	    set_workgroup (rq, NULL);
	    free (rq->screen);
	    rq->screen = xstrdup ("authfailed");
	    return;
	  }
      }
  }

  if (!rq->cx.usr[0] && getenv ("REMOTE_USER"))
    set_user (rq, getenv ("REMOTE_USER"));

  if (!rq->cx.usr[0])
    {
      set_workgroup (rq, NULL);
      return;
    }

  /* At this point we have authorized the user.
     Now we must select one of the user's workgroups. */
  db_begin ();

  /* This is a kluge to handle links from emails properly by requiring
     the proper associated current workgroup. */
  if (rq->cx.grp[0] == '\0'
      && std_contains (rq->cgi, "needgrp")
      && user_belongs_to (rq->cx.usr, std_get (rq->cgi, "needgrp")))
    set_workgroup (rq, std_get (rq->cgi, "needgrp"));
  
  /* Users with more than one workgroup may change workgroups by
     selecting one from the "Main Menu..." drop-down on every
     TeamSCOPE page.
     We do this regardless of whether the user already has a
     workgroup, of course. */
  if (rq->screen != NULL && !strncmp (rq->screen, "grp-", 4))
    {
      char *cgi_workgroup = xstrdup (rq->screen + 4);
      char *p = strchr (cgi_workgroup, '-');

      if (p != NULL)
	{
	  *p = '\0';
	  free (rq->screen);
	  rq->screen = xstrdup (p + 1);
	}
      else
	rq->screen = xstrdup ("");

      if (user_belongs_to (rq->cx.usr, cgi_workgroup))
	set_workgroup (rq, cgi_workgroup);
    }

  /* If the user doesn't have a workgroup, see if authorization for a
     particular group is needed, and if so then select one of those
     groups if the user belongs to one of them. */
  if (rq->cx.grp[0] == '\0')
    {
      char *groups = local_strdup (std_get (rq->cgi, "groups"));
      if (strtok (groups, " \t\r\n\v") != NULL)
	for (;;)
	  {
	    char *token = strtok (NULL, " \t\r\n\v");
	    if (token == NULL)
	      break;

	    if (user_belongs_to (rq->cx.usr, token))
	      {
		set_workgroup (rq, token);
		break;
	      }
	  }
      local_free (groups);
    }

  /* If the user still doesn't have a workgroup, select the user's
     primary group. */
  if (rq->cx.grp[0] == '\0')
    {
      db_result *r;
      const char *primarygrp;

      db_query ("SELECT primarygrp FROM users WHERE usr = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

      primarygrp = dbr_value (r, 0, 0);
      if (primarygrp[0] != '\0' && user_belongs_to (rq->cx.usr, primarygrp))
	set_workgroup (rq, primarygrp);

      dbr_destroy (r);
    }

  /* If the user still doesn't have a workgroup, select the first
     group in alphabetical order. */
  if (rq->cx.grp[0] == '\0')
    {
      db_result *r;

      db_query ("SELECT grp FROM user_group WHERE usr = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

      if (dbr_nrows (r) > 0)
	set_workgroup (rq, dbr_value (r, 0, 0));

      dbr_destroy (r);
    }

  /* Set idle time from database.  This can change within a session,
     so we have to redo this with each page load. */
  {
    db_result *r;

    db_query ("SELECT idle FROM users WHERE usr = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

    if (stoi (dbr_value (r, 0, 0))
	!= stoi (std_get (rq->trusted, "idle")))
      std_replace (rq->trusted, "idle", dbr_value (r, 0, 0));

    dbr_destroy (r);
  }

  db_end ();
}

/* Change user's password if everything's good. */
void
auth_password_change (struct request *rq)
{
  const char *cur = std_get_null (rq->cgi, "curpasswd");
  const char *pw1 = std_get (rq->cgi, "newpasswd1");
  const char *pw2 = std_get (rq->cgi, "newpasswd2");
  
  http_header ();
  html_banner (rq);

  if (cur == NULL)
    return;
  if (!auth_check_password (rq->cx.usr, cur))
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P>%s</P>\n",
	    _("Password Change Failed"),
	    _("Your password was not changed because the current password\n"
	      "supplied was not correct.  Please try again."));
      return;
    }

  if (strcmp (pw1, pw2))
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P>%s</P>\n",
	    _("Password Change Failed"),
	    _("Your password was not changed because the two copies of\n"
	      "the new password did not match.  Please try again."));
      return;
    }

  if (strspn (pw1, " \t\r\n\v") == strlen (pw1)
      || blp_strcasestr (pw1, rq->cx.usr) != NULL
      || strlen (pw1) < 6)
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P>%s</P>\n",
	    _("Password Change Failed"),
	    _("Your new password was rejected because it was determined\n"
	      "to be too weak.\n"
	      "You must use a password at least 6 characters long not\n"
	      "based on your username.\n"
	      "Select a new password and try again."));
      return;
    }

  auth_set_web_password (rq->cx.usr, pw1);
  auth_set_unix_password (rq->cx.usr, pw1);
      
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s</P>\n",
	_("Password Change Successful"),
	_("Your password has been changed."));

  html_trailer (rq);
  exit (EXIT_SUCCESS);
}

/* Password changing screen. */
void
scr_password (struct request *rq)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s</P>\n",

	_("Change Password"),
	_("To change your TeamSCOPE password, enter your current password\n"
	  "below, then the new password twice for verification.\n"));

  html_start_form ("password", HT_POST, HT_URL_ENCODED);
  htmls ("<TABLE ALIGN=CENTER CELLSPACING=0>\n");
  htmls ("<TR><TD ALIGN=RIGHT>\n");
  html_label ("curpasswd", _("Current Password:"));
  htmls ("</TD><TD>\n");
  html_password_field ("curpasswd", 16);
  htmls ("</TD></TR>\n"
	 "<TR><TD ALIGN=RIGHT>\n");
  html_label ("newpasswd1", _("New Password:"));
  htmls ("</TD><TD>\n");
  html_password_field ("newpasswd1", 16);
  htmls ("</TD></TR>\n"
	 "<TR><TD ALIGN=RIGHT>\n");
  html_label ("newpasswd2", _("New Password (again):"));
  htmls ("</TD><TD>\n");
  html_password_field ("newpasswd2", 16);
  htmls ("</TD></TR>\n"
	 "<TR><TD COLSPAN=2 ALIGN=CENTER>\n");
  html_submit (_("Change Password"));
  htmls ("</TD></TR>\n"
	 "</TABLE>\n");
  html_end_form ();

  htmls ("</CENTER>\n");

  html_trailer (rq);
}

/* Library functions. */

/* Try to authorize USER against the password file given password
   PASSWORD.  Return 1 if authentication is accepted, 0 otherwise. */
int
auth_check_password (const char *user, const char *password)
{
  const char *filename = PASSWD_FILE;
  char *line = NULL;
  size_t line_size = 0;

  FILE *f = fopen (filename, "r");

  int good = 0;

  /* Make sure we opened the password file okay. */
  if (f == NULL)
    {
      log ("opening %s: %s", filename, strerror (errno));
      goto exit;
    }
  verify_file_owner (filename, f, 0);

  /* Process each line in the password file. */
  for (;;)
    {
      char *file_passwd;

      /* Read a line. */
      long line_len = getline (&line, &line_size, f);
      if (-1 == line_len)
	{
	  if (ferror (f))
	    log ("reading %s: %s", filename, strerror (errno));
	  else
	    log ("%s: user %s not found", filename, user);

	  goto exit;
	}

      /* Check that the line is in the proper format.  If the username
	 doesn't match that of the user we're interested in, skip back
	 to the beginning of the loop and try the next line. */
      {
	char *colon = strchr (line, ':');
	if (colon == NULL)
	  {
	    static int warned = 0;

	    if (!warned)
	      {
		log ("%s: line has wrong format", filename);
		warned = 1;
	      }
	    
	    continue;
	  }

	*colon = '\0';
	if (strcmp (line, user))
	  continue;

	file_passwd = colon + 1;
      }

      /* Drop the trailing newline so that the line can be compared
	 properly. */
      assert (line[line_len - 1] == '\n');
      line[line_len - 1] = '\0';
      
      /* Encrypt the user-supplied password against the salt given in
	 the password file, then compare this encrypted password to
	 the encrypted password in the password file.  If there's a
	 match then accept authentication, otherwise reject. */
      {
	char *user_passwd = crypt (password, file_passwd);
	if (user_passwd && !strcmp (file_passwd, user_passwd))
	  {
	    good = 1;
	    log ("accepted authorization of %s", user);
	  }
	else
	  log ("rejected authorization of %s", user);
      }

      break;
    }

 exit:
  free (line);
  fclose (f);
  return good;
}

/* Set USR's TeamSCOPE password to PASSWORD.  */
void
auth_set_web_password (const char *usr, const char *password)
{
  char salt[13];
  char *crypted;

  char *line;
  size_t line_size;

  FILE *src, *dst;
  int lock_fd;

  rand_salt (salt);
  crypted = crypt (password, salt);
  assert (crypted != NULL);
  
  src = NULL;
  dst = NULL;

  lock_fd = lock_file (PASSWD_FILE ".lock");
  
  src = fopen (PASSWD_FILE, "r");
  if (src == NULL)
    fail ("can't open " PASSWD_FILE " for reading: %s", strerror (errno));
  verify_file_owner (PASSWD_FILE, src, 0);

  dst = fopen (PASSWD_FILE ".new", "w");
  if (dst == NULL)
    fail ("can't create " PASSWD_FILE ".new: %s", strerror (errno));
  
  line = NULL;
  line_size = 0;
  for (;;)
    {
      ssize_t line_len = getline (&line, &line_size, src);
      if (-1 == line_len)
	{
	  if (!ferror (src))
	    break;
	  else
	    fail ("error reading " PASSWD_FILE ": %s", strerror (errno));
	}

      if (strncmp (line, usr, strlen (usr)) || line[strlen (usr)] != ':')
	if (EOF == fputs (line, dst))
	  fail ("error writing " PASSWD_FILE ".new: %s", strerror (errno));
    }

  if (fprintf (dst, "%s:%s\n", usr, crypted) < 0)
    fail ("error writing " PASSWD_FILE ".new: %s", strerror (errno));

  if (fclose (dst) == EOF)
    fail ("error closing " PASSWD_FILE ".new: %s", strerror (errno));
  dst = NULL;

  if (fclose (src) == EOF)
    fail ("error closing " PASSWD_FILE ": %s", strerror (errno));
  src = NULL;

  unlink (PASSWD_FILE ".old");
  if (link (PASSWD_FILE, PASSWD_FILE ".old") == -1
      || rename (PASSWD_FILE ".new", PASSWD_FILE) == -1)

  unlock_file (PASSWD_FILE ".lock", lock_fd);
}

/* Set USR's UNIX password to PASSWORD. */
void
auth_set_unix_password (const char *usr, const char *password)
{
  /* Yes, this code is redundant.  We obtain a new salt because that
     way it's not obvious that the user has the same password on
     TeamSCOPE as on their Unix account. */
  char salt[13];
  char *crypted;

  FILE *chpasswd;
  int status;

  rand_salt (salt);
  crypted = crypt (password, salt);
  assert (crypted != NULL);
  
  chpasswd = popen ("chpasswd -e", "w");
  if (chpasswd == NULL)
    fail ("error running chpasswd: %s", strerror (errno));
  fprintf (chpasswd, "%s:%s\n", usr, crypted);

  status = pclose (chpasswd);
  if (!WIFEXITED (status) || WEXITSTATUS (status) != 0)
    fail ("chpasswd returned status %04x", status);
}

/* Private functions. */

/* Returns nonzero only if USR belongs to GRP. */
static int
user_belongs_to (const char *usr, const char *grp)
{
  db_result *r;
  int belongs;
  
  db_query ("SELECT * FROM user_group "
	    "WHERE usr = '%sq' AND grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, usr, grp);
  belongs = dbr_nrows (r) > 0;
  dbr_destroy (r);

  return belongs;
}

/* Set the current user to NEW_USER. */
static void
set_user (struct request *rq, const char *new_usr)
{
  if (new_usr != NULL)
    {
      strncpy (rq->cx.usr, new_usr, 8);
      rq->cx.usr[8] = '\0';
    }
  else
    rq->cx.usr[0] = '\0';

  rq->cx.grp[0] = '\0';
}

/* Set the current workgroup to NEW_WORKGROUP. */
static void
set_workgroup (struct request *rq, const char *new_grp)
{
  free (rq->cx.fullgrp);
  
  if (new_grp != NULL && new_grp[0] != '\0')
    {
      db_result *r;

      strncpy (rq->cx.grp, new_grp, 8);
      rq->cx.grp[8] = '\0';

      db_query ("SELECT name FROM groups WHERE grp = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, rq->cx.grp);
      rq->cx.fullgrp = xstrdup (dbr_value (r, 0, 0));
      dbr_destroy (r);
    }
  else
    {
      rq->cx.grp[0] = '\0';
      rq->cx.fullgrp = NULL;
    }
}

/* Create a lockfile with a safe algorithm. */
static int
lock_file (const char *lockfile)
{
  int fd = open (lockfile, O_WRONLY | O_CREAT, 0600);
  if (fd == -1)
    fail ("can't create lockfile %s: %s", lockfile, strerror (errno));

  if (!lock (F_WRLCK, fd))
    fail ("can't lock %s: %s", lockfile, strerror (errno));

  return fd;
}

/* Delete a lockfile.
   FIXME? */
static void
unlock_file (const char *lockfile, int fd)
{
  if (-1 == unlink (lockfile))
    {
      if (errno != ENOENT)
	fail ("error deleting lockfile %s: %s", lockfile, strerror (errno));
      else
	log ("error deleting lockfile %s: %s", lockfile, strerror (errno));
    }

  close (fd);  
}

/* Generate a random MD5 salt into SALT. */
static void
rand_salt (char salt[13])
{
  unsigned char bits[8];
  
  /* Read 8 bytes of random noise into `bits'. */
  {
    int fd = open ("/dev/urandom", O_RDONLY);
    if (fd == -1)
      fail ("opening /dev/urandom: %s", strerror (errno));

    if (read (fd, bits, 8) != 8)
      fail ("reading /dev/urandom: %s", strerror (errno));

    if (close (fd) == -1)
      fail ("closing /dev/urandom: %s", strerror (errno));
  }

  /* Convert it to desired format. */
  {
    static const char salts[64]
      = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";
    int i;

    strcpy (salt, "$1$........$");
    for (i = 0; i < 8; i++)
      salt[i + 3] = salts[bits[i] & 63];
  }
}
