/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <gd.h>
#include <gdfonts.h>
#include "activity.h"
#include "alloc.h"
#include "calendar.h"
#include "cgi.h"
#include "db.h"
#include "datetime.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "messages.h"
#include "resource.h"
#include "scope.cgi.h"
#include "str.h"
#include "std.h"

/* Represents an event. */
struct event
  {
    /* Event begin time in various formats. */
    struct tm b_tm;
    time_t b_time;
    const char *b_string;	/* As retrieved from database record. */

    /* Event end time in various formats. */
    struct tm e_tm;
    time_t e_time;
    const char *e_string;	/* As retrieve from database record. */

    const char *title;		/* Event title. */
    const char *note;		/* Notes. */

    const char *owner;		/* Not necessarily the person who posted it. */
    const char *grp;		/* Owning workgroup. */

    int *res;			/* Resources. */
    int n_res;			/* Number of resources. */

    int seq;			/* Sequence number. */
  };

static void show_calendar (struct request *);
static void show_goto (struct request *);
static void show_import_msp (struct request *);
static void add_calendar_event (struct request *);
static void show_calendar_event (struct request *, int seq);
static void new_calendar_event (struct request *, const struct event *);
static void find_event_days (struct context *,
			     int year, int month, short events[31]);

static void show_event (struct request *, const struct event *e);
static void check_conflicts (struct context *, const struct event *e);
static void determine_interval (struct request *,
				struct tm *beg, struct tm *end);
static db_result *calendar_query (struct context *,
				  const struct tm *begin,
				  const struct tm *end);
static void print_calendar (struct request *,
			    struct tm *cur, const struct tm *begin,
			    const struct tm *end);
static void catchup_title (const struct tm *date, struct tm *last);

static db_result *retrieve_event (struct context *, int seq, struct event *);
static void interpret_event (struct event *e, db_result *r1, int row);
static int list_contains_value (int *list, int n_list, int value);

static void center_text (gdImagePtr, int line, int x, int color, const char *);

static time_t tz_offset (void);
static time_t round_date_down (time_t);
static time_t time_of_day (time_t);
static int tm_ge_day (const struct tm *, const struct tm *);
static int tm_ge_month (const struct tm *, const struct tm *);
static struct tm *prev_day (const struct tm *);
static void inc_month (struct tm *);
static void dec_month (struct tm *);

/* Public functions. */

/* Timezone offset from UTC, in seconds. */
time_t tzoff;

/* Handles the main calendar screen. */
void
scr_calendar (struct request *rq)
{
  tzoff = tz_offset ();

  if (!strcmp (rq->op, "goto"))
    show_goto (rq);
  else if (!strcmp (rq->op, "msp"))
    show_import_msp (rq);
  else if (!strcmp (rq->op, "new"))
    new_calendar_event (rq, NULL);
  else if (!strcmp (rq->op, "add"))
    add_calendar_event (rq);
  else if (std_contains (rq->cgi, "seq"))
    show_calendar_event (rq, stoi (std_get (rq->cgi, "seq")));
  else
    show_calendar (rq);
}

/* Handles calendargif URLs, which produce a graphical illustration of
   a traditional-looking wall calendar. */
void
scr_calendargif (struct request *rq)
{
  /* Calendar size metrics. */
  #define C_DAY_WIDTH 20		/* Width of a day rectangle. */
  #define C_DAY_HEIGHT C_DAY_WIDTH	/* Height of same. */

  #define C_MARGIN 2			/* Spacing between days. */

  #define C_WIDTH 7 * (C_DAY_WIDTH + C_MARGIN) - C_MARGIN

  #define C_TEXT_HEIGHT 12		/* Height of font. */
  #define C_TEXT_WIDTH 6		/* Width of font. */

  /* Height of the top caption ("January 1999"). */
  #define C_CAPTION_HEIGHT (C_TEXT_HEIGHT + 2 * C_MARGIN)

  /* Y-offset of the top row of day boxes. */
  #define C_FIRST_ROW_Y (2 * (C_CAPTION_HEIGHT + C_MARGIN))

  /* Height of a row of day boxes. */
  #define C_ROW_HEIGHT (C_DAY_HEIGHT + C_MARGIN)

  /* Information on the dates represented by this calendar. */
  int year, month;		/* Year and month of calendar to display. */
  int bday, eday;		/* First and last day of month to highlight. */
  int dow;			/* Day of week of the day in question. */
  int n_days;			/* Number of days in this month. */
  struct tm t;			/* First day of this month. */
  short events[31];		/* Nonzero values are days with events. */

  /* Information on the image itself. */
  gdImagePtr im;		/* The image. */
  int height;			/* Height of image. */
  int n_rows;			/* Number of rows of days. */
  int clear, white, gray, black, red;
				/* Color indexes. */

  /* Figure out what dates we're interested in. */
  tzoff = tz_offset ();

  year = stoi (std_get (rq->cgi, "year"));
  month = stoi (std_get (rq->cgi, "month"));
  bday = stoi (std_get (rq->cgi, "bday"));
  eday = stoi (std_get (rq->cgi, "eday"));

  if (year < 1900 || year > 2100)
    {
      log ("bad year %d", year);
      return;
    }
  year -= 1900;

  if (month < 1 || month > 12)
    {
      log ("bad month %d", month);
      return;
    }
  month--;

  n_days = datetime_month_days (year, month);
  if (bday < 1 || bday > n_days
      || eday < bday || eday > n_days)
    {
      log ("bad bday/eday %d/%d for year/month %d/%d",
	   bday, eday, year + 1900, month + 1);
      return;
    }

  find_event_days (&rq->cx, year, month, events);
  
  memset (&t, 0, sizeof t);
  t.tm_mday = 1;
  t.tm_mon = month;
  t.tm_year = year;
  mktime (&t);
  dow = t.tm_wday;

  /* Calculate lots of stuff about the image, and create the image. */
  n_rows = (dow + n_days - 1) / 7 + 1;

  height = (2 * (C_CAPTION_HEIGHT + C_MARGIN)
	    + n_rows * (C_DAY_HEIGHT + C_MARGIN)
	    - C_MARGIN);
  im = gdImageCreate (C_WIDTH, height);

  /* Allocate colors. */
  clear = gdImageColorAllocate (im, 0000, 0000, 0377);
  gdImageColorTransparent (im, clear);

  white = gdImageColorAllocate (im, 0377, 0377, 0377);
  gray = gdImageColorAllocate (im, 0300, 0300, 0300);
  black = gdImageColorAllocate (im, 0000, 0000, 0000);
  red = gdImageColorAllocate (im, 0377, 0000, 0000);

  /* Title. */
  {
    const char *title = datetime_format ("%B %Y", &t);
    
    gdImageFilledRectangle (im, 0, 0, C_WIDTH - 1, C_CAPTION_HEIGHT - 1,
			    black);
    gdImageString (im, gdFontSmall,
		   (C_WIDTH - C_TEXT_WIDTH * strlen (title)) / 2,
		   C_MARGIN, (char *) title, white);
  }

  /* < and > for next, previous month. */
  gdImageChar (im, gdFontSmall, C_MARGIN, C_MARGIN, '<', white);
  gdImageChar (im, gdFontSmall, C_WIDTH - C_TEXT_WIDTH - C_MARGIN, C_MARGIN,
	       '>', white);  

  /* Weekday names. */
  {
    int day;
    
    for (day = 0; day < 7; day++)
      {
	static const char weekdays[7][3]
	  = {N_("Su"), N_("Mo"), N_("Tu"), N_("We"),
	     N_("Th"), N_("Fr"), N_("Sa")};
	
	gdImageFilledRectangle (im,
				day * (C_DAY_WIDTH + C_MARGIN),
				C_CAPTION_HEIGHT + C_MARGIN,
				(day * (C_DAY_WIDTH + C_MARGIN)
				 + C_DAY_WIDTH - 1),
				C_MARGIN + 2 * C_CAPTION_HEIGHT - 1,
				black);
	gdImageString (im,
		       gdFontSmall,
		       ((C_DAY_WIDTH + C_MARGIN) * day
			+ C_DAY_WIDTH / 2 - C_TEXT_WIDTH),
		       (C_CAPTION_HEIGHT + C_MARGIN
			+ (C_CAPTION_HEIGHT - C_TEXT_HEIGHT) / 2),
		       (char *) (weekdays[day]),
		       white);
      }
  }

  /* All the individual days. */
  {
    int day;
    int y = C_FIRST_ROW_Y;

    for (day = 1; day <= n_days; day++) 
      {
	/* Draw the box around the day. */
	int x = dow * (C_DAY_WIDTH + C_MARGIN);
	int x2 = x + C_DAY_WIDTH - 1;
	int y2 = y + C_DAY_HEIGHT - 1;

	gdImageRectangle (im, x, y, x2, y2, black);
	gdImageRectangle (im, x + 1, y + 1, x2 - 1, y2 - 1, black);
	gdImageFilledRectangle (im, x + 2, y + 2, x2 - 2, y2 - 2,
				day >= bday && day <= eday ? gray : white);

	/* Draw a little circle in the center if there's an event on
           this day. */
	if (events[day - 1])
	  {
	    int cx = (x + x2) / 2 + 1;
	    int cy = (y + y2) / 2 + 1;

	    gdImageArc (im, cx, cy, 8, 8, 0, 360, red);
	    gdImageFill (im, cx, cy, red);
	  }

	/* The day number. */
	{
	  char day_string[INT_DIGITS + 1];

	  sprintf (day_string, "%d", day);
	  
	  gdImageString (im,
			 gdFontSmall,
			 x2 - (C_TEXT_WIDTH * strlen (day_string) + C_MARGIN),
			 y + C_MARGIN,
			 day_string,
			 black);
	}

	/* Increment day of week. */
	dow++;
	if (dow >= 7)
	  {
	    assert (dow == 7);
	    dow = 0;
	    y += C_ROW_HEIGHT;
	  }
      }
  }

  /* Output the gif image. */
  fputs ("Content-Type: image/gif\n\n", stdout);
  gdImageGif (im, stdout);
}

/* Draw a pseudo-Gantt chart representing the events between two
   particular dates. */
void
scr_ganttgif (struct request *rq)
{
  /* Image measurements. */
  #define G_CHAR_WIDTH 6		/* Font character width. */
  #define G_CHAR_HEIGHT 12		/* Font height. */
  #define G_MARGIN 4			/* Space between things. */

  #define G_ROW_HEIGHT (G_CHAR_HEIGHT + G_MARGIN)
					/* Height of a Gantt chart row. */
  #define G_WIDTH 600			/* Gantt chart width. */

  #define G_TITLE_CHARS 20		/* Length of displayed event titles. */
  #define G_TITLE_X (G_WIDTH - G_CHAR_WIDTH * G_TITLE_CHARS)
					/* X-location of title. */
  #define G_BAR_WIDTH (G_TITLE_X - G_MARGIN)
					/* Width of duration bar. */

  /* Timezone offset. */
  time_t tzoff;

  /* Beginning and ending dates in different formats. */
  struct tm begin, end;
  time_t b, e;

  /* Database query result representing the events that occur in the
     duration in question. */
  db_result *r;	
  int count;				/* Number of events. */

  /* Image info. */
  gdImagePtr im;			/* The image itself. */
  int height;				/* Image height. */
  int clear, black, outline, fill;	/* Color indexes. */

  /* Calculate timezone offset. */
  tzoff = tz_offset ();

  /* Determine date interval. */
  determine_interval (rq, &begin, &end);
  b = mktime (&begin);
  e = mktime (&end);

  /* Retrieve event list. */
  r = calendar_query (&rq->cx, &begin, &end);
  count = (int) dbr_nrows (r);

  /* Create and set up image. */
  height = G_ROW_HEIGHT * count + G_CHAR_HEIGHT * 2;
  im = gdImageCreate (G_WIDTH, height);

  clear = gdImageColorAllocate (im, 0000, 0000, 0377);
  gdImageColorTransparent (im, clear);

  black = gdImageColorAllocate (im, 0000, 0000, 0000);
  outline = gdImageColorAllocate (im, 0377, 0000, 0000);
  fill = gdImageColorAllocate (im, 0000, 0000, 0377);

  /* Annotate the time axis with tick marks and labels. */
  if (G_BAR_WIDTH / ((e - b) / (3600 * 4)) >= G_CHAR_WIDTH * 4 + G_MARGIN)
    {
      /* Tri-hourly markings with a tick mark for every hour. */
      time_t i = (b + tzoff + 3599) / 3600 * 3600 - tzoff;
      int h = (i + tzoff) / 3600 % 24;

      for (;;)
	{
	  int x = (double) (i - b) / (e - b) * G_BAR_WIDTH;
	  if (x > G_BAR_WIDTH)
	    break;
	  gdImageLine (im, x, G_CHAR_HEIGHT * 2, x, height - 1, black);
	  
	  if (h % 3 == 0)
	    {
	      char buf[INT_DIGITS + 16];

	      sprintf (buf, "%d%s", (h % 12 ? h % 12 : 12),
		       (h >= 12 ? "pm" : "am"));
	      center_text (im, 1, x, black, buf);
	    }
	  
	  i += 3600;
	  h++;
	}
    }
  else if (G_BAR_WIDTH / ((e - b) / (86400 * 3))
	   >= G_CHAR_WIDTH * 2 + G_MARGIN)
    {
      /* Tri-daily markings with a tick mark for every day. */
      time_t i = (b + tzoff + 86399) / 86400 * 86400 - tzoff;
      int omonth = -1;

      for (;;)
	{
	  int x = (double) (i - b) / (e - b) * G_BAR_WIDTH;
	  struct tm *tm = localtime (&i);

	  if (x > G_BAR_WIDTH)
	    break;
	  gdImageLine (im, x, G_CHAR_HEIGHT * 2, x, height - 1, black);

	  if ((tm->tm_mday - 1) % 3 == 0)
	    {
	      char buf[INT_DIGITS + 1];
	      sprintf (buf, "%d", tm->tm_mday);
	      center_text (im, 1, x, black, buf);
	    }
	  
	  if (tm->tm_mon != omonth)
	    {
	      center_text (im, 0, x, black, datetime_format ("%b %Y", tm));
	      omonth = tm->tm_mon;
	    }
	  i += 86400;
	}
    }
  else if (G_BAR_WIDTH / ((e - b) / (86400 * 30 * 2))
	   >= G_CHAR_WIDTH * 3 + G_MARGIN)
    {
      /* Bi-monthly markings with a tick mark for every month. */
      time_t i = (b + tzoff + 86399) / 86400 * 86400 - tzoff;
      struct tm *tm = localtime (&i);

      if (tm->tm_mday != 1)
	{
	  i += 86400 * (datetime_month_days (tm->tm_year, tm->tm_mon)
			- tm->tm_mday + 1);
	  inc_month (tm);
	}
      
      for (;;)
	{
	  int x = (double) (i - b) / (e - b) * G_BAR_WIDTH;
	  if (x > G_BAR_WIDTH)
	    break;
	  gdImageLine (im, x, G_CHAR_HEIGHT * 2, x, height - 1, black);
	  
	  if (tm->tm_mon % 2 == 0)
	    center_text (im, 1, x, black, datetime_format ("%b", tm));
	  if (tm->tm_mon == 0)
	    center_text (im, 0, x, black, datetime_format ("%Y", tm));

	  i += 86400 * datetime_month_days (tm->tm_year, tm->tm_mon);
	  inc_month (tm);
	}
    }

  /* Draw all of the events. */
  {
    double scale = 1.0 / (e - b) * G_BAR_WIDTH;
    int y = G_CHAR_HEIGHT * 2 + G_MARGIN;
    int row;

    for (row = 0; row < count; row++)
      {
	time_t s = datetime_parse (dbr_value (r, row, 0), 0, 0, NULL);
	time_t f = datetime_parse (dbr_value (r, row, 1), 0, 0, NULL);
	int x1 = (s - b) * scale;
	int x2 = (f - b) * scale;

	if (x1 < 0)
	  x1 = 0;
	else if (x1 > G_BAR_WIDTH)
	  x1 = G_BAR_WIDTH;

	if (x2 < 0)
	  x2 = 0;
	else if (x2 > G_BAR_WIDTH)
	  x2 = G_BAR_WIDTH;

	gdImageRectangle (im, x1, y, x2, y + G_CHAR_HEIGHT - 1, outline);
	gdImageFilledRectangle (im,
				x1 + 1, y + 1, x2 - 1, y + G_CHAR_HEIGHT - 2,
				fill);
	gdImageString (im, gdFontSmall, G_TITLE_X, y,
		       dbr_value (r, row, 2), black);

	y += G_ROW_HEIGHT;
      }
  }

  /* Send gif image to the user. */
  fputs ("Content-Type: image/gif\n\n", stdout);
  gdImageGif (im, stdout);
}

/* Handle the main calendar screen by show events on the dates of
   interest and displaying graphical calendars too. */
static void
show_calendar (struct request *rq)
{
  /* Duration of interest in two formats. */
  struct tm begin, end;
  struct st beg_st, end_st;

  /* =0: display traditional calendar; !=0: display pseudo-Gantt
     chart. */
  int gantt;

  /* Figure out the dates we're interested in. */
  determine_interval (rq, &begin, &end);
  st_init (&beg_st, 16);
  st_strftime (&beg_st, "%d %b %Y", &begin);
  st_init (&end_st, 16);
  st_strftime (&end_st, "%d %b %Y", &end);

  /* Tell the user what range of dates we're showing. */
  if (!rq->abbrev)
    {
      if (begin.tm_year != end.tm_year
	  || begin.tm_mon != end.tm_mon
	  || begin.tm_mday != end.tm_mday)
	{
	  /* Don't combine these statements: datetime_format() uses a
	     static buffer. */
	  html ("<H2 ALIGN=CENTER>%s %s ", _("Calendar of Events from"),
		st_value (&beg_st));
	  html ("%s %s</H2>\n", _("to"), st_value (&end_st));
	}
      else
	html ("<H2 ALIGN=CENTER>%s %s</H2>\n", _("Calendar of Events for"),
	      st_value (&beg_st));
    }

  /* Determine whether we'll display a traditional calendar or a
     pseudo-Gantt chart. */
  gantt = !strcmp (std_get (rq->cgi, "type"), "gantt");

  /* Display a menu that describes user's date choice options. */
  if (!rq->abbrev)
    {
      htmls ("<TABLE WIDTH=\"100%\" CELLSPACING=0 CELLPADDING=0>\n"
	     "<TR><TD WIDTH=\"65%\">\n"
	     "<P ALIGN=LEFT>\n");
      html_start_form ("calendar", HT_GET, HT_URL_ENCODED);
      html_hidden ("begin", st_value (&beg_st));
      html_hidden ("end", st_value (&end_st));
      html
	("<SELECT NAME=op ONCHANGE=\"form.submit()\">\n"
	 "<OPTION SELECTED>%s...</OPTION>\n"
	 "<OPTGROUP LABEL=\"%s...\">\n"
	 "  <OPTION VALUE=\"day\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"week\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"month\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"goto\">%s...</OPTION>\n"
	 "</OPTGROUP>\n"
	 "<OPTGROUP LABEL=\"%s...\">\n"
	 "  <OPTION VALUE=\"fday\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"fweek\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"fmonth\">%s</OPTION>\n"
	 "</OPTGROUP>\n"
	 "<OPTGROUP LABEL=\"%s...\">\n"
	 "  <OPTION VALUE=\"bday\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"bweek\">%s</OPTION>\n"
	 "  <OPTION VALUE=\"bmonth\">%s</OPTION>\n"
	 "</OPTGROUP>\n"
	 "<OPTION VALUE=\"msp\">%s</OPTION>\n"
	 "</SELECT>\n",

	 _("Calendar Menu"),
	 _("Go to"), _("Today"), _("This week"), _("This month"), _("Other"),
	 _("Go forward"), _("Forward a day"), _("Forward a week"),
	 _("Forward a month"),
	 _("Go back"), _("Back a day"), _("Back a week"), _("Back a month"),
	 _("Import MS Project timeline"));
	 
      html_submit (_("Go!"));
      htmls ("</P>\n");
      html_end_form ();
      htmls ("</TD>\n");
      htmls ("<TD WIDTH=\"35%\">\n");

      /* Allow the user to choose a pseudo-Gantt chart or traditional
         calendar view. */
      {
	static const char *calview = N_("Calendar View");
	static const char *gntview = N_("Gantt View");
	struct url url;
	
	url_create (&url, "calendar");

	url_mark (&url);
	url_arg_int (&url, "beginday", begin.tm_mday);
	url_arg_int (&url, "beginmonth", begin.tm_mon + 1);
	url_arg_int (&url, "beginyear", begin.tm_year + 1900);
	url_arg_int (&url, "endday", end.tm_mday);
	url_arg_int (&url, "endmonth", end.tm_mon + 1);
	url_arg_int (&url, "endyear", end.tm_year + 1900);
	url_arg (&url, "op", "new");
	if (getenv ("REQUEST_URI"))
	  url_arg (&url, "back", getenv ("REQUEST_URI"));
	html ("<P ALIGN=RIGHT>[ <A HREF=\"%s\">%s</A> ]<BR>\n",
	      url_string (&url), _("New Event"));

	url_release (&url);
	url_arg (&url, "begin", st_value (&beg_st));
	url_arg (&url, "end", st_value (&end_st));
	
	if (gantt)
	  {
	    url_arg (&url, "type", "calendar");
	    html ("[ <A HREF=\"%s\">%s</A> | %s ]</P>\n",
		   url_string (&url), gettext (calview), gettext (gntview));
	  }
	else
	  {
	    url_arg (&url, "type", "gantt");
	    html ("[ %s | <A HREF=\"%s\">%s</A> ]</P>\n",
		  gettext (calview), url_string (&url), gettext (gntview));
	  }
	
	url_destroy (&url);
      }
      
      htmls ("</TD></TR>\n"
	     "</TABLE>\n");
    }

  /* Display the Gantt chart if requested. */
  if (gantt)
    {
      struct url url;

      url_create (&url, "ganttgif");
      url_arg (&url, "begin", st_value (&beg_st));
      url_arg (&url, "end", st_value (&end_st));
      html ("<IMG SRC=\"%s\">\n", url_string (&url));
      url_destroy (&url);

      return;
    }

  if (!rq->abbrev)
    htmls ("<TABLE>\n");

  /* Display the event titles and durations within the duration of
     interest, along with calendars showing the dates. */
  {
    /* Events that take place during the entire duration of interest
       should be displayed at the end, but we encounter them at the
       beginning, so we buffer their content here. */
    struct st ongoing;
    size_t n_ongoing;		/* Number of ongoing events so far. */

    /* Scratch buffer. */
    struct st s;

    /* Date of the first month for which we need to output a graphical
       calendar image, or a month title, respectively. */
    struct tm gif;
    struct tm title;

    /* Events within duration of interest. */
    db_result *r;
    size_t n_rows;		/* Number of rows. */
    size_t row;			/* Current row. */

    /* URL to calendar event. */
    struct url url;

    /* Initialize all our local variables and perform the query. */
    gif = begin;

    memset (&title, 0, sizeof title);

    r = calendar_query (&rq->cx, &begin, &end);
    n_rows = dbr_nrows (r);

    st_init (&ongoing, 0);
    n_ongoing = 0;

    st_init (&s, 128);

    url_create (&url, "calendar");
    url_mark (&url);

    /* Loop over all the events, displaying their titles and
       durations. */
    for (row = 0; row < n_rows; row++)
      {
	/* Extract the event from the database result. */
	struct event e;

	e.b_time = datetime_parse (dbr_value (r, row, 0), 0, 0, &e.b_tm);
	e.e_time = datetime_parse (dbr_value (r, row, 1), 0, 0, &e.e_tm);
	e.title = dbr_value (r, row, 2);
	e.seq = stoi (dbr_value (r, row, 3));
	e.owner = dbr_value (r, row, 4);
	e.grp = dbr_value (r, row, 5);

	/* Make the URL for the event. */
	url_release (&url);
	url_arg_int (&url, "seq", e.seq);

	/* Display the event's duration and title.  This is intricate,
	   but not difficult; there are just lots of special cases to
	   allow abbreviation of times and dates. */
	if (tm_ge_day (&e.b_tm, &begin))
	  {
	    /* Event does not start before the period of interest. */

	    while (tm_ge_month (&e.b_tm, &gif))
	      print_calendar (rq, &gif, &begin, &end);
	    catchup_title (&e.b_tm, &title);

	    html_list (1);
	    /* For several of the following cases, notice that some
	       statements can't be combined since datetime_format()
	       uses a static buffer. */
	    if (time_of_day (e.b_time) == 0
		&& e.e_time == e.b_time + 86400)
	      {
		/* Single-day, full-day, day-aligned event.  Nothing to
                   print. */
		;
	      }
	    else if (time_of_day (e.b_time) == 0
		     && time_of_day (e.e_time) == 0)
	      {
		/* Multiple-day, full-day, day-aligned event. */
		html ("%s - ", datetime_format ("%d %b %Y", &e.b_tm));
		html ("%s: ", datetime_format ("%d %b %Y",
						prev_day (&e.e_tm)));
	      }
	    else if (e.e_time <= round_date_down (e.b_time + 86400))
	      {
		/* Partial-day event that does not span days. */
		html ("%s - ", datetime_format ("%H:%M", &e.b_tm));
		html ("%s: ", datetime_format ("%H:%M", &e.e_tm));
	      }
	    else 
	      {
		/* Multiple-day, non-day-aligned event. */
		html ("%s - ", datetime_format ("%H:%M", &e.b_tm));
		html ("%s: ", datetime_format ("%d %b %Y %H:%M", &e.e_tm));
	      }

	    if (!strcmp (e.grp, rq->cx.grp))
	      {
		html ("<A HREF=\"%s\">", url_string (&url));
		html_escape (e.title);
		htmls ("</A>\n");
	      }
	    else
	      {
		html_escape (e.title);
		html (" (%s %s)", e.grp, _("group"));
	      }
	  }
	else
	  /* Event begins before the period of interest. */
	  {
	    st_clear (&s);
	    st_concat (&s, "<LI>");

	    if (time_of_day (e.b_time) == 0) 
	      st_strftime (&s, "%d %b %Y - ", &e.b_tm);
	    else
	      st_strftime (&s, "%d %b %Y %H:%M - ", &e.b_tm);
	    if (time_of_day (e.e_time) == 0)
	      st_strftime (&s, "%d %b %Y", prev_day (&e.e_tm));
	    else 
	      st_strftime (&s, "%d %b %Y %H:%M", &e.e_tm);

	    if (!strcmp (e.grp, rq->cx.grp))
	      {
		st_printf (&s, ": <A HREF=\"%s\">", url_string (&url));
		st_concat_html_quoted (&s, e.title);
		st_concat (&s, "</A>");
	      }
	    else
	      {
		st_concat (&s, ": ");
		st_concat_html_quoted (&s, e.title);
		st_printf (&s, " (%s %s)", e.grp, _("group"));
	      }
	    st_concat (&s, "</LI>\n");

	    if (tm_ge_day (&end, &e.e_tm))
	      {
		while (tm_ge_month (&e.b_tm, &gif))
		  print_calendar (rq, &gif, &begin, &end);
		catchup_title (&e.b_tm, &title);
		htmls (st_value (&s));
	      }
	    else
	      {
		st_concat (&ongoing, st_value (&s));
		n_ongoing++;
	      }
	  }
      }

    /* If all the events were ongoing events, or if there were no
       events at all (same condition), say so. */
    if (n_ongoing == n_rows)
      {
	while (tm_ge_month (&begin, &gif))
	  print_calendar (rq, &gif, &begin, &end);
	catchup_title (&begin, &title);
	htmls ("<LI>No events.\n");
      }

    /* Print any leftover calendars from the duration of interest not
       printed yet. */
    while (tm_ge_month (&end, &gif))
      print_calendar (rq, &gif, &begin, &end);

    /* Print ongoing events if any. */
    if (n_ongoing && !rq->abbrev)
      {
	html_list (0);
	html ("<TR>\n"
	      "<TD>\n"
	      "<TD>\n"
	      "<P>%s</P>\n"
	      "<UL>\n"
	      "%s"
	      "</UL>",
	      _("Ongoing events"),
	      st_value (&ongoing));
      };

    html_list (0);

    if (!rq->abbrev)
      htmls ("</TABLE>\n");
  }

  if (rq->abbrev)
    {
      html ("<P>%s</P>\n", _("Recent calendar activities:"));

      activity_abbreviated (rq, AC_CAT_EVENT);
    }
}

/* Show the screen that allows a user to pick an arbitrary date
   interval. */
static void
show_goto (struct request *rq unused)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s</P>\n",
	_("Go to Date Interval"),
	_("To view calendar events for a particular range of dates,\n"
	  "enter the beginning and ending dates for the interval\n"
	  "below, choose the type of display you'd like to see,\n"
	  "then click on <STRONG>Submit</STRONG>."));

  html_start_form ("calendar", HT_GET, HT_URL_ENCODED);
  html ("<TABLE>\n"
	"<TR><TH>%s<TD>\n"
	"<TABLE CELLPADDING=0 CELLSPACING=0>\n"
	"<TR><TD><SMALL>%s</SMALL></TD>\n"
	"<TR><TD>\n",
	_("Interval"), _("Example: 4 Dec 1998"));
  html_text_field ("begin", "", 11);
  htmls ("</TD><TD>");
  html_text_field ("end", "", 11);
  html ("</TD></TR>\n"
	"<TR><TD>%s</SMALL></TD><TD><SMALL>%s</SMALL></TD></TR></TABLE>\n"
	"<TR><TH>%s<TD>\n"
	"<SELECT NAME=type>\n"
	"  <OPTION VALUE=calendar>Calendar</OPTION>\n"
	"  <OPTION VALUE=gantt>Gantt Chart</OPTION>\n"
	"</SELECT>\n"
	"</TABLE>\n",
	_("begin date"), _("end date"), _("Display Type"));
  html_submit (_("Submit"));
  html_end_form ();
}

/* Displays a graphical calendar clickable imagemap for the month
   represented by CUR.  The dates between BEGIN and END are
   highlighted.  Advances CUR forward one month. */
static void 
print_calendar (struct request *rq, struct tm *cur,
		const struct tm *begin, const struct tm *end)
{
  struct url gif;
  struct url map;
  
  html_list (0);

  /* Create the URL of the image. */
  {
    /* Figure out which days should be highlighted. */
    int ds, de;

    if (cur->tm_year == begin->tm_year && cur->tm_mon == begin->tm_mon)
      ds = begin->tm_mday;
    else
      ds = 1;

    if (cur->tm_year == end->tm_year && cur->tm_mon == end->tm_mon)
      de = end->tm_mday;
    else
      de = datetime_month_days (cur->tm_year, cur->tm_mon);

    /* Create the URL. */
    url_create (&gif, "calendargif");
    url_arg_int (&gif, "year", cur->tm_year + 1900);
    url_arg_int (&gif, "month", cur->tm_mon + 1);
    url_arg_int (&gif, "bday", ds);
    url_arg_int (&gif, "eday", de);
  }

  /* Create the URL of the imagemap. */
  {
    char buf[INT_DIGITS * 2 + 16];
    sprintf (buf, "calendar-%d-%d", cur->tm_year + 1900, cur->tm_mon + 1);
    url_create (&map, buf);
  }

  /* Write the necessary HTML. */
  htmls (rq->abbrev ? "<P ALIGN=CENTER>" : "<TR VALIGN=TOP><TD>");
  html ("<A HREF=\"%s\"><IMG BORDER=0 SRC=\"%s\" ISMAP ALT=\"\"></A>",
	url_string (&map), url_string (&gif));
  htmls (rq->abbrev ? "</P>" : "</TD><TD>\n");

  /* Advance CUR by one month. */
  cur->tm_mon++;
  if (cur->tm_mon >= 12)
    {
      assert (cur->tm_mon == 12);
      cur->tm_mon = 0;
      cur->tm_year++;
    }

  /* Garbage collection. */
  url_destroy (&gif);
  url_destroy (&map);
}

/* Display the calendar event with sequence number SEQ. */
static void
show_calendar_event (struct request *rq, int seq)
{
  struct event e;		/* Event in question. */
  db_result *r;			/* Database result. */

  int edit;			/* 1=Allow user to edit event. */

  /* Handle event deletion. */
  db_begin ();
  if (std_contains (rq->cgi, "delete")) 
    {
      db_result *r;

      r = retrieve_event (&rq->cx, seq, &e);
      if (r != NULL)
	{
	  if (strcmp (rq->cx.grp, e.grp))
	    fail ("user is trying to delete event that doesn't belong "
		  "to his workgroup");

	  db_query ("DELETE FROM events WHERE seq = %d", DB_FORMAT, NULL, seq);
	  db_query ("DELETE FROM event_resource WHERE event = %d",
		    DB_FORMAT, NULL, seq);

	  log_activity_by_name (&rq->cx, e.title, NULL, AC_CAT_EVENT,
				AC_EVENT_DELETE, seq);

	  html ("<H2>%s</H2>\n"
		"<P>%s</P>\n",
		_("Event Deleted"),
		_("The selected event has been deleted."));

	  free (e.res);
	  dbr_destroy (r);
	  db_end ();
	  return;
	}

    }

  edit = std_contains (rq->cgi, "edit");

  /* Handle change requests. */
  if (std_contains (rq->cgi, "change"))
    {
      int success;
      int title;

      title = get_file_seq (&rq->cx, std_get (rq->cgi, "title"));

      success = db_query ("UPDATE events "
			  "SET start = '%sq', finish = '%sq', title = %d, "
			  "note = '%sq', owner = '%sq' "
			  "WHERE seq = %d AND grp = '%sq'",
			  DB_FORMAT | DB_FAIL_OK, NULL,
			  std_get (rq->cgi, "start"),
			  std_get (rq->cgi, "finish"),
			  title,
			  std_get (rq->cgi, "note"),
			  std_get (rq->cgi, "owner"),
			  seq, rq->cx.grp) != -1;

      if (success)
	success &= db_query ("DELETE FROM event_resource WHERE event = %d",
			     DB_FORMAT | DB_FAIL_OK, NULL, seq) != -1;

      /* Update resources related to the event if successful. */
      if (success)
	{
	  const char **res = std_get_many (rq->cgi, "resources", NULL);
	  if (res != NULL)
	    while (*res != NULL)
	      {
		success &= db_query ("INSERT INTO event_resource "
				     "VALUES (%d, %d, nextval ('seq'))",
				     DB_FORMAT | DB_FAIL_OK, NULL,
				     seq, stoi (*res)) != -1;
		if (!success)
		  break;
		res++;
	      }

	  log_activity (&rq->cx, title, 0, AC_CAT_EVENT, AC_EVENT_CHANGE, seq);
	}
      
      if (!success)
	{
	  html ("<H2 ALIGN=CENTER>%s</H2>\n"
		"<P>%s:</P>\n"
		"<BLOCKQUOTE>%s</BLOCKQUOTE>\n"
		"<P>%s</P>\n",
		_("Error Updating Record"),
		_("The SQL backend returned the following error message"),
		db_error (),
		_("Fix the error in the input data below and resubmit "
		  "your changes."));
	  edit = 1;
	}
      else
	html ("<H2 ALIGN=CENTER>%s</H2>\n"
	      "<P>%s</P>\n",
	      _("Record Updated"),
	      _("The data for this calendar event was successfully "
		"changed.  Please verify your changes below."));
    }

  /* Retrieve the event and make sure that it exists. */
  r = retrieve_event (&rq->cx, seq, &e);
  if (r == NULL)
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P>%s</P>\n",
	    _("Event Not Found"),
	    _("The event you selected was not found.\n"
	      "This probably indicates that it has been deleted\n"
	      "by another member of your workgroup."));
      db_end ();
      return;
    }
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("Event View"));

  html_start_form ("calendar", HT_POST, HT_URL_ENCODED);
  html_hidden ("seq", std_get_null (rq->cgi, "seq"));

  /* Display the event. */
  if (edit)
    {
      int *res;
      int n_res;
      
      html_hidden ("change", "yes");

      html ("<TABLE>\n"
	    "<TR><TH>%s:<TD>",
	    _("Title"));
      html_text_field ("title", e.title, 50);
      html ("<TR><TH>%s:<TD>", _("Begins"));
      html_text_field ("start", e.b_string, 50);
      html ("<TR><TH>%s:<TD>", _("Ends"));
      html_text_field ("finish", e.e_string, 50);
      html ("<TR><TH>%s:<TD>", _("Notes"));
      html_text_area ("note", e.note, 5, 50);
      
      read_user_list (&rq->cx.wg, rq->cx.grp);
      html ("<TR><TH>%s:<TD>\n"
	    "<SELECT NAME=owner>", _("User"));
      {
	int i;

	html ("<OPTION VALUE=\"\">---%s---</OPTION>\n", _("none"));
	for (i = 0; i < rq->cx.wg->n; i++)
	  html ("<OPTION VALUE=\"%s\"%s>%s</OPTION>\n",
		rq->cx.wg->name[i],
		strcmp (rq->cx.wg->name[i], e.owner) ? "" : " SELECTED",
		rq->cx.wg->fullname[i]);
      }
      html ("</SELECT>\n"
	    "<TR><TH>%s:<TD>\n",
	    _("Resources"));
      
      {
	db_result *r;
	int i;

	db_query ("SELECT resource FROM event_resource "
		  "WHERE event = %d ORDER BY resource",
		  DB_TUPLES | DB_FORMAT, &r, seq);

	n_res = dbr_nrows (r);
	res = xmalloc (sizeof *res * n_res);
	for (i = 0; i < n_res; i++)
	  res[i] = stoi (dbr_value (r, i, 0));
	dbr_destroy (r);
      }

      read_resource_list ();
      {
	int i;

	htmls ("<SELECT NAME=resources MULTIPLE SIZE=5>\n");
	for (i = 0; i < n_resources; i++)
	  {
	    int selected = list_contains_value (res, n_res, resource[i].seq);
						
	    html ("<OPTION VALUE=\"%d\"%s>%s</OPTION>\n",
		  resource[i].seq, selected ? " SELECTED" : "",
		  resource[i].name);
	  }
	htmls ("</SELECT>\n");
      }

      html ("<A HREF=\"%s\">%s</A>\n",
	    url_this ("resources", (char *) NULL),
	    _("Edit Resource List"));

      htmls ("</TABLE>\n");

      html_submit (_("Submit Changes"));
    }
  else
    {
      show_event (rq, &e);
      show_viewed_by (&rq->cx, seq, e.title, NULL,
		      AC_CAT_EVENT, AC_EVENT_VIEW);

      html_submit_alt ("edit", _("Edit Event"));
      htmlc (' ');
      html_submit_alt ("delete", _("Delete Event"));
    }

  free (e.res);
  dbr_destroy (r);

  html_end_form ();

  db_end ();
}

/* Display a set of three drop-down menus representing a date.  They
   will be stored into fields named `day', `month', and `year'
   prefixed by BASENAME.  Use the contents of correspondingly named
   CGI parameters as default if ENABLE_DEFAULT is nonzero, otherwise
   no default date given. */
static void
date_field (struct request *rq, const char *basename, int enable_default)
{
  struct st s;
  int i;
  
  st_init (&s, 16);
  
  {
    int day;
    
    st_printf (&s, "%sday", basename);
    if (!enable_default
	|| !sscanf (std_get (rq->cgi, st_value (&s)), "%d", &day))
      day = -1;
  
    html ("<SELECT NAME=%s>\n"
	  "<OPTION%s>--</OPTION>",
	  st_value (&s),
	  day == -1 ? " SELECTED" : "");
    for (i = 1; i <= 31; i++)
      html ("  <OPTION%s>%d</OPTION>\n", 
	    day == i ? " SELECTED" : "",
	    i);
    html ("</SELECT>\n");
  }
  
  {
    int month;

    st_clear (&s);
    st_printf (&s, "%smonth", basename);
    if (!enable_default
	|| !sscanf (std_get (rq->cgi, st_value (&s)), "%d", &month))
      month = -1;
    else
      month--;
    
    html ("<SELECT NAME=%smonth>\n"
	  "<OPTION%s>--</OPTION>\n",
	  basename,
	  month == -1 ? " SELECTED" : "");
    for (i = 0; i < 12; i++)
      html ("  <OPTION VALUE=%d%s>%.3s</OPTION>\n",
	    i + 1, 
	    month == i ? " SELECTED" : "", 
	    "JanFebMarAprMayJunJulAugSepOctNovDec" + 3 * i);
    html ("</SELECT>\n");
  }
  
  {
    int year;

    st_clear (&s);
    st_printf (&s, "%syear", basename);
    if (!enable_default
	|| !sscanf (std_get (rq->cgi, st_value (&s)), "%d", &year))
      year = -1;
    else
      year -= 1900;
  
    html ("<SELECT NAME=%syear>\n"
	  "<OPTION%s>----</OPTION>\n",
	  basename,
	  year == -1 ? " SELECTED" : "");
    for (i = cur_tm.tm_year - 3; i <= cur_tm.tm_year + 3; i++)
      html ("  <OPTION%s>%d</OPTION>\n",
	    year == i ? " SELECTED" : "",
	    i + 1900);
    htmls ("</SELECT>\n");
  }
}
  
/* Display a form that allows users to add new calendar events. */
static void
new_calendar_event (struct request *rq, const struct event *e)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("New Calendar Event"));

  html_start_form ("calendar", HT_POST, HT_URL_ENCODED);
  html_hidden ("op", "add");
  html_hidden ("back", std_get (rq->cgi, "back"));
  
  html ("<TABLE>\n"
	"<TR VALIGN=TOP>\n"
	"  <TH>%s</TH>\n"
	"  <TD>%s<BR>\n"
	"    <TABLE CELLPADDING=0 CELLSPACING=0>\n"
	"      <TR>\n"
	"        <TD>",
	_("Date"),
	_("Enter the beginning and ending dates for the new event.\n"
	  "If the event takes place on a single day, then the end\n"
	  "date may be omitted."));

  date_field (rq, "begin", 1);
  htmls ("        </TD>\n"
	 "        <TD>");
  date_field (rq, "end", 0);
  
  html ("        </TD>\n"
	"      </TR>\n"
	"      <TR>\n"
	"        <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"        <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"      </TR>\n"
	"    </TABLE>\n"
	"  </TD>\n"
	"<TR VALIGN=TOP>\n"
	"  <TH>%s\n"
	"  <TD>%s<BR>\n"
	"    <TABLE CELLPADDING=0 CELLSPACING=0>\n"
	"      <TR>\n"
	"         <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"         <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"      </TR>\n"
	"      <TR>\n"
	"        <TD>",
	_("begin date"),
	_("end date"),
	_("Time"),
	_("Enter the beginning and ending times for the beginning and\n"
	  "ending dates, respectively.\n"
	  "Both times are optional.\n"
	  "If either time is omitted then the entire beginning or\n"
	  "ending date, as appropriate, will be included in the\n"
	  "event duration.\n"),
	_("Example: 11:30am"),
	_("Example: 3:00pm"));
  html_text_field ("begintime", std_get (rq->cgi, "begintime"), 10);
  htmls ("        </TD>\n"
	 "        <TD>\n");
  html_text_field ("endtime",
		   std_get (rq->cgi, "endtime"), 10);
  html ("        </TD>\n"
	"      <TR>\n"
	"        <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"        <TD><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	"      </TR>\n"
	"    </TABLE>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n",
	_("begin time"),
	_("end time"),
	_("Title"),
	_("Enter a short title for the task or event."));
  html_text_field ("title", e && e->title ? e->title : "", 50);
  html ("  </TR>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n",
	_("Notes"),
	_("Optionally, you can enter any amount of additional\n"
	  "explanatory text about this event or task in the\n"
	  "space below."));
  html_text_area ("note", e && e->note ? e->note : "", 5, 50);
  html ("  </TR>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n",
	_("Resources"),
	_("If this event requires the use of one or more\n"
	  "shared resources, select those resources in the\n"
	  "box below.\n"
	  "You may need to hold down Control or Alt on\n"
	  "your keyboard to select multiple items."));

  read_resource_list ();
  {
    int i;
    
    htmls ("<SELECT NAME=resources MULTIPLE SIZE=5>\n");
    for (i = 0; i < n_resources; i++)
      html ("<OPTION VALUE=\"%d\"%s>%s</OPTION>\n",
	    resource[i].seq,
	    (e && list_contains_value (e->res, e->n_res, resource[i].seq)
	     ? " SELECTED" : ""),
	    resource[i].name);
    htmls ("</SELECT>\n");
  }
  
  html ("  </TR>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n",
	_("User"),
	_("If a particular group member is the contact person or "
	  "otherwise responsible for this task\n"
	  "or event, select the group member's name below."
	  "If resources are used (see below), this field is mandatory."));

  read_user_list (&rq->cx.wg, rq->cx.grp);
  {
    int i;

    html ("      <SELECT NAME=owner>\n"
	  "        <OPTION VALUE=\"\">---%s---</OPTION>\n", _("none"));
    for (i = 0; i < rq->cx.wg->n; i++)
      html ("        <OPTION VALUE=\"%s\"%s>%s</OPTION>\n",
	    rq->cx.wg->name[i],
	    (e && e->owner && !strcmp (e->owner, rq->cx.wg->name[i])
	     ? " SELECTED" : ""),
	    rq->cx.wg->fullname[i]);
    htmls ("      </SELECT>\n");
  }

  html ("<A HREF=\"%s\">%s</A>\n",
	url_this ("resources", (char *) NULL),
	_("Edit Resource List"));

  html ("    </TD>\n"
	"  </TR>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n",
	_("Submit"),
	_("When you're ready to add this event to the calendar,\n"
	  "click on the <STRONG>Submit</STRONG> button below."));
  html_submit (_("Submit"));

  htmls ("  </TR>\n"
	 "</TABLE>\n");

  html_end_form ();
}

static const char *try_add_calendar_event (struct request *, struct event *);
static void clear_event (struct event *);

/* Add a new calendar event as specified by the user. */
static void
add_calendar_event (struct request *rq)
{
  /* Event structure populated according to user instructions. */
  struct event e;
  const char *error;

  clear_event (&e);
  error = try_add_calendar_event (rq, &e);

  /* Finish up by displaying the results. */
  if (error != NULL)
    {
      /* Report an error. */
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P>%s: %s</P>\n"
	    "<P>%s</P>\n"
	    "<H2 ALIGN=CENTER>%s</H2>\n",
	    _("Error Adding New Calendar Event"),
	    _("Error"),
	    error,
	    _("Examine the data supplied for the new event below,\n"
	      "then adjust the data for the event as required and\n"
	      "click the <STRONG>Submit</STRONG> button again.\n"),
	    _("Data Supplied"));

      new_calendar_event (rq, &e);
    }
  else
    {
      /* Add the event. */
      int seq = get_seq ();
      int title = get_file_seq (&rq->cx, e.title);
      
      db_query ("INSERT INTO events VALUES "
		"('%sq', '%sq', %d, '%sq', '%sq', '%sq', %d, 0)",
		DB_FORMAT, NULL, e.b_string, e.e_string, title,
		e.note, e.owner ? e.owner : "", rq->cx.grp, seq);
      
      {
	const char **res = std_get_many (rq->cgi, "resources", NULL);
	if (res != NULL)
	  while (*res != NULL)
	    {
	      db_query ("INSERT INTO event_resource "
			"VALUES (%d, %d, nextval ('seq'))",
			DB_FORMAT | DB_FAIL_OK, NULL,
			seq, stoi (*res));
	      res++;
	    }
      }

      log_activity (&rq->cx, title, 0, AC_CAT_EVENT, AC_EVENT_NEW, seq);

      html ("<H2 ALIGN=CENTER>%s</H2>\n",
	    _("New Calendar Event Scheduled"));

      show_event (rq, &e);

      if (std_get_nonempty (rq->cgi, "back") != NULL)
	html ("<P ALIGN=CENTER>\n"
	      "  <A HREF=\"%s\">Return to previous calendar view</A>\n"
	      "</P>\n",
	      std_get (rq->cgi, "back"));
    }

  free ((char *) e.b_string);
  free ((char *) e.e_string);
}

/* Try to parse a calendar event from CGI.  If successful, return a
   null pointer.  Otherwise, return an error message. */
static const char *
try_add_calendar_event (struct request *rq, struct event *e)
{
  /* Calculate the beginning and ending dates.
     Not complicated, just ugly. */
  {
    const char *begin_day, *begin_month, *begin_year;
    const char *end_day, *end_month, *end_year;
    const char *begin_time, *end_time;

    struct st s;

    begin_day = std_get (rq->cgi, "beginday");
    begin_month = std_get (rq->cgi, "beginmonth");
    begin_year = std_get (rq->cgi, "beginyear");
    end_day = std_get (rq->cgi, "beginday");
    end_month = std_get (rq->cgi, "beginmonth");
    end_year = std_get (rq->cgi, "beginyear");
    begin_time = std_get (rq->cgi, "begintime");
    end_time = std_get (rq->cgi, "endtime");
    
    {
      int x;
      
      x = empty (begin_day) + empty (begin_month) + empty (begin_year);
      if (x != 0 && x != 3)
	return _("Begin date specified partially.");

      x = empty (end_day) + empty (end_month) + empty (end_year);
      if (x != 0 && x != 3)
	return _("End date specified partially.");
    }
    
    st_init (&s, 128);
    st_printf (&s, "%s/%s/%s %s",
	       begin_year, begin_month, begin_day, begin_time);
    e->b_time = datetime_parse (st_value (&s), DT_DATE_REQUIRED, 0, &e->b_tm);
    if (!e->b_time)
      {
	st_destroy (&s);
	return datetime_error;
      }
    
    e->b_string = xstrdup (datetime_format ("%a %b %d %H:%M:%S %Y %Z",
					    &e->b_tm));

    st_clear (&s);
    if (empty (end_day))
      st_printf (&s, "%s/%s/%s %s",
		 begin_year, begin_month, begin_day, end_time);
    else
      st_printf (&s, "%s/%s/%s %s",
		 end_year, end_month, end_day, end_time);
    e->e_time = datetime_parse (st_value (&s), 0, 0, &e->e_tm);
    if (e->e_time)
      {
	if (empty (end_time))
	  {
	    e->e_tm.tm_mday++;
	    e->e_time = mktime (&e->e_tm);
	  }

	e->e_string = xstrdup (datetime_format ("%a %b %d %H:%M:%S %Y %Z",
						&e->e_tm));
      }

    st_destroy (&s);
    
    if (!e->e_time)
      return datetime_error;
  }

  /* Fill out the rest of the structure. */
  e->title = std_get (rq->cgi, "title");
  e->note = std_get (rq->cgi, "note");
  e->owner = std_get (rq->cgi, "owner");
  if (has_prefix (e->owner, "---"))
    e->owner = NULL;
  e->grp = rq->cx.grp;
  {
    const char **res = std_get_many (rq->cgi, "resources", &e->n_res);
    if (e->n_res)
      {
	int i;

	e->res = xmalloc (sizeof *e->res * e->n_res);
	for (i = 0; i < e->n_res; i++)
	  e->res[i] = stoi (res[i]);
      }
    else
      e->res = NULL;
  }

  /* Check for errors. */
  if (e->b_time == 0 || e->e_time == 0)
    return _("Error parsing start or end date/time.");
  if (e->b_time > e->e_time)
    return _("Begin date/time must precede end date/time.\n"
	     "Pay careful attention to AM/PM qualifiers when\n"
	     "considering time order.");
  if (e->b_time == e->e_time)
    return _("Begin date/time cannot equal end date/time.");
  if (empty (e->title))
    return _("Title field is required.");
  if (e->owner == NULL && e->res != NULL)
    return _("Contact person must be specified if shared resources "
	     "are required.");

  return NULL;
}

/* Initialize event E. */
static void
clear_event (struct event *e)
{
  e->b_time = 0;
  e->b_string = NULL;
  e->e_time = 0;
  e->e_string = NULL;
  e->title = NULL;
  e->note = NULL;
  e->owner = NULL;
  e->grp = NULL;
  e->res = NULL;
  e->n_res = 0;
  e->seq = 0;
}
  
/* Figure out which days in MONTH of YEAR have events scheduled for
   them.  Puts the number of events on each day in that month in
   EVENTS[day]. */
static void 
find_event_days (struct context *cx, int year, int month, short events[31])
{
  struct tm begin, end;
  time_t beg_t, end_t;

  db_result *r;
  size_t n_rows;
  size_t row;

  memset (&begin, 0, sizeof begin);
  memset (&end, 0, sizeof end);
  begin.tm_mday = 1;
  end.tm_mday = datetime_month_days (year, month);
  begin.tm_mon = end.tm_mon = month;
  begin.tm_year = end.tm_year = year;
  beg_t = mktime (&begin);
  end_t = mktime (&end);

  r = calendar_query (cx, &begin, &end);
  n_rows = dbr_nrows (r);
  memset (events, 0, sizeof *events * 31);
  for (row = 0; row < n_rows; row++)
    {
      time_t b, e;

      b = datetime_parse (dbr_value (r, row, 0), 0, 0, NULL);
      e = datetime_parse (dbr_value (r, row, 1), 0, 0, NULL);

      if (b < beg_t)
	b = beg_t;
      if (e > end_t)
	e = end_t;

      for (b = round_date_down (b); b < e; b += 86400)
	events[(b - beg_t - tzoff) / 86400]++;
    }
}

/* Displays event E for the user. */
static void
show_event (struct request *rq, const struct event *e)
{
  const char *owner;

  if (e->owner && e->owner[0])
    {
      read_user_list (&rq->cx.wg, rq->cx.grp);
      owner = username_to_fullname (rq->cx.wg, e->owner);
    }
  else
    owner = _("(none)");
      
  html ("<TABLE>\n"
	"<TR><TH>%s:<TD>", _("Title"));
  html_escape (e->title);

  html ("\n"
	"<TR><TH>%s:<TD>%s\n"
	"<TR><TH>%s:<TD>%s\n"
	"<TR><TH>%s:<TD>%s\n",
	_("Begins"), e->b_string,
	_("Ends"), e->e_string,
	_("User"), owner);
  html ("<TR><TH>%s:<TD>", _("Notes"));
  if (e->note && e->note[0])
    html_pretty_print (e->note);
  else
    htmls (_("(none)"));

  html ("<TR><TH>%s:<TD>\n", _("Resources"));

  if (e->n_res)
    {
      db_result *r;
      struct st query;

      int n_rows;
      int i;

      st_create (&query, "SELECT name FROM resources WHERE seq IN (");
      for (i = 0; i < e->n_res; i++)
	{
	  if (i != 0)
	    st_concat (&query, ", ");
	  st_printf (&query, "%d", e->res[i]);
	}
      st_putchar (&query, ')');
      db_query (st_value (&query), DB_TUPLES, &r);
      st_destroy (&query);

      n_rows = dbr_nrows (r);
      for (i = 0; i < n_rows; i++)
	{
	  html_escape (dbr_value (r, i, 0));
	  htmls ("<BR>\n");
	}

      dbr_destroy (r);
    }
  else
    htmls (_("(none)"));

  htmls ("</TABLE>\n");

  check_conflicts (&rq->cx, e);
}

/* Add a textual description of event E to string S. */
void
append_event_string (struct st *s, const struct event *e)
{
  int i;
  
  st_printf (s, "(%s, %d/\"%s\", %s - %s, %s: ",
	     e->grp, e->seq, e->title, e->b_string, e->e_string,
	     e->owner && e->owner[0] ? e->owner : "nobody");

  for (i = 0; i < e->n_res; i++)
    {
      if (i)
	st_concat (s, ", ");
      
      st_printf (s, "\"%s\"", resource_name (e->res[i]));
    }

  st_putchar (s, ')');
}

/* Check for conflicts with event E, and display them if any. */
void
check_conflicts (struct context *cx, const struct event *e)
{
  struct st log_msg;
  
  db_result *r;
  int count;
  int row;

  db_query ("SELECT start, finish, string AS title, note, owner, "
	    "events.grp, events.seq "
	    "FROM events, strings "
	    "WHERE events.seq IN "
	    "(SELECT DISTINCT event FROM event_resource WHERE resource IN "
	    "(SELECT resource FROM event_resource WHERE event = %d)) "
	    "AND (abstime (start) <#> abstime (finish)) "
	    "&& ('%sq' <#> '%sq') "
	    "AND events.seq <> %d "
	    "AND events.title = strings.seq",
	    DB_FORMAT | DB_TUPLES, &r,
	    e->seq, e->b_string, e->e_string, e->seq);

  count = dbr_nrows (r);
  if (count == 0)
    {
      dbr_destroy (r);
      return;
    }

  read_resource_list ();
  html ("<P><STRONG>%s</STRONG>: %s\n"
	"<TABLE>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"    <TH ALIGN=LEFT>%s</TH>\n"
	"  </TR>\n",
	_("Warning"),
	_("The following resource conflicts have been discovered.\n"
	  "You may resolve these conflicts by editing the event details.\n"
	  "If conflict resolution is uncertain, contact the listed contact\n"
	  "person."),
	_("Start"), _("End"), _("Event"), _("Team"),
	_("Contact"), _("Resources"));

  st_init (&log_msg, 128);
  st_printf (&log_msg, "resource conflict: ");
  append_event_string (&log_msg, e);

  for (row = 0; row < count; row++)
    {
      struct event e;
      int i;

      interpret_event (&e, r, row);

      st_concat (&log_msg, ", ");
      append_event_string (&log_msg, &e);

      html ("  <TR>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n"
	    "    <TD>%s</TD>\n",
	    e.b_string, e.e_string, e.title, e.grp);

      if (e.owner && !empty (e.owner))
	html ("    <TD><A HREF=\"mailto:%s@%s\">%s@%s</A>\n",
	      e.owner, mail_domain, e.owner, mail_domain);
      else
	html ("    <TD>(%s)</TD>\n", _("none"));

      htmls ("    <TD>");
      for (i = 0; i < e.n_res; i++)
	{
	  if (i != 0)
	    htmls ("<BR>\n"
		   "        ");

	  htmls (resource_name (e.res[i]));
	}
      htmls ("</TD>\n"
	     "  </TR>\n");

      free (e.res);
    }
  htmls ("</TABLE>\n");

  st_truncate (&log_msg, 4096);
  db_query ("INSERT INTO uselog VALUES ('nobody', 'now'::datetime, '%sq', "
	    "'%sq', 'resource-conflict')",
	    DB_FORMAT, NULL, st_value (&log_msg), cx->grp);
  st_destroy (&log_msg);
}


/* Figure out what the duration of interest is, and put it into BEG
   and END.  This is more complicated than it might be since we put
   the smarts onto this end instead of the other end, but there's no
   way to avoid having the complication somewhere. */
static void
determine_interval (struct request *rq, struct tm *beg, struct tm *end)
{
  time_t b, e;

  /* There are two basic ways to determine the initial duration of
     interest.  The most common way is through the `begin' and `end'
     CGI arguments, which specify a pair of dates directly.  The other
     way is that the actual screen name passed as part of the main URL
     (not part of the query string) can have a year and a day coded
     into it.  This kluge is used because there doesn't seem to be any
     other universally supported way to pass arguments to an ISMAP. */
  if (strchr (rq->screen, '-'))
    {
      int year, month;
      int x, y;

      if (sscanf (rq->screen, "calendar-%d-%d", &year, &month) == 2
	  && sscanf (std_get (rq->cgi, "loc"), "%d,%d", &x, &y) == 2)
	{
	  beg->tm_mon = month - 1;
	  beg->tm_year = year - 1900;
	  beg->tm_sec = beg->tm_min = 0;
	  beg->tm_hour = 12;
	  beg->tm_isdst = -1;
	  
	  if (y >= C_FIRST_ROW_Y) 
	    {
	      int col = x / (C_DAY_WIDTH + C_MARGIN);
	      int row = (y - C_FIRST_ROW_Y) / C_ROW_HEIGHT;

	      /* Determine weekday at start of month. */
	      beg->tm_mday = 1;
	      mktime (beg);

	      /* Determine day of month based on the starting weekday. */
	      beg->tm_mday = col + 7 * row - beg->tm_wday + 1;
	      b = e = mktime (beg);
	      *end = *beg;
	    }
	  else
	    {
	      beg->tm_mday = 1;
	      *end = *beg;
	      end->tm_mday = datetime_month_days (end->tm_year, end->tm_mon);
	      
	      if (x <= C_TEXT_WIDTH + C_MARGIN)
		{
		  dec_month (beg);
		  dec_month (end);
		}
	      else if (x >= C_WIDTH - C_TEXT_WIDTH - C_MARGIN)
		{
		  inc_month (beg);
		  inc_month (end);
		}
	      b = mktime (beg);
	      e = mktime (end);
	    }
	}
      else
	fail ("bad location: %s, %s, %d, %d, %d, %d",
	      rq->screen, std_get (rq->cgi, "loc"),
	      year, month, x, y);
    }
  else
    {
      b = e = 0;
      if (std_contains (rq->cgi, "begin"))
	b = datetime_parse (std_get (rq->cgi, "begin"),
			    DT_DATE_REQUIRED, 0, beg);
      if (std_contains (rq->cgi, "end"))
	{
	  const char *end_string = std_get (rq->cgi, "end");

	  if (!empty (end_string))
	    e = datetime_parse (end_string, DT_DATE_REQUIRED, 0, end);
	  else
	    {
	      e = b;
	      *end = *beg;
	    }
	}
    }

  /* Now consider whether any modifiers or special specifiers were
     given.  If so, apply them to the date already determined. */

  /* Absolute specifiers. */
  if (!strcmp (rq->op, "nextmonth")) /* Next month. */
    {
      *beg = *end = cur_tm;
      inc_month (beg);
      inc_month (end);
      beg->tm_mday = 1;
      end->tm_mday = datetime_month_days (end->tm_year, end->tm_mon);
    }
  else if (!strcmp (rq->op, "lastmonth")) /* Last month. */
    {
      *beg = *end = cur_tm;
      dec_month (beg);
      dec_month (end);
      beg->tm_mday = 1;
      end->tm_mday = datetime_month_days (end->tm_year, end->tm_mon);
    }
  else if (!strcmp (rq->op, "month")) /* This month. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday = 1;
      end->tm_mday = datetime_month_days (end->tm_year, end->tm_mon);
    }
  else if (!strcmp (rq->op, "nextweek")) /* Next week. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday += -beg->tm_wday + 7;
      end->tm_mday = beg->tm_mday + 6;
    }
  else if (!strcmp (rq->op, "lastweek")) /* Last week. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday += -beg->tm_wday - 7;
      end->tm_mday = beg->tm_mday + 6;
    }
  else if (!strcmp (rq->op, "week")) /* This week. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday -= beg->tm_wday;
      end->tm_mday = beg->tm_mday + 6;
    }
  else if (!strcmp (rq->op, "tomorrow")) /* Tomorrow. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday++;
      end->tm_mday++;
    }
  else if (!strcmp (rq->op, "yesterday")) /* Yesterday. */
    {
      *beg = *end = cur_tm;
      beg->tm_mday--;
      end->tm_mday--;
    }
  else if (b == 0 || e == 0 || b > e || !strcmp (rq->op, "day")) /* Today. */
    *beg = *end = cur_tm;

  /* Relative specifiers. */
  else if (!strcmp (rq->op, "fmonth")) /* Forward a month. */
    {
      inc_month (beg);
      inc_month (end);
    }
  else if (!strcmp (rq->op, "bmonth")) /* Back a month. */
    {
      dec_month (beg);
      dec_month (end);
    }
  else if (!strcmp (rq->op, "fweek")) /* Forward a week. */
    {
      beg->tm_mday += 7;
      end->tm_mday += 7;
    }
  else if (!strcmp (rq->op, "bweek")) /* Back a week. */
    {
      beg->tm_mday -= 7;
      end->tm_mday -= 7;
    }
  else if (!strcmp (rq->op, "fday")) /* Forward a day. */
    {
      beg->tm_mday++;
      end->tm_mday++;
    }
  else if (!strcmp (rq->op, "bday")) /* Back a day. */
    {
      beg->tm_mday--;
      end->tm_mday--;
    }

  mktime (beg);
  mktime (end);

  /* Include the entire ending day. */
  end->tm_sec = 59;
  end->tm_min = 59;
  end->tm_hour = 23;

#if DEBUGGING
  log ("beg: %d %d %d; end: %d %d %d",
       beg->tm_year, beg->tm_mon, beg->tm_mday,
       end->tm_year, end->tm_mon, end->tm_mday);
#endif
}

/* Retrieve all the events between BEGIN and END and
   return the database result. */
static db_result *
calendar_query (struct context *cx,
		const struct tm *begin, const struct tm *end)
{
  char b[32], e[32];
  db_result *r;

  if (!strftime (b, 32, "%b %d %Y", begin)
      || !strftime (e, 32, "%b %d %Y 23:59:59", end))
    fail ("date too long");

  db_query ("SELECT start, finish, string AS title, events.seq, owner,"
	    " events.grp "
	    "FROM events, strings "
	    "WHERE"
	    " events.grp IN (SELECT grp FROM user_group WHERE usr = '%sq')"
	    " AND events.title = strings.seq"
	    " AND (abstime (start) <#> abstime (finish)) && ('%sq' <#> '%sq') "
	    "ORDER BY start",
	    DB_FORMAT | DB_TUPLES, &r, cx->usr, b, e);

  return r;
}

/* Retrieve the event with sequence number SEQ from the database and
   put all of its info into event E. */
static db_result *
retrieve_event (struct context *cx, int seq, struct event *e)
{
  db_result *r;
  
  db_query ("SELECT start, finish, string AS title, note, owner, "
	    "events.grp, events.seq "
	    "FROM events, strings "
	    "WHERE events.seq = %d AND events.grp = '%sq' AND "
	    "events.title = strings.seq",
	    DB_FORMAT | DB_TUPLES, &r, seq, cx->grp);
  if (dbr_nrows (r) == 0)
    {
      dbr_destroy (r);
      return NULL;
    }

  interpret_event (e, r, 0);

  return r;
}

static void
interpret_event (struct event *e, db_result *r1, int row)
{
  db_result *r2;
  int i;
  
  e->b_string = dbr_value (r1, row, 0);
  e->e_string = dbr_value (r1, row, 1);
  e->title = dbr_value (r1, row, 2);
  e->note = dbr_value (r1, row, 3);
  e->owner = dbr_value (r1, row, 4);
  e->grp = dbr_value (r1, row, 5);
  e->seq = stoi (dbr_value (r1, row, 6));

  db_query ("SELECT resource FROM event_resource WHERE event = %d",
	    DB_FORMAT | DB_TUPLES, &r2, e->seq);
  e->n_res = (int) dbr_nrows (r2);
  e->res = xmalloc (sizeof *e->res * (e->n_res + 1));
  for (i = 0; i < e->n_res; i++)
    e->res[i] = stoi (dbr_value (r2, i, 0));
  e->res[e->n_res] = 0;

  dbr_destroy (r2);
}

/* Return nonzero only if sorted list LIST with N_LIST elements
   contains VALUE. */
static int 
list_contains_value (int *list, int n_list, int value)
{
  /* Knuth's Algorithm 6.2.1B. */
  int l, u;

  l = 1;
  u = n_list;
  for (;;)
    {
      int i;
      
      if (u < l)
	return 0;

      i = (l + u) / 2;
      if (value < list[i - 1])
	u = i - 1;
      else if (value > list[i - 1])
	l = i + 1;
      else
	return 1;
    }
}

/* Center text string S in color C at horizontal position X on
   character line LINE on image IM. */
static void 
center_text (gdImagePtr im, int line, int x, int color, const char *s)
{
  int pos = x - G_CHAR_WIDTH / 2 * strlen (s);
  if (pos < 0)
    pos = 0;
  gdImageString (im, gdFontSmall, pos, G_CHAR_HEIGHT * line,
		 (char *) s, color);
}

/* Calculates the difference from GMT of the current timezone.  There
   are various horrible problems with the whole concept behind this,
   but nobody else has noticed yet, so I haven't fixed it. */
/* Concept from Perl Time::Timezone. */
static time_t
tz_offset (void)
{
  struct tm local;
  struct tm gm;
  time_t off;
  
  local = cur_tm;
  gm = *gmtime (&cur_time);

  off = ((local.tm_sec - gm.tm_sec)
	 + (local.tm_min - gm.tm_min) * 60
	 + (local.tm_hour - gm.tm_hour) * 3600);

  if (local.tm_wday != gm.tm_wday)
    {
      if (local.tm_wday == gm.tm_wday + 1)
	off += 86400;
      else if (local.tm_wday == gm.tm_wday - 1)
	off -= 86400;
      else if (local.tm_wday < gm.tm_wday)
	off += 86400;
      else
	off -= 86400;
    }

  return off;
}

/* Returns DATE rounded down to the closest midnight. */
static time_t
round_date_down (time_t date)
{
  return (date + tzoff) / 86400 * 86400 - tzoff;
}

/* Returns the time since the beginning of the day represented by
   WHEN. */
static time_t
time_of_day (time_t when)
{
  return (when + tzoff) % 86400;
}

/* Returns nonzero only if the day represented by B is later than the
   day represented by B. */
static int 
tm_ge_day (const struct tm *a, const struct tm *b)
{
  if (a->tm_year > b->tm_year)
    return 1;
  else if (a->tm_year < b->tm_year)
    return 0;

  if (a->tm_mon > b->tm_mon)
    return 1;
  else if (a->tm_mon < b->tm_mon)
    return 0;

  return a->tm_mday >= b->tm_mday;
}

/* Returns nonzero only if the month represented by B is later than
   the month represented by B. */
static int 
tm_ge_month (const struct tm *a, const struct tm *b)
{
  if (a->tm_year > b->tm_year)
    return 1;
  else if (a->tm_year < b->tm_year)
    return 0;

  return a->tm_mon >= b->tm_mon;
}

/* Returns a new date in a static buffer that represents one day
   before T. */
static struct tm *
prev_day (const struct tm *t)
{
  static struct tm tmp;

  tmp.tm_year = t->tm_year;
  tmp.tm_mon = t->tm_mon;
  tmp.tm_mday = t->tm_mday;

  tmp.tm_mday--;
  if (tmp.tm_mday == 0)
    {
      tmp.tm_mon--;
      if (tmp.tm_mon == -1)
	{
	  tmp.tm_year--;
	  tmp.tm_mon = 11;
	  tmp.tm_mday = 31;
	}
      else
	tmp.tm_mday = datetime_month_days (tmp.tm_year, tmp.tm_mon);
    }

  return &tmp;
}

/* If the year, month, or day in LAST are different from DATE, then
   display a title giving the date, and set LAST appropriately. */
static void 
catchup_title (const struct tm *date, struct tm *last)
{
  if (date->tm_year != last->tm_year
      || date->tm_mon != last->tm_mon
      || date->tm_mday != last->tm_mday)
    {
      html_list (0);
      html ("<P>%s</P>\n", datetime_format ("%d %b %Y", date));

      last->tm_year = date->tm_year;
      last->tm_mon = date->tm_mon;
      last->tm_mday = date->tm_mday;
    }
}

/* Advance T by one month.  If T is at the end of a particular month
   before advancement, it will be at the end of the next month as well. */
static void
inc_month (struct tm *t)
{
  int at_end = t->tm_mday == datetime_month_days (t->tm_year, t->tm_mon);
  
  t->tm_mon++;
  if (t->tm_mon > 11)
    {
      assert (t->tm_mon == 12);
      t->tm_mon = 0;
      t->tm_year++;
    }
  
  if (at_end)
    t->tm_mday = datetime_month_days (t->tm_year, t->tm_mon);
  else 
    {
      int n_days = datetime_month_days (t->tm_year, t->tm_mon);
      if (t->tm_mday > n_days)
	t->tm_mday = n_days;
    }
}

/* Back T up by one month.  If T is at the end of a particular month
   before advancement, it will be at the end of the next month as
   well. */
static void
dec_month (struct tm *t)
{
  int at_end = t->tm_mday == datetime_month_days (t->tm_year, t->tm_mon);
  
  t->tm_mon--;
  if (t->tm_mon < 0)
    {
      assert (t->tm_mon == -1);
      t->tm_mon = 11;
      t->tm_year--;
    }
  
  if (at_end)
    t->tm_mday = datetime_month_days (t->tm_year, t->tm_mon);
  else 
    {
      int n_days = datetime_month_days (t->tm_year, t->tm_mon);
      if (t->tm_mday > n_days)
	t->tm_mday = n_days;
    }
}

/* MS Project Import. */

static int do_msp_import (struct request *);
static struct msp_event *read_msp_import (void);
static int break_up (char *line, char **field, size_t n_fields);
static void import_error (char *msg, ...)
     attribute ((format (printf, 1, 2)));
     
static void 
show_import_msp (struct request *rq)
{
  if (cgi_file ("import", NULL))
    {
      if (do_msp_import (rq))
	return;
    }
    
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s</P>\n",
	_("Import MS Project Timeline"),
	_("To import a timeline from MS Project, you must first\n"
	  "save it in a tab-delimited format.  For instructions,\n"
	  "click on the <STRONG>Help</STRONG> link at the top of\n"
	  "this page.  Once you've done that, select the file\n"
	  "you saved below, set the options as you like, then\n"
	  "click on <STRONG>Import</STRONG>."));

  html_start_form ("calendar", HT_POST, HT_FORM_DATA);
  html_hidden_abbrev (rq);
  html_hidden ("op", "msp");
  html ("<P>%s:</P>\n", _("Select <EM>tab-delimited</EM> file to import"));
  html_file_field ("import");
  htmls ("</P>\n");

  html ("<P>%s:<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=type VALUE=keep CHECKED>%s</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=type VALUE=drop>%s</LABEL>\n"
	"</P>",
	_("Treatment of existing MS Project imported events"),
	_("<STRONG>Keep them.</STRONG> "
	  "Newly imported events are added to the existing imported events."),
	_("<STRONG>Drop them.</STRONG> "
	  "Newly imported events replace existing imported events."));

  html_submit (_("Import"));
  html_end_form ();  
}

struct msp_event 
  {
    struct msp_event *next;
    char *name;
    time_t beg;
    time_t end;
  };

static int
do_msp_import (struct request *rq)
{
  struct msp_event *i, *next;
  time_t min, max;
  int drop;

  i = read_msp_import ();
  if (i == NULL)
    return 0;

  db_begin ();
  drop = !strcmp (std_get (rq->cgi, "type"), "drop");
  if (drop)
    db_query ("DELETE FROM events WHERE grp = '%sq' AND source = 1",
	      DB_FORMAT, NULL, rq->cx.grp);

  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s.  %s</P>\n\n"
	"<CENTER>\n"
	"<TABLE>\n"
	"<TR><TH>%s<TH>%s<TH>%s\n",
	_("Import Results"),
	_("The following events were imported into the calendar."),
	(drop
	 ? _("Any existing imported events were dropped.")
	 : _("Any existing imported events were retained.")),
	_("Event"),
	_("Begins"),
	_("Ends"));

  min = i->beg;
  max = i->end;
  for (; i != NULL; i = next)
    {
      char b[64], e[64];

      if (i->beg < min)
	min = i->beg;
      if (i->end > max)
	max = i->end;
      
      strftime (b, sizeof b, "%b %d %Y %H:%M", localtime (&i->beg));
      strftime (e, sizeof e, "%b %d %Y %H:%M", localtime (&i->end));

      db_query ("INSERT INTO events VALUES "
		"('%sq', '%sq', %d, '', '%sq', '%sq', nextval ('seq'), 1)",
		DB_FORMAT, NULL,
		b, e, get_file_seq (&rq->cx, i->name),
		rq->cx.usr, rq->cx.grp);

      htmls ("<TR><TD>");
      html_escape (i->name);
      htmls ("<TD>");
      html_escape (b);
      htmls ("<TD>");
      html_escape (e);
      htmlc ('\n');

      next = i->next;
      free (i->name);
      free (i);
    }

  htmls ("</TABLE>\n"
	 "</CENTER>\n");
  
  {
    char b[64], e[64];
    
    strftime (b, sizeof b, "%b %d %Y %H:%M", localtime (&min));
    strftime (e, sizeof e, "%b %d %Y %H:%M", localtime (&max));
    
    html_button ("calendar", _("View Calendar for Range of Imported Data"),
		 "begin", b, "end", e, (char *) NULL);
  }

  db_end ();

  return 1;
}

static struct msp_event *
read_msp_import (void)
{
  struct msp_event *head = NULL;
  struct msp_event *tail = NULL;
  
  char *line = NULL;
  size_t size = 0;
  int ln = 0;
  
  const char *fn;
  FILE *f = cgi_file ("import", &fn);
  assert (f != NULL);

  for (;;)
    {
      char *fields[7];
      const int n_fields = sizeof fields / sizeof *fields;

      time_t beg, end;
      struct msp_event *new;
      
      long len = getline (&line, &size, f);
      if (len == -1)
	break;
      while (len > 0 && (line[len - 1] == '\n' || line[len - 1] == '\r'))
	line[--len] = '\0';
      ln++;

      if (!break_up (line, fields, n_fields))
	{
	  import_error (_("%s:%d: Line does not contain %d "
			  "tab-delimited fields."),
			fn, ln, n_fields);
	  continue;
	}

      /* First line in file is column titles. */
      if (ln == 1)
	continue;

      beg = datetime_parse (fields[3], DT_DATE_REQUIRED, 0, NULL);
      if (beg == 0)
	{
	  import_error (_("%s:%d: Could not parse \"%s\" as a date: %s"),
			fn, ln, fields[3], datetime_error);
	  continue;
	}
      
      end = datetime_parse (fields[4], DT_DATE_REQUIRED, 0, NULL);
      if (end == 0)
	{
	  import_error (_("%s:%d: Could not parse \"%s\" as a date: %s"),
			fn, ln, fields[4], datetime_error);
	  continue;
	}
      if (end == round_date_down (end))
	end += 86400;

      new = xmalloc (sizeof *new);
      new->next = NULL;
      new->name = xstrdup (fields[1]);
      new->beg = beg;
      new->end = end;
      if (head == NULL)
	head = new;
      else
	tail->next = new;
      tail = new;
    }

  free (line);
  return head;
}

static int 
break_up (char *line, char **field, size_t n_fields)
{
  for (; n_fields; n_fields--, field++)
    {
      *field = strsep (&line, "\t");
      if (*field == NULL)
	return 0;
    }

  return strsep (&line, "\t") == NULL;
}

static void 
import_error (char *msg, ...)
{
  va_list args;
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s:\n",
	_("Error in Import"),
	_("The following error was encountered while processing the\n"
	  "file for import"));

  va_start (args, msg);
  vprintf (msg, args);
  va_end (args);

  html ("\n"
	"%s</P>\n",
	_("Please check that the submitted file was properly exported from\n"
	  "MS Project.  For instructions, click on the <STRONG>Help</STRONG>\n"
	  "link at the top of this page."));
}
