/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include "alloc.h"
#include "cgi.h"
#include "error.h"
#include "str.h"
#include "std.h"

/* This module has been debugged pretty well by now. */
/*#undef DEBUGGING*/

/* 0=CGI module has not yet been initialized; !0=CGI module is
   ready. */
static int inited;

/* A file upload CGI parameter. */
struct file
  {
    struct file *next;
    
    char *name;
    FILE *file;
    char *filename;
  };

/* All the file upload parameters. */
static struct file *file_list;

/* Public functions. */

static void parse_POST_query (struct std *);
static void parse_multipart (struct std *, const char *content_type,
			     size_t content_length unused);
static void parse_offline_query (struct std *);
static void parse_query_string (struct std *, const char *query);

#if DEBUGGING
static void debug_print (struct std *);
#endif

/* Initialize the CGI module.  This causes the CGI parameters to be
   read from the environment or stdin (interactively or
   noninteractively).  This function must be called before any other
   function in this module, and before cgi is accessed. */
struct std * 
cgi_initialize (void)
{
  struct std *cgi;
  
  assert (!inited);
  inited = 1;

  cgi = std_create (1);

  {
    const char *request_method = getenv ("REQUEST_METHOD");
    if (request_method != NULL)
      {
	if (!strcmp ("GET", request_method)
	    || !strcmp ("HEAD", request_method))
	  {
	    const char *query = getenv ("QUERY_STRING");
	    if (query != NULL)
	      parse_query_string (cgi, query);
	  }
	else if (!strcmp ("POST", request_method))
	  parse_POST_query (cgi);
	else
	  fail (("unknown request method %s"), request_method);
      }
    else
      parse_offline_query (cgi);
  }

#if DEBUGGING
  debug_print (cgi);
#endif

  return cgi;
}

/* Looks for a CGI file upload field named NAME.  If found, sets
   *FILENAME to the suggested filename (if FILENAME is non-null) and
   *returns the associated file; otherwise, returns a null pointer. */
FILE *
cgi_file (const char *name, const char **const filename)
{
  struct file *iter;

  for (iter = file_list; iter; iter = iter->next)
    if (!strcmp (iter->name, name))
      {
	if (filename != NULL)
	  *filename = iter->filename;
	return iter->file;
      }
  return NULL;
}

/* Private functions. */

/* Return the value of hex digit CH,
   or -1 if it's not a hex digit. */
static int
xdigit_value (int ch)
{
  static const char letters[] = "0123456789abcdef";
  const char *p = strchr (letters, tolower ((unsigned char) ch));
  return p != NULL ? p - letters : -1;
}

/* Parse http query string QUERY, as specified by the CGI spec at
   <URL:http://www.princeton.edu/~lieske/www/cgi-spec.html> and the
   HTTP 1.1 specification (RFC 2068).  Puts all retrieved query data
   into cgi. */
static void
parse_query_string (struct std *cgi, const char *query)
{
  struct st key, value;
  struct st *cur;
  const char *s;

  st_init (&key, 16);
  st_init (&value, 64);

  cur = &key;
  for (s = query; ; s++)
    switch (*s)
      {
      case '=':
	cur = &value;
	break;

      case '&':
      case '\0':
	if (st_length (&key))
	  {
	    if (st_length (&value) == 0 && strchr (st_value (&key), ','))
	      std_put (cgi, "loc", st_value (&key));
	    else
	      std_put (cgi, st_value (&key), st_value (&value));
	  }
	if (*s == '\0')
	  {
	    st_destroy (&key);
	    st_destroy (&value);
	    
	    return;
	  }

	st_clear (&key);
	st_clear (&value);
	cur = &key;

	break;

      case '+':
	st_putchar (cur, ' ');
	break;

      case '%':
	{
	  int a = xdigit_value (s[1]);
	  if (a != -1)
	    {
	      int b = xdigit_value (s[2]);
	    
	      if (b != -1)
		{
		  st_putchar (cur, a * 16 + b);
		  s += 2;
		  break;
		}
	    }
	}
	/* Fall through. */

      default:
	st_putchar (cur, (unsigned char) *s);
	break;
      }
}

/* Reads a POST type query from stdin and parses it into
   cgi. */
static void
parse_POST_query (struct std *cgi)
{
  long content_length;
  const char *content_type = getenv ("CONTENT_TYPE");

  if (getenv ("CONTENT_LENGTH") == NULL)
    {
      log ("POST query missing CONTENT_LENGTH");
      return;
    }
  content_length = strtol (getenv ("CONTENT_LENGTH"), NULL, 10);
  if (content_length < 0)
    {
      log ("POST query has negative CONTENT_LENGTH");
      return;
    }
  
  if (content_type == NULL)
    content_type = "";

  if (!strncasecmp (content_type,
		    "multipart/form-data", strlen ("multipart/form-data")))
    parse_multipart (cgi, content_type, content_length);
  else
    {
      char *buf = xmalloc (content_length + 1);
      if (fread (buf, content_length, 1, stdin) != 1)
	fail ("error reading POST data from stdin: %s", strerror (errno));
      buf[content_length] = '\0';
      
      parse_query_string (cgi, buf);

      free (buf);
    }
}

/* Support for multipart/form-data MIME format as specified by RFC
   2388. */

/* The data stream for multipart/form-data is difficult to parse in a
   fast, efficient, robust manner.  The biggest problem is that it
   mixes line-oriented text with arbitrary 8-bit binary data in an
   ugly way.

   There are several inelegant ways that I thought of using to handle
   the problems associated with this, but in the end I settled on
   using plain line-oriented input.  If it turns out that there's an
   inportant number of large files out there that have a very low
   frequency of newlines, then maybe it'll be time to rethink this.

   At one point my primary concern with this routine was
   memory-consumption DoS attacks, but since local users can do far
   worse things, I'm no longer so worried. */
   
static void parse_boundary (struct st *boundary, const char *content_type);
static void parse_content_disposition (const char *line, struct st *name,
				       struct st *filename);
static int write_line_without_CRLF (FILE *file, struct st *line);
static void read_error (FILE *, const char *filename)
     attribute ((noreturn));
static void add_file (const char *name, FILE *file, const char *filename);

/* Parse multipart/form-data script arguments. */
static void
parse_multipart (struct std *cgi,
		 const char *content_type, size_t content_length unused)
{
  struct st boundary;	/* The boundary between MIME parts. */
  struct st name;	/* The name of the part. */
  struct st line;	/* Line buffer. */
  struct st content;	/* Content of a non-file part. */
  struct st filename;	/* File name associated with this part. */
  FILE *file;		/* File associated with this part. */
  int state;		/* State machine state. */
  int eof;		/* At end of input? */

  st_init (&boundary, 64);
  st_init (&name, 64);
  st_init (&line, 64);
  st_init (&content, 64);
  st_init (&filename, 64);
  file = NULL;
  state = 0;
  eof = 0;

  parse_boundary (&boundary, content_type);

  do
    {
      if (!st_getline (&line, stdin))
	read_error (stdin, "stdin");

      state &= 0xf;
      if (!strncmp (st_value (&line), st_value (&boundary),
		    st_length (&boundary)))
	{
	  const char *line_str = st_value (&line);
	  size_t line_len = st_length (&line);
	  size_t boundary_len = st_length (&boundary);
	  
	  /* According to RFC 2046 section 5.1.1 the boundary string
	     need not be followed directly by a CRLF to qualify as a
	     boundary. */
	  state |= 0x10;
	  if (line_str[boundary_len] == '-'
	      && line_str[boundary_len + 1] == '-'
	      && (strspn (&line_str[boundary_len + 2], " \t\r\n")
		  == line_len - boundary_len - 2))
	    eof = 1;
	}
      else if (st_value (&line)[st_length (&line) - 1] != '\n')
	{
	  state |= 0x10;
	  eof = 1;
	}

      switch (state)
	{
	  /* We are in the preamble.  Ignore everything until we see a
	     boundary. */
	case 0x00:
	  break;

	  /* We are in the preamble, but we just saw the boundary.
	     Now transition to state 1 to start looking at headers. */
	case 0x10:
	  state = 1;
	  break;
	  
	  /* We are in the headers of one of the parts.
	     Look for interesting headers or the end-of-headers marker. */
	case 0x01:
	  if (!strncasecmp (st_value (&line), "Content-Disposition:",
			    strlen ("Content-Disposition:")))
	    parse_content_disposition (st_value (&line), &name, &filename);
	  else if (st_value (&line)[0] == '\r' && st_value (&line)[1] == '\n')
	    {
	      if (st_length (&name) > 0)
		{
		  if (st_length (&filename) == 0)
		    state = 2;
		  else
		    {
		      state = 3;
		      file = tmpfile ();
		      if (file == NULL)
			fail ("error creating temporary file: %s",
			      strerror (errno));
		    }
		}
	      else
		state = 0;
	    }
	  break;
	  
	  /* We've encountered a boundary in the middle of headers.
	     Start over. */
	case 0x11:
	  st_clear (&name);
	  st_clear (&filename);
	  break;

	  /* Read content into a string buffer.  */
	case 0x02:
	  st_concat_buffer (&content, st_value (&line), st_length (&line));
	  break;

	  /* End of content.  Strip CRLF from end of content and save
             it. */
	case 0x12:
	  if (st_last_char (&content) == '\n')
	    st_trim (&content, 1);
	  if (st_last_char (&content) == '\r')
	    st_trim (&content, 1);
	  std_put (cgi, st_value (&name), st_value (&content));
	  st_clear (&name);
	  st_clear (&content);

	  state = 1;
	  break;

	  /* This is the first line of content in a file. */
	case 0x03:
	  state = write_line_without_CRLF (file, &line);
	  break;

	  /* This is another line of content from the file.  The
	     previous line ended with \r\n. */
	case 0x04:
	  if (EOF == putc ('\r', file)
	      || EOF == putc ('\n', file))
	    fail ("error writing temporary file: %s", strerror (errno));
	  state = write_line_without_CRLF (file, &line);
	  break;
	  
	  /* This is another line of content from the file.  The
	     previous line ended with \n. */
	case 0x05:
	  if (EOF == putc ('\n', file))
	    fail ("error writing temporary file: %s", strerror (errno));
	  state = write_line_without_CRLF (file, &line);
	  break;

	  /* We've encountered a boundary while reading file contents.
	     Add the file to the list and start over. */
	case 0x13:
	case 0x14:
	case 0x15:
	  add_file (st_value (&name), file, st_value (&filename));
	  st_clear (&name);
	  st_clear (&filename);
	  state = 1;
	  break;

	default:
	  assert (0);
	}
    }
  while (!eof);

  st_destroy (&boundary);
  st_destroy (&name);
  st_destroy (&line);
  st_destroy (&content);
  st_destroy (&filename);
}

/* Writes LINE of length LINE_LEN to FILE.  The end-of-line sequence
   (either LF or CRLF) is not written.  Returns 4 if the end-of-line
   sequence was CRLF or 5 if it was LF only. */
static int
write_line_without_CRLF (FILE *file, struct st *line)
{
  int next_state;

  assert (st_last_char (line) == '\n');
  st_trim (line, 1);
  
  if (st_last_char (line) == '\r')
    {
      st_trim (line, 1);
      next_state = 4;
    }
  else
    next_state = 5;

  if (fwrite (st_value (line), 1, st_length (line), file) != st_length (line))
    fail ("error writing temporary file: %s", strerror (errno));

  return next_state;
}

/* Attempt to parse the Content-Type: header as supplied in
   CONTENT_TYPE to determine the MIME multipart boundary, which is
   stored into BOUNDARY. */
static void
parse_boundary (struct st *boundary, const char *content_type)
{
  /* See RFC 2045 for a description of the format of the Content-Type:
     header.  See RFC 822 for a description of quoted-string. */

  /* FIXME: This is *not* fully RFC 2045 compliant.  However, I
     suspect that no one else is, either, so it's not a big deal. */

  const char *boundary_beg = blp_strcasestr (content_type, "boundary=");
  if (boundary_beg != NULL)
    {
      boundary_beg += strlen ("boundary=");
      if (*boundary_beg == '"')
	boundary_beg++;

      st_concat (boundary, "--");
      st_concat_buffer (boundary, boundary_beg,
			strcspn (boundary_beg, " \t\r\n\""));
    }
  else
    fail ("missing boundary= for multipart/form-data");
}

/* Parses an RFC 2388 Content-Disposition: line specified in LINE.  If
   a name= property is given in LINE, *NAME is set appropriately to a
   malloc()'d string representing its value; similarly for filename=
   and *FILENAME. */
static void
parse_content_disposition (const char *line, struct st *name,
			   struct st *filename)
{
  /* This is not even close to being RFC 2045 compliant.  However, it
     works with Netscape-generated multipart/form-data, which is
     important.  FIXME: Detect Netscape and handle it differently from
     correct browers. */
  
  {
    const char *name_beg = blp_strcasestr (line, "name=");
    if (name_beg != NULL)
      {
	size_t name_len;

	name_beg += strlen ("name=");
	if (*name_beg == '"')
	  {
	    name_beg++;
	    name_len = strcspn (name_beg, "\"");
	  }
	else
	  name_len = strcspn (name_beg, "\r\n");

	if (st_length (name) > 0)
	  {
	    log ("duplicate name= headers");
	    st_clear (name);
	  }
	st_concat_buffer (name, name_beg, name_len);
      }
  }
  
  {
    const char *filename_beg = blp_strcasestr (line, "filename=\"");
    if (filename_beg != NULL)
      {
	size_t filename_len;

	filename_beg += strlen ("filename=\"");
	filename_len = strrchr (line, '"') - filename_beg;
	if (filename_len > 0)
	  {
	    if (st_length (filename) > 0)
	      {
		log ("duplicate filename= headers");
		st_clear (filename);
	      }
	    st_concat_buffer (filename, filename_beg, filename_len);
	  }
      }
  }
}

/* Adds FILE with its corresponding Content-Disposition: name= and
   filename= fields NAME and FILENAME to the list of files. */
static void
add_file (const char *name, FILE *file, const char *filename)
{
  struct file *f = xmalloc (sizeof *f);
  f->next = file_list;
  f->name = xstrdup (name);
  f->file = file;
  f->filename = xstrdup (filename);
  file_list = f;

  if (EOF == fflush (file))
    fail ("can't flush temporary file: %s", strerror (errno));
  if (fseek (file, 0, SEEK_SET) != 0)
    fail ("can't rewind temporary file: %s", strerror (errno));
}

/* Report a fatal error reading from STREAM having name FILENAME. */
void
read_error (FILE *stream, const char *filename)
{
  if (ferror (stream))
    fail ("error reading %s: %s", filename, strerror (errno));
  else
    fail ("unexpected end of file reading %s", filename);
}

/* Parse an `offline query' where the user runs the CGI script
   interactively from the command line.  */
static void
parse_offline_query (struct std *cgi)
{
  struct st line;
  
  printf (_("CGI offline execution detected.\n"
	    "Enter parameters below in application/x-www-url-encoded form.\n"
	    "End the parameter list by typing Ctrl+D.\n\n"));

  st_init (&line, 64);
  for (;;)
    {
      if (!st_getline (&line, stdin))
	break;
      if (st_last_char (&line) == '\n')
	st_trim (&line, 1);
      
      parse_query_string (cgi, st_value (&line));
    }
  st_destroy (&line);
}

/* Debug code. */

#if DEBUGGING
/* Print out the values of the CGI parameters and information about
   the CGI files. */
static void
debug_print (struct std *cgi)
{
  {
    struct std_iterator iter;
    const char *key;
    char *value;

    fprintf (stderr, "CGI parameters:\n");
    std_init_iterator (cgi, &iter);
    while (std_iterate (&iter, &key, &value))
      {
	if (strcmp (key, "password"))
	  fprintf (stderr, "\t\"%s\" = \"%s\"\n", key, value);
	else
	  fprintf (stderr, "\t\"%s\" = <censored>\n", key);
      }
  }

  if (file_list != NULL)
    {
      struct file *f;
    
      fprintf (stderr, "CGI files:\n");
      for (f = file_list; f; f = f->next)
	fprintf (stderr, "\t\"%s\": filename=\"%s\"\n", f->name, f->filename);
    }
}
#endif /* DEBUGGING */
      
