/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <sys/wait.h>
#include "cgi.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "help.h"
#include "html.h"
#include "identity.h"
#include "scope.cgi.h"
#include "session.h"
#include "std.h"
#include "str.h"

static void dump_file (const char *filename);
static void toc_header (void);

void 
scr_help (struct request *rq)
{
  struct st filename;
  st_init (&filename, 64);
  st_printf (&filename, "%s/", HELP_DIR);
  if (strlen (rq->screen) > 5)
    {
      const char *cp = rq->screen + 5;
      while (isalnum ((unsigned char) *cp) || *cp == '_' || *cp == '-')
	{
	  st_putchar (&filename, *cp);
	  cp++;
	}

      if (!strcmp (rq->screen + 5, "toc"))
	toc_header ();
    }
  else
    {
      st_concat (&filename, "toc");
      toc_header ();
    }
  
  dump_file (st_value (&filename));
  st_destroy (&filename);
}

static void
dump_file (const char *filename)
{
  FILE *f = fopen (filename, "r");
  if (f == NULL)
    {
      html ("<H2 ALIGN=CENTER>%s: %s</H2>"
	    "<P>%s: %s.",
	    _("Error accessing help file"),
	    strerror (errno),
	    _("TeamSCOPE encountered an error accessing the following\n"
	      "help file"),
	    filename);
      html (_("If you think that this is a bug in TeamSCOPE, report\n"
	      "it to the system administrator, %s.\n"
	      "Thanks for using TeamSCOPE.\n"),
	    admin_email);
      htmls ("</P>\n");
      return;
    }
  verify_file_owner (filename, f, 0);
  
  for (;;)
    {
      char buffer[4096];

      if (fwrite (buffer, 1, fread (buffer, 1, 4096, f), stdout) != 4096)
	{
	  fclose (f);
	  return;
	}
    }
}

static void
toc_header (void)
{
  /* Ugly kluge. */
  html ("<H2 ALIGN=CENTER>\n"
	"[ <A HREF=\"%s://%s%s/doc\">%s</A> ]\n"
	"</H2>\n",
	server_protocol,
	std_get (conf, "server_name"),
	std_get (conf, "www_root"),
	_("Other TeamSCOPE Manuals | Alternate Formats"));
}

static void send_feedback (struct request *);

void
scr_feedback (struct request *rq)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n",
	_("User Feedback"));

  if (std_get_null (rq->cgi, "title") != NULL)
    {
      send_feedback (rq);
      return;
    }

  html ("<P>%s</P>\n",
	_("Use this form to report bugs and suggest new features\n"
	  "for TeamSCOPE.  Once you've filled it out, click on\n"
	  "<STRONG>Submit</STRONG> to have it sent directly to\n"
	  "the TeamSCOPE development team.\n"
	  "If you prefer, you can mail the developers directly at\n"
	  "<A HREF=\"mailto:scope@cscw.msu.edu\">scope@cscw.msu.edu</A>.\n"));

  html_start_form ("feedback", HT_POST, HT_URL_ENCODED);
  html_hidden ("user", rq->cx.usr);
  html_hidden ("team", rq->cx.grp);
  html_hidden ("fscreen", std_get (rq->cgi, "fscreen"));
  html_hidden ("query", std_get (rq->cgi, "fquery"));
  html_hidden ("server", server_name);
  html_hidden ("admin", admin_email);
  html_hidden ("mail-domain", mail_domain);

  html ("<TABLE>\n"
	"  <TR VALIGN=TOP>"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n"
	"      ",
	_("Subject"),
	_("Enter a descriptive title for your bug report here."
	  "``TeamSCOPE bug'' and ``bug report'' are "
	  "<EM>not</EM> descriptive titles."));
  
  html_text_field ("title", "", 55);

  html ("    </TD>\n"
	"  </TR>\n"
	"  <TR VALIGN=TOP>\n"
	"    <TH>%s</TH>\n"
	"    <TD>%s<BR>\n"
	"      ",
	_("Description"),
	_("Enter a full, detailed description of the bug encountered\n"
	  "or feature request here.\n"
	  "``I can't log in'' and ``The activity screen doesn't work''\n"
	  "are <EM>not</EM> full, detailed descriptions; such bug\n"
	  "reports will be ignored."));

  html_text_area ("description", "", 20, 55);

  htmls ("    </TD>\n"
	 "  </TR>\n"
	 "</TABLE>\n");

  html_submit (_("Submit"));

  html_end_form ();
}

static void
send_feedback (struct request *rq)
{
  FILE *f;
  
  {
    struct st s;
    const char *cp;

    st_create (&s, "mail -s '");
    for (cp = std_get (rq->cgi, "title"); *cp; cp++)
      if (strchr ("abcdefghijklmnopqrstuvwxyz"
		  "ABCDEFGHIJKLMNOPRSTUVWXYZ"
		  " 	:.,/?!@%()", (unsigned char) *cp))
	st_putchar (&s, (unsigned char) *cp);
    st_concat (&s, "' scope@cscw.msu.edu");

    f = popen (st_value (&s), "w");
    if (f == NULL)
      goto error;

    st_destroy (&s);
  }
  
  {
    struct std_iterator i;

    std_init_iterator (rq->cgi, &i);
    for (;;)
      {
	const char *key;
	char *value;

	if (!std_iterate (&i, &key, &value))
	  break;

	fprintf (f, "%s:\n%s\n\n", key, value ? value : "<no value>");
      }
  }
  
  {
    int status = pclose (f);
    if (!WIFEXITED (status) || WEXITSTATUS (status))
      goto error;
  }

  html ("<P>%s</P>\n",
	_("Your comments have been passed on to the TeamSCOPE development\n"
	  "team.  You should receive a response within the next few days.\n"
	  "Thanks for helping to improve TeamSCOPE."));
  
  return;

 error:
  html ("<P>%s</P>\n",
	_("An error was encountered by the server while attempting to send\n"
	  "your comments to the TeamSCOPE development team.  You can still\n"
	  "send your feedback by writing directly to the team at\n"
	  "<A HREF=\"mailto:scope@cscw.msu.edu\">scope@cscw.msu.edu</A>.\n"));
}

void
scr_welcome (struct request *rq)
{
  db_result *r;

  if (std_contains (rq->cgi, "enable"))
    db_query ("UPDATE users SET welcome = '%sq' WHERE usr = '%sq'",
	      DB_FORMAT, NULL,
	      strcmp (std_get (rq->cgi, "enable"), "yes") ? "f" : "t",
	      rq->cx.usr);

  db_query ("SELECT name, email, recordable FROM users "
	    "WHERE usr = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("Welcome to TeamSCOPE!"));
  html ("<P>%s</P>\n\n",
	_("Welcome to TeamSCOPE, a system to facilitate asynchronous\n"
	  "collaborative activities among geographically distributed\n"
	  "teams.  To start using the system at once, pick one of the\n"
	  "options from the menu above, then click on the\n"
	  "<STRONG>Go!</STRONG> button, if necessary.  Alternatively, keep\n"
	  "reading for more information on what TeamSCOPE is and\n"
	  "what it can do for you.\n"));

  html ("<P>%s</P>\n\n",
	_("Here are some starting points for learning and using\n"
	  "TeamSCOPE.  You can explore by clicking on them.\n"
	  "If you want to come back here, just select\n"
	  "<STRONG>Welcome</STRONG> from the drop-down menu at the top\n"
	  "of the page, then click on the <STRONG>Go!</STRONG> button,\n"
	  "if necessary.\n"));

  htmls ("<CENTER>\n");
  if (show_welcome (&rq->cx))
    html_button ("welcome", _("Don't show this page again at next login"),
		 "enable", "no", (char *) NULL);
  else
    html_button ("welcome", _("Show this page at next login"),
		 "enable", "yes", (char *) NULL);
  htmls ("</CENTER>\n");

  htmls ("<UL>\n");

  html ("<LI><P>%s <A HREF=\"%s\" TARGET=\"_blank\">%s</A>.\n"
	"<BR>%s</P>\n",
	_("Go through the"),
	url_this ("help-tutorial_activity", (char *) NULL),
	_("TeamSCOPE tutorial"),
	_("This link will bring up a new browser window,\n"
	  "so that you can conveniently view\n"
	  "TeamSCOPE and the tutorial side by side."));

  html ("<LI><P>%s <A HREF=\"%s\">%s</A>.\n"
	"<BR>%s</P>\n",
	_("Change your"),
	url_this ("password", (char *) NULL),
	_("password"),
	_("Chances are that you were automatically assigned a random,\n"
	  "unmemorable password.  With this page, you can change it to\n"
	  "something more familiar.\n"));

  html ("<LI><P>%s <A HREF=\"%s\">%s</A>.\n"
	"<BR>%s</P>\n",
	_("Visit the"),	
	url_this ("options", (char *) NULL),
	_("Options screen"),
	_("You can use this screen to set your\n"
	  "timezone, email preferences, and more.\n"));

  html ("<LI><P>%s <A HREF=\"%s\">%s</A>.\n"
	"<BR>",
	_("Edit your"),
	url_this ("admin", "op", "editusr", "usr", rq->cx.usr, (char *) NULL),
	_("user profile"));

  if (strcmp (std_get (conf, "CC"), ""))
    html (_("This allows you to set your full name, email address, or\n"
	    "%s recording of your TeamSCOPE email.\n"),
	  dbr_value (r, 0, 2)[0] == 't' ? "disable" : "enable");
  else
    htmls (_("This allows you to set your full name or email address.\n"));

  html ("%s <STRONG>%s</STRONG>\n",
	_("You probably don't want to change these settings unless\n"
	  "they are incorrect.  Currently your full name is set as\n"),
	dbr_value (r, 0, 0));

  if (strcmp (dbr_value (r, 0, 1), ""))
    html ("%s <STRONG>%s</STRONG>",
	  _("and your email address as\n"), dbr_value (r, 0, 1));
  else
    htmls (_("and your email is set for storage in a local mailbox"));
  htmls (".</P>\n");

  html ("<LI><P>%s <A HREF=\"%s\">%s</A>.\n"
	"<BR>%s</P>\n",
	_("Look at an"),
	url_this ("help-tutorial_overview", (char *) NULL),
	_("overview of TeamSCOPE features"),
	_("This will give you a better idea of what\n"
	  "TeamSCOPE is designed to do.\n"));

  html ("<LI><P>%s <A HREF=\"%s\">%s</A>.\n"
	"<BR>%s: <A HREF=\"%s://%s%s/doc\">%s</A>.\n",
	_("View the"),
	url_this ("help", (char *) NULL),
	_("TeamSCOPE user manual"),
	_("This manual covers using TeamSCOPE in detail.\n"
	  "Also available"),
	server_protocol,
	std_get (conf, "server_name"), std_get (conf, "www_root"),
	_("TeamSCOPE installation manual and internals manual"));

  htmls ("</UL>\n");

  dbr_destroy (r);
}

int
show_welcome (struct context *cx)
{
  db_result *r;
  int show;

  db_query ("SELECT welcome FROM users WHERE usr = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, cx->usr);
  show = dbr_value (r, 0, 0)[0] == 't';
  dbr_destroy (r);

  return show;
}
  
