/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "alloc.h"
#include "conf.h"
#include "html.h"
#include "identity.h"
#include "scope.cgi.h"
#include "session.h"
#include "std.h"

#if GLOBAL_DEBUGGING
#include <time.h>
#include <sys/time.h>
clock_t start_clock;
struct timeval start_timeval;
#endif

/* Various stuff about ourselves. */
const char *server_protocol;	/* http */
const char *server_name;	/* cscw.msu.edu */
const char *script_name;	/* /scope/scope.cgi */
int server_port;		/* 80 */

const char *www_root;
const char *mail_domain;
const char *admin_email;

const char *our_abs_url;	/* http://cscw.msu.edu */

/* Public functions. */

/* Initialize the HTML module. */
void
html_initialize (void)
{
  assert (conf != NULL);

  server_protocol = std_get_assert (conf, "server_protocol");
  server_name = std_get_assert (conf, "server_name");
  server_port = stoi (std_get_assert (conf, "server_port"));
  script_name = std_get_assert (conf, "script_name");
  www_root = std_get_assert (conf, "www_root");
  mail_domain = std_get_assert (conf, "mail_domain");
  admin_email = std_get_assert (conf, "admin_email");
  
  if (server_port == 80)
    our_abs_url = blp_asprintf ("%s://%s%s",
				server_protocol, server_name, script_name);
  else
    our_abs_url = blp_asprintf ("%s://%s:%d%s",
				server_protocol, server_name, server_port,
				script_name);
}

/* Write a standard HTML header to the HTML output stream. */
void
html_header (void)
{
  html
    ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\n"
     "        \"http://www.w3.org/TR/REC-html40/loose.dtd\">\n\n"
     "<HTML>\n"
     "<HEAD>\n"
     "<TITLE>%s</TITLE>\n"
     "<BASE HREF=\"%s\">\n"
     "<META NAME=\"generator\" CONTENT=\"TeamSCOPE\">\n"
     "<META NAME=\"teamscope-version\" CONTENT=\"0.9\">\n"
     "</HEAD>\n\n"
     "<BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\" LINK=\"#1F00FF\"\n"
     "      ALINK=\"#FF0000\" VLINK=\"#1F00FF\">\n",
     _("TeamSCOPE: Team Software for Collaborative Project Environments"),
     our_abs_url);
}


/* Write S to the HTML output stream, escaping all the possibly
   problematic characters as HTML &entities; */
void
html_escape (const char *s)
{
  for (; ; s++)
    switch (*s)
      {
      case 0:
	return;
	
      case '<':
	htmls ("&lt;");
	break;
	
      case '>':
	htmls ("&gt;");
	break;

      case '&':
	htmls ("&amp;");
	break;

      case '"':
	htmls ("&quot;");
	break;

      default:
	if (isprint ((unsigned char) *s))
	  htmlc (*s);
	else
	  html ("&#%d;", (unsigned char) *s);
	break;
      }
}

/* Write S to the HTML output stream, escaping all the possibly
   problematic characters as HTML &entities; and translating special
   <URL:www.example.com> syntax into a real HTML <A> tag. */
void
html_pretty_print (const char *s)
{
  size_t ll = 0;
  
  for (; ; s++)
    switch (*s)
      {
      case 0:
	return;
	
      case '<':
	if (!strncasecmp (s, "<URL:", 5))
	  {
	    const char *end = strchr (s, '>');
	    
	    if (end != NULL)
	      {
		struct st url;
		size_t literal_ofs;

		s += 5;
		st_init (&url, end - s + 8);

		if (!memchr (s, ':', end - s))
		  {
		    if (memchr (s, '@', end - s))
		      st_concat (&url, "mailto:");
		    else
		      st_concat (&url, "http://");
		  }
		literal_ofs = st_length (&url);
		st_concat_buffer (&url, s, end - s);
		  
		htmls ("<A HREF=\"");
		html_escape (st_value (&url));
		htmls ("\">");
		html_escape (st_value (&url) + literal_ofs);
		htmls ("</A>");

		ll += 9 + 2 + 4 + 2 * st_length (&url) - literal_ofs;

		st_destroy (&url);
		s = end;
	      }
	  }
	else
	  {
	    htmls ("&lt;");
	    ll += 4;
	  }
	break;
	
      case '>':
	htmls ("&gt;");
	ll += 4;
	break;

      case '&':
	htmls ("&amp;");
	ll += 5;
	break;

      case '"':
	htmls ("&quot;");
	ll += 6;
	break;

      case '\n':
	htmls ("<P>\n");
	ll = 0;
	break;

      case '\r':
      case ' ':
      case '\t':
      case '\v':
	if (ll <= 65)
	  {
	    htmlc (' ');
	    ll++;
	  }
	else
	  {
	    htmlc ('\n');
	    ll = 0;
	  }
	break;

      default:
	if (isprint ((unsigned char) *s))
	  {
	    htmlc (*s);
	    ll++;
	  }
	else
	  {
	    html ("&#%d;", (unsigned char) *s);
	    ll += 6;
	  }
	break;
      }
}

/* Introduce a <UL> list at nesting level LEVEL.  Keeps tracks of the
   nesting level last referenced and outputs the proper number of <UL>
   and </UL> tags. */
void
html_list (int level)
{
  static int list_level = 0;

  if (level != list_level)
    {
      if (level > list_level)
	{
	  htmls ("<UL>");
	  list_level++;

	  while (level > list_level)
	    {
	      htmls ("<LI><UL>");
	      list_level++;
	    }
	}
      while (level < list_level)
	{
	  htmls ("</UL>");
	  list_level--;
	}
      htmlc ('\n');
    }
  
  if (level != 0)
    htmls ("<LI>");
}

/* Write FILENAME to the HTML stream.  If FILENAME is an actual file,
   write a file manager link to it, too.  If non-null, URL is used as
   a base URL for the link; it should have a `screen' of "files". */
void
html_a_file (struct context *cx, const char *filename, const struct std *url)
{
  int exists;

  assert (cx != NULL && filename != NULL);
  
  {
    struct stat statbuf;
    exists = lstat (filename, &statbuf) == 0;
  }

  if (exists)
    html ("<A HREF=\"%s\">", url_file (filename, url));
  if (has_prefix (filename, cx->shared))
    {
      filename += strlen (cx->shared);
      if (*filename == '/')
	filename++;
      if (*filename == '\0')
	html ("(%s)", _("shared folder"));
    }
  else if (has_prefix (filename, cx->home))
    {
      filename += strlen (cx->home);
      if (*filename == '/')
	filename++;
      html ("(%s) ", _("personal folder"));
    }
  html_escape (filename);
  if (exists)
    htmls ("</A>");
}

/* Display a button for screen SCREEN with label LABEL.
   Additional arguments are taken as name-value pairs,
   up to the first null pointer argument. */
void
html_button (const char *screen, const char *label, ...)
{
  va_list args;
  
  html_start_form (screen, HT_GET, HT_URL_ENCODED);
  
  va_start (args, label);
  for (;;)
    {
      const char *name, *value;
      
      name = va_arg (args, const char *);
      if (name == NULL)
	break;

      value = va_arg (args, const char *);
      html_hidden (name, value);
    }
  va_end (args);

  html_submit (label);
  html_end_form ();
}

#ifndef NDEBUG
/* 0=not in a form, 1=in a form. */
static int in_form;
#endif

/* Start an HTML form.  The form will be submitted to screen SCREEN
   using method METH (HT_GET or HT_POST) and with submission encoding
   type ENCTYPE (HT_URL_ENCODED or HT_FORM_DATA). */
void
html_start_form (const char *screen, int method, int enctype)
{
  assert (method == HT_GET || method == HT_POST);
  assert (enctype == HT_URL_ENCODED || enctype == HT_FORM_DATA);
  assert (++in_form == 1);

  html ("<FORM NAME=form ACTION=\"%s\"\n"
	" METHOD=\"%s\" ENCTYPE=\"%s\">\n",
	url_this (screen, (char *) NULL),
	method == HT_GET ? "GET" : "POST",
	(enctype == HT_URL_ENCODED
	 ? "application/x-www-form-urlencoded" : "multipart/form-data"));
}

/* End an HTML form. */
void
html_end_form (void)
{
  assert (--in_form == 0);
  htmls ("</FORM>\n\n");
}

/* Output an HTML form text field with name NAME, default contents
   DEF, with displayed size SIZE. */
void
html_text_field (const char *name, const char *def, int size)
{
  assert (in_form);
  html ("<INPUT NAME=%s SIZE=%d VALUE=\"", name, size);
  if (def != NULL)
    html_escape (def);
  htmls ("\">\n");
}

/* Output an HTML form password field with name NAME and displayed
   size SIZE. */
void
html_password_field (const char *name, int size)
{
  assert (in_form);
  html ("<INPUT TYPE=PASSWORD NAME=%s SIZE=%d>\n", name, size);
}

/* Output an HTML form file upload field with name NAME. */
void
html_file_field (const char *name)
{
  assert (in_form);
  html ("<INPUT TYPE=file NAME=%s>\n", name);
}

/* Output an HTML form radio button named NAME with value VALUE,
   labeled with LABEL.  If CHECKED is nonzero then the radio button
   will be initially selected. */
void 
html_radio (const char *name, const char *value, const char *label,
	    int checked)
{
  assert (in_form);
  html ("<LABEL><INPUT TYPE=radio NAME=%s VALUE=\"", name);
  html_escape (value);
  html ("\"%s>%s</LABEL>\n", checked ? " CHECKED" : "", label);
}

/* Output an HTML form checkbox named NAME and labeled with LABEL.  If
   CHECKED is nonzero then the checkbox will be initially checked. */
void
html_checkbox (const char *name, int checked, const char *label)
{
  assert (in_form);
  html ("<LABEL><INPUT TYPE=checkbox NAME=%s VALUE=yes%s>%s</LABEL>\n",
	name, checked ? " CHECKED" : "", label);
}

/* Output an HTML form text area named NAME with ROWS rows and COLUMNS
   columns.  By default the text area will contain DEF.

   WRAP=PHYSICAL is not HTML 4.0 compliant, but users complain if it
   isn't present. */
void
html_text_area (const char *name, const char *def, int rows, int columns)
{
  assert (in_form);
  html ("<TEXTAREA NAME=%s ROWS=%d COLS=%d WRAP=PHYSICAL>\n",
	name, rows, columns);
  html_escape (def);
  htmls ("</TEXTAREA>\n");
}

/* Output an HTML form submit button labeled LABEL. */
void
html_submit (const char *label)
{
  assert (in_form);
  html ("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", label);
}

/* Output an HTML form submit button named NAME and labeled as
   LABEL. */
void
html_submit_alt (const char *name, const char *label)
{
  assert (in_form);
  html ("<INPUT TYPE=SUBMIT NAME=%s VALUE=\"%s\">\n", name, label);
}

/* Output an HTML hidden text field with name NAME and value VALUE. */
void
html_hidden (const char *name, const char *value)
{
  assert (in_form);
  html ("<INPUT TYPE=HIDDEN NAME=%s VALUE=\"", name);
  if (value != NULL)
    html_escape (value);
  htmls ("\">\n");
}

/* Output a label LABEL for HTML form element name NAME. */
void
html_label (const char *name unused, const char *label)
{
  assert (in_form);
#if 1
  htmls (label);
#else
  /* This is disabled because it screws up Mozilla rendering of the
     login page. */
  html ("<LABEL FOR=%s>%s</LABEL>\n", name, label);
#endif
}

/* Functions for URLs. */

/* Initialize URL as a url referring to this script and screen
   SCREEN. */
void
url_create (struct url *url, const char *screen)
{
  st_create (&url->s, SCRIPT_NAME);
  if (screen != NULL)
    {
      st_putchar (&url->s, '/');
      st_concat_url_query_quoted (&url->s, screen);
    }
  url->n_args = 0;
}

/* Free URL. */
void 
url_destroy (struct url *url)
{
  st_destroy (&url->s);
}

/* Add `NAME=VALUE' to the query string for URL, or simply `NAME' if
   VALUE is a null pointer. */
void
url_arg (struct url *url, const char *name, const char *value)
{
  st_concat (&url->s, url->n_args++ ? "&amp;" : "?");
  st_concat_url_query_quoted (&url->s, name);
  if (value != NULL)
    {
      st_putchar (&url->s, '=');
      st_concat_url_query_quoted (&url->s, value);
    }
}

/* Add `NAME=VALUE' to the query string for URL.  VALUE is formatted
   as an string. */
void
url_arg_int (struct url *url, const char *name, int value)
{
  st_concat (&url->s, url->n_args++ ? "&amp;" : "?");
  st_concat_url_query_quoted (&url->s, name);
  st_printf (&url->s, "=%d", value);
}

/* Puts the values in DICT as arguments to URL, or if URL is a null
   pointer then as hidden form elements. */
void
url_arg_std (struct url *url, const struct std *dict)
{
  struct std_iterator i;

  if (dict == NULL)
    return;
  
  std_init_iterator (dict, &i);
  for (;;)
    {
      const char *key;
      char *value;

      if (!std_iterate (&i, &key, &value))
	return;

      if (url != NULL)
	url_arg (url, key, value);
      else
	html_hidden (key, value);
    }
}

/* Copy URL from SRC to DST.  Don't initialize SRC before calling this
   function. */
void
url_copy (struct url *dst, const struct url *src)
{
  *dst = *src;
  st_create (&dst->s, st_value (&src->s));
}

/* Mark the current location in URL for later release. */
void
url_mark (struct url *url)
{
  url->mark_length = st_length (&url->s);
  url->mark_n_args = url->n_args;
}

/* Delete all the arguments in URL up to the last marked location. */
void
url_release (struct url *url)
{
  st_truncate (&url->s, url->mark_length);
  url->n_args = url->mark_n_args;
}

/* Return the string represented by URL. */
const char *
url_string (const struct url *url)
{
  return st_value (&url->s);
}

/* Returns a string URL in a static buffer for SCREEN with specified
   varargs arguments giving names and values in pairs.  Values may be
   null to indicate a name-only argument.  Terminate the list of
   varargs arguments with a null pointer by itself. */
const char *
url_this (const char *screen, ...)
{
  static struct url *url;
  va_list args;

  if (url == NULL)
    {
      url = xmalloc (sizeof *url);
      st_init (&url->s, 128);
    }
  else
    st_clear (&url->s);

  st_concat (&url->s, our_abs_url);
  st_putchar (&url->s, '/');
  st_concat_url_query_quoted (&url->s, screen);
  url->n_args = 0;

  va_start (args, screen);
  for (;;)
    {
      const char *name;
      const char *value;

      name = va_arg (args, const char *);
      if (name == NULL)
	{
	  va_end (args);
	  return st_value (&url->s);
	}

      value = va_arg (args, const char *);
      url_arg (url, name, value);
    }
}

/* Returns a string for a URL for the given FILENAME under the
   TeamSCOPE file manager.
   Extra arguments for the URL may be provided in EXTRA. */
const char *
url_file (const char *filename, const struct std *extra)
{
  static struct url *url;

  if (url == NULL)
    {
      url = xmalloc (sizeof *url);
      url_create (url, "files");
      url_mark (url);
    }

  url_release (url);
  url_arg_std (url, extra);
  url_arg (url, "file", filename);
  return st_value (&url->s);
}

/* Writes FORMAT to the HTML stream, formatting it with printf. */
void 
html (const char *format, ...)
{
  if (html_buffer == NULL)
    {
      va_list args;

      va_start (args, format);
      vprintf (format, args);
      va_end (args);
    }
  else
    {
      /* It's too bad that ANSI C's va_list is so fscked up that no
	 practical function st_vprintf() can be written. */
      static char *buf;
      static int size;

      va_list args;

      int needed;

      if (buf == NULL)
	{
	  size = 128;
	  buf = xmalloc (size);
	}

      va_start (args, format);
      needed = vsnprintf (buf, size, format, args);
      va_end (args);

      if (needed >= size)
	{
	  size *= 2;
	  if (size < needed)
	    size = needed + 1;
	  buf = xrealloc (buf, size);
      
	  va_start (args, format);
	  vsprintf (buf, format, args);
	  va_end (args);
	}
      else
	while (needed == -1)
	  {
	    size = size * 2 + 32;
	    buf = xrealloc (buf, size);
	
	    va_start (args, format);
	    needed = vsnprintf (buf, size, format, args);
	    va_end (args);
	  }

      st_concat (html_buffer, buf);
    }
}

struct st *html_buffer;

/* Direct HTML output to BUFFER. */
void
html_push_buffer (struct st *buffer)
{
  assert (html_buffer == NULL);
  html_buffer = buffer;
}

/* Cancel redirection of HTML output. */
void
html_pop_buffer (void)
{
  assert (html_buffer != NULL);
  html_buffer = NULL;
}

/* Writes STRING to the HTML stream. */
void 
htmls (const char *s)
{
  if (html_buffer == NULL)
    fputs (s, stdout);
  else
    st_concat (html_buffer, s);
}

/* Writes character C to the HTML stream. */
void htmlc (int c)
{
  if (html_buffer == NULL)
    putchar (c);
  else
    st_putchar (html_buffer, c);
}

/* Ensures that everything written to the HTML stream thus far has
   actually appeared on it. */
void 
html_flush (void)
{
  if (html_buffer == NULL)
    fflush (stdout);
}

