/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <stdlib.h>
#include "alloc.h"
#include "db.h"
#include "html.h"
#include "id-list.h"
#include "std.h"
#include "str.h"

/* ID list ADT implementation. */

/* Makes LIST an empty list. */
void 
id_list_null (struct id_list *list)
{
  list->n = 0;
  list->s = NULL;
}

/* Returns nonzero only if LIST is empty. */
int
id_list_null_p (const struct id_list *list)
{
  return list->n == 0;
}

/* Frees memory owned by LIST. */
void 
id_list_destroy (struct id_list *list)
{
  size_t i;

  for (i = 0; i < list->n; i++)
    free (list->s[i]);
  free (list->s);
}

/* Makes LIST a copy of SRC. */
void
id_list_copy (struct id_list *list, const struct id_list *src)
{
  size_t i;
  
  list->n = src->n;
  list->s = xmalloc (sizeof *list->s);
  for (i = 0; i < list->n; i++)
    list->s[i] = xstrdup (src->s[i]);
}

/* Makes LIST a list of length one, containing only ID. */
void 
id_list_one (struct id_list *list, const char *id)
{
  list->n = 1;
  list->s = xmalloc (sizeof *list->s);
  list->s[0] = xstrdup (id);
}

/* Returns nonzero only if lists A and B have the same contents. */
int
id_list_equal (const struct id_list *a, const struct id_list *b)
{
  size_t i;

  if (a->n != b->n)
    return 0;
  for (i = 0; i < a->n; i++)
    if (strcmp (a->s[i], b->s[i]))
      return 0;
  return 1;
}

/* Grabs the first column from R and makes LIST out of it.
   The column must be in sorted order.
   Destroys R. */
void
id_list_from_query (struct id_list *list, db_result *r)
{
  size_t i;
  
  list->n = dbr_nrows (r);
  list->s = xmalloc (sizeof *list->s * list->n);
  for (i = 0; i < list->n; i++)
    list->s[i] = xstrdup (dbr_value (r, i, 0));

  dbr_destroy (r);
}

/* Makes LIST a list of all users. */
void
id_list_all (struct id_list *list)
{
  db_result *r;

  db_query ("SELECT usr FROM users ORDER BY usr", DB_TUPLES, &r);
  id_list_from_query (list, r);
}

/* Makes LIST from the values associated with KEY within DICT.
   If MAX is nonzero then it is taken as a limit on the number
   of ids that will be put into LIST. */
void 
id_list_from_std (struct id_list *list, const struct std *dict,
		  const char *key, size_t max)
{
  const char **s;
  size_t i;

  s = std_get_many (dict, key, &list->n);
  if (max && list->n > max)
    list->n = max;

  list->s = xmalloc (sizeof *list->s * list->n);
  for (i = 0; i < list->n; i++)
    list->s[i] = xstrdup (s[i]);
  id_list_sort (list);
}

/* Puts LIST into DICT associated with KEY.
   Any existing key with the same name is retained. */
void 
id_list_to_std (struct id_list *list, struct std *dict, const char *key)
{
  size_t i;

  for (i = 0; i < list->n; i++)
    std_put (dict, key, list->s[i]);
}

/* Puts LIST into QUERY as a list of the form ('a','b','c'). */
void
id_list_to_sql (struct id_list *list, struct st *query)
{
  size_t i;
  
  st_putchar (query, '(');
  for (i = 0; i < list->n; i++)
    {
      if (i > 0)
	st_putchar (query, ',');
      st_format (query, "'%sq'", list->s[i]);
    }
  st_putchar (query, ')');
}

/* Makes LIST a list of all users in group GRP. */
void 
id_list_for_group (struct id_list *list, const char *grp)
{
  db_result *r;
  
  db_query ("SELECT usr FROM user_group WHERE grp = '%sq' ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r, grp);
  id_list_from_query (list, r);
}

/* Makes LIST a list of all users in a group USR belongs to. */
void
id_list_for_user (struct id_list *list, const char *usr)
{
  db_result *r;
  
  db_query ("SELECT DISTINCT usr FROM user_group "
	    "WHERE grp IN (SELECT grp FROM user_group WHERE usr = '%sq') "
	    "ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r, usr);
  id_list_from_query (list, r);
}

/* Makes LIST a list of all users who belong to only one group. */
void 
id_list_one_group (struct id_list *list)
{
  db_result *r;
  size_t i;

  /* We really want HAVING count(*) = 1 here, but old PostgreSQL
     doesn't support HAVING clause. */
  db_query ("SELECT usr, count(*) FROM user_group GROUP BY usr ORDER BY usr",
	    DB_TUPLES, &r);

  list->n = 0;
  list->s = xmalloc (sizeof *list->s * dbr_nrows (r));
  for (i = 0; i < dbr_nrows (r); i++)
    if (stol (dbr_value (r, i, 1)) == 1)
      list->s[list->n++] = xstrdup (dbr_value (r, i, 0));

  dbr_destroy (r);
}

/* Removes from LA the ids that are not also in LB.
   Destroys LB. */
void 
id_list_intersect (struct id_list *la, struct id_list *lb)
{
  size_t a, b, d;

  b = d = 0;
  for (a = 0; a < la->n; a++)
    {
      while (b < lb->n && strcmp (la->s[a], lb->s[b]) > 0)
	b++;
      if (b < lb->n && strcmp (la->s[a], lb->s[b]) == 0)
	{
	  la->s[d++] = la->s[a];
	  b++;
	}
      else
	free (la->s[a]);
    }
  la->n = d;
  id_list_destroy (lb);
}

#if 0
/* This function is untested, so don't trust it. */

/* Removes from LA the ids in LB. */
void
id_list_subtract (struct id_list *la, struct id_list *lb)
{
  size_t d, s, p;

  d = s = p = 0;
  for (;;)
    {
      int cmp = strcmp (la->s[p], lb->s[s]);
      if (cmp > 0)
	{
	  if (d != s)
	    {
	      free (la->s[d]);
	      la->s[d++] = la->s[s++];
	    }
	  if (s >= la->n)
	    break;
	}
      else if (cmp < 0)
	{
	  p++;
	  if (p >= lb->n)
	    break;
	}
      else /* cmp == 0 */
	{
	  s++;
	  if (s >= la->n)
	    break;
	}
    }

  if (d != s)
    while (s < la->n)
      {
	free (la->s[d]);
	la->s[d++] = la->s[s++];
      }

  for (p = d; p < la->n; p++)
    free (la->s[p]);
  la->n = d;
}
#endif /* 0 */

/* qsort() and bsearch() comparison function for ids. */
int
compare_ids (const void *pa, const void *pb)
{
  const char *const *a = pa;
  const char *const *b = pb;

  return strcmp (*a, *b);
}

/* Sorts LIST. */
void
id_list_sort (struct id_list *list)
{
  qsort (list->s, list->n, sizeof *list->s, compare_ids);
}

/* Returns the index of ID within LIST, or -1 if ID is not in LIST. */
int
id_list_find (const struct id_list *list, const char *id)
{
  char **p;

  p = bsearch (&id, list->s, list->n, sizeof *list->s, compare_ids);
  return p ? p - list->s : -1;
}

#if 0
/* Returns nonzero if ID is in LIST. */
int 
id_list_contains (struct id_list *list, const char *id)
{
  return bsearch (&id, list->s, list->n, sizeof *list->s, compare_ids) != NULL;
}
#endif

/* Displays LIST as a comma-separated list. */
void
id_list_print (const struct id_list *list)
{
  size_t i;

  for (i = 0; i < list->n; i++)
    {
      if (i > 0)
	htmls (", ");
      html_escape (list->s[i]);
    }
}
