/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <gd.h>
#include <gdfonts.h>
#include <stdio.h>
#include <stdlib.h>
#include "activity.h"
#include "alloc.h"
#include "cgi.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "info.h"
#include "scope.cgi.h"
#include "std.h"

/* List of categories that users can request information on. */
enum
  {
    IT_USERNAME,
    IT_FULL_NAME,
    IT_LAST_LOGIN,
    IT_USAGE_PATTERN,
    IT_DAYS_USED,
    IT_UD_COUNT,
    IT_RU_COUNT,

    IT_N_TYPES
  };

/* One category of information. */
struct info_type
  {
    int id;		/* One of IT_*. */
    char *name;		/* Name used as CGI variable. */
    char *desc;		/* Description put on form for user to check. */
    char *title;	/* Title used on output table; NULL if same as desc. */
  };

/* An entire user information request. */
struct info_request 
  {
    const char **teams;		/* Teams of interest. */
    int n_teams;		/* Number of teams. */
    char *cond;			/* SQL condition for inclusion of user. */

    struct user_data *users;	/* Info on users of interest. */
    int n_users;		/* Number of users. */

    unsigned which;		/* Bitmask from 1 << IT_*. */
    int duration;		/* Number of weeks in period of interest. */

    int n_posted;		/* Number of messages posted in period. */
  };

/* Information on a user. */
struct user_data
  {
    char *username;		/* Username. */
    char *full_name;		/* Full name. */
    char *last_login;		/* Last login date, time. */
    int days_used;		/* Number of days used in period. */
    int ud_count;		/* Upload/download count in period. */
    int p_count;		/* Messages posted in period of interest. */
    int r_count;		/* Messages read in period of interest. */
  };

/* All the information types. */
struct info_type info_types[] =
  {
    {IT_USERNAME,
     "username",
     N_("Username"),
     NULL},

    {IT_FULL_NAME,
     "fullname",
     N_("Full name"),
     NULL},

    {IT_LAST_LOGIN,
     "lastlogin",
     N_("Last login"),
     NULL},
    
    {IT_USAGE_PATTERN,
     "usagepattern",
     N_("Usage pattern over time period"),
     N_("Usage pattern")},
    
    {IT_DAYS_USED,
     "daysused",
     N_("Number of days used over time period"),
     N_("# of days used")},
    
    {IT_UD_COUNT,
     "udcount",
     N_("File upload/download count for distinct files during time period"),
     N_("File activity count")},
    
    {IT_RU_COUNT,
     "rucount",
     N_("Read and unread messages during time period"),
     N_("Read/posted messages<BR>Total posted: %d<SUP>*</SUP>")},

    {0, 0, 0, 0},
  };

static void show_info (struct request *);

static void info_form (struct request *);
static void offer_workgroups (struct request *);

/* Handles `info' CGI screen. */
void
scr_info (struct request *rq)
{
  if (std_contains (rq->cgi, "duration"))
    show_info (rq);
  else
    info_form (rq);
}

/* Handles `infousagegif' CGI screen by dumping a GIF graph of to the
   client. */
void 
scr_infousagegif (struct request *rq)
{
  /* Basics. */
  #define I_TEXT_HEIGHT 10		/* Height of font. */
  #define I_TEXT_Y_OFS -3		/* Offset to add to font y coord. */
  #define I_TEXT_WIDTH 6		/* Width of font. */
  #define I_MARGIN 3			/* Margin between elements. */

  /* Vertical. */
  #define I_HEADER_HEIGHT (I_TEXT_HEIGHT + I_MARGIN + 1)
  #define I_BODY_HEIGHT (4 * I_TEXT_HEIGHT + 3 * I_MARGIN)
  #define I_TRAILER_HEIGHT 1
  #define I_HEIGHT (I_HEADER_HEIGHT + I_BODY_HEIGHT + I_TRAILER_HEIGHT)
  #define I_CELL_HEIGHT (I_BODY_HEIGHT / 7)

  /* Horizontal. */
  #define I_HEADER_WIDTH (2 * I_TEXT_WIDTH + 2 * I_MARGIN + 1)
  #define I_BODY_WIDTH (24 * I_CELL_WIDTH)
  #define I_TRAILER_WIDTH 1
  #define I_WIDTH (I_HEADER_WIDTH + I_BODY_WIDTH + I_TRAILER_WIDTH)
  #define I_CELL_WIDTH 2

  /* Image to draw on. */
  gdImagePtr im;
  int clear, black, blue;		/* Color palette values. */

  /* Histogram of usage counts per hour per week. */
  int hist[7 * 24];

  /* Scale factor for graph; max value in hist[]. */
  double scale = 1.0;

  /* Determine histogram values and scale factor. */
  {
    db_result *r;

    /* Database query. */
    {
      int duration;
      const char *graph_user;

      /* Parse duration and desired target user out of screen in
	 format `infousagegif-%d-%s'.  For some reason, couldn't get
	 sscanf() to work properly with this. */
      {
	const char *p;

	p = strchr (rq->screen, '-');
	if (!p)
	  return;
	duration = stoi (++p);

	p = strchr (p, '-');
	if (!p)
	  return;
	graph_user = ++p;
      }
      
      /* Make sure that the graph user is in a workgroup with the
       current user. */
      db_begin ();
      db_query ("SELECT a.grp FROM user_group a, user_group b "
		"WHERE a.grp = b.grp AND a.usr = '%sq' and b.usr = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, rq->cx.usr, graph_user);
      if (dbr_nrows (r) == 0)
	{
	  db_end ();
	  return;
	}
      dbr_destroy (r);

      /* Actual query. */
      db_query ("SELECT "
		"date_part ('hour', t) + date_part ('dow', t) * 24 AS ofs, "
		"COUNT(*) "
		"FROM uselog "
		"WHERE usr = '%sq' "
		"AND t > ('now'::datetime - '%d weeks'::timespan) "
		"GROUP BY ofs",
		DB_FORMAT | DB_TUPLES, &r, graph_user, duration);
      db_end ();
    }

    /* Parse database result into hist[].
       There ought to be a way to do this that avoids our having to
       summarize it by hand, and ISTR that there was with a
       sufficiently new version of PostgreSQL, but that it wouldn't
       work with older versions. */
    {
      int count;
      int row;

      count = dbr_nrows (r);
      memset (hist, 0, sizeof hist);
      for (row = 0; row < count; row++)
	{
	  int index = stoi (dbr_value (r, row, 0));
	  int value = stoi (dbr_value (r, row, 1));

	  if (index >= 0 && index < 24 * 7)
	    {
	      hist[index] += value;
	      if (hist[index] > scale)
		scale = hist[index];
	    }
	}
    }

    dbr_destroy (r);
  }

  /* Create image and color palette. */
  im = gdImageCreate (I_WIDTH, I_HEIGHT);
  clear = gdImageColorAllocate (im, 0000, 0000, 0377);
  gdImageColorTransparent (im, clear);

  black = gdImageColorAllocate (im, 0000, 0000, 0000);
  blue = gdImageColorAllocate (im, 0000, 0000, 0377);
  
  /* Draw diagram. */
  {
    int index;

    for (index = 0; index < 24 * 7; index++)
      if (hist[index])
	{
	  int dow = index / 24;
	  int tod = index % 24;
	  int x1 = I_HEADER_WIDTH + I_CELL_WIDTH * tod;
	  int y1 = I_HEADER_HEIGHT + I_CELL_HEIGHT * dow + 1;
	  int x2 = x1 + I_CELL_WIDTH - 1;
	  int y2 = y1 + I_CELL_HEIGHT - 2;

	  gdImageRectangle (im, x1, y1, x2, y2, blue);
	}
  }

  /* Draw days of the week y-axis legend. */
  {
    int i;

    for (i = 0; i < 4; i++)
      gdImageChar (im, gdFontSmall,
		   0,
		   (I_HEADER_HEIGHT + i * (I_MARGIN + I_TEXT_HEIGHT)
		    + I_TEXT_Y_OFS + i),
		   "STTS"[i], black);

    for (i = 0; i < 3; i++)
      gdImageChar (im, gdFontSmall,
		   I_TEXT_WIDTH + I_MARGIN,
		   (I_HEADER_HEIGHT
		    + (2 * i + 1) * (I_MARGIN + I_TEXT_HEIGHT) / 2
		    + i + I_TEXT_Y_OFS),
		   "MWF"[i], black);

    for (i = 0; i < 7; i++)
      {
	int y = I_HEADER_HEIGHT + I_CELL_HEIGHT * (2 * i + 1) / 2;
	
	gdImageLine (im,
		     I_HEADER_WIDTH - I_MARGIN - 1, y,
		     I_HEADER_WIDTH - 1, y,
		     black);
      }
  }

  /* Draw time-of-day x-axis legend. */
  gdImageChar (im, gdFontSmall,
	       I_HEADER_WIDTH + I_BODY_WIDTH / 4 - I_TEXT_WIDTH / 2 + 1,
	       I_TEXT_Y_OFS,
	       '6', black);

  gdImageString (im, gdFontSmall,
		 I_HEADER_WIDTH + I_BODY_WIDTH / 2 - I_TEXT_WIDTH + 1,
		 I_TEXT_Y_OFS,
		 "12", black);

  gdImageChar (im, gdFontSmall,
	       I_HEADER_WIDTH + I_BODY_WIDTH * 3 / 4 - I_TEXT_WIDTH / 2 + 1,
	       I_TEXT_Y_OFS,
	       '6', black);
  {
    int i;

    for (i = 1; i <= 3; i++)
      {
	int x = I_HEADER_WIDTH + I_BODY_WIDTH * i / 4;

	gdImageLine (im,
		     x, I_HEADER_HEIGHT - I_MARGIN - 1,
		     x, I_HEADER_HEIGHT + I_BODY_HEIGHT,
		     black);
      }
  }
  
  /* Rectangle. */
  gdImageRectangle (im,
		    I_HEADER_WIDTH - 1,
		    I_HEADER_HEIGHT - 1,
		    I_HEADER_WIDTH + I_BODY_WIDTH,
		    I_HEADER_HEIGHT + I_BODY_HEIGHT,
		    black);

  /* Output the gif image. */
  fputs ("Content-Type: image/gif\n\n", stdout);
  gdImageGif (im, stdout);
}

/* Present a form to the user offering search options. */
static void
info_form (struct request *rq)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("Team Info Search"));

  html_start_form ("info", HT_POST, HT_URL_ENCODED);

  offer_workgroups (rq);

  html ("<P>%s<BR>\n"
	"<INPUT TYPE=radio NAME=duration VALUE=1>%s</INPUT><BR>\n"
	"<INPUT TYPE=radio NAME=duration VALUE=2 CHECKED>%s</INPUT><BR>\n"
	"<INPUT TYPE=radio NAME=duration VALUE=3>%s</INPUT><BR>\n"
	"<INPUT TYPE=radio NAME=duration VALUE=4>%s</INPUT>\n"
	"</P>\n\n",
	_("Some of the available types of information are summaries over\n"
	  "a time period.  For use in these summaries, select a time period\n"
	  "from the list below."),
	_("The last 7 days"),
	_("The last 14 days"),
	_("The last 21 days"),
	_("The last 28 days"));
  
  {
    struct info_type *i;

    html ("<P>%s<BR>\n"
	  "<TABLE CELLSPACING=0>\n",
	  _("Select the types of information you want for each member of "
	    "the selected teams."));

    for (i = info_types; i->name; i++)
      html ("  <TR>\n"
	    "    <TD><INPUT TYPE=checkbox NAME=%s VALUE=on%s></TD>\n"
	    "    <TD>%s</TD>\n"
	    "  </TR>\n",
	    i->name,
	    (i->id == IT_USERNAME || i->id == IT_LAST_LOGIN) ? " CHECKED" : "",
	    gettext (i->desc));
    html ("</TABLE>\n");
  }

  html_submit (_("Submit"));

  html_end_form ();
}

/* Display a selection box offering all the workgroups that the user
   belongs to. */
static void
offer_workgroups (struct request *rq)
{
  db_result *r;
  size_t count, i;

  db_query ("SELECT grp, name FROM groups "
	    "WHERE grp IN (SELECT grp FROM user_group WHERE usr = '%sq')",
	    DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

  count = dbr_nrows (r);
  if (count == 1)
    {
      dbr_destroy (r);
      return;
    }

  html ("<P>%s<BR>\n"
	"<SELECT MULTIPLE SIZE=4 NAME=teams>\n",
	_("Select the team or teams for which you want to see\n"
	  "information from the list below."));

  for (i = 0; i < count; i++)
    html ("  <OPTION VALUE=\"%s\">%s</OPTION>\n",
	  dbr_value (r, i, 0), dbr_value (r, i, 1));

  htmls ("</SELECT>\n"
	 "</P>\n\n");

  dbr_destroy (r);
}

static void info_request_from_cgi (struct request *, struct info_request *);
static void get_users (struct request *, struct info_request *);
static void get_days_used (struct info_request *);
static void get_ud_count (struct info_request *);
static void get_ru_count (struct info_request *);
static void display_request (struct info_request *);

/* Perform team info search and display results. */
static void
show_info (struct request *rq)
{
  struct info_request request;

  /* Figure out what user wants. */
  info_request_from_cgi (rq, &request);

  /* Get what the user wants. */
  db_begin ();
  get_users (rq, &request);
  if (request.which & (1 << IT_DAYS_USED))
    get_days_used (&request);
  if (request.which & (1 << IT_UD_COUNT))
    get_ud_count (&request);
  if (request.which & (1 << IT_RU_COUNT))
    get_ru_count (&request);
  db_end ();

  /* Show what the user wants. */
  display_request (&request);

  free (request.cond);
}

/* Figure out what the user wants by reading the CGI variables. */
static void
info_request_from_cgi (struct request *rq, struct info_request *request)
{
  struct info_type *i;

  request->teams = std_get_many (rq->cgi, "teams", &request->n_teams);
  
  request->which = 0;
  for (i = info_types; i->name; i++)
    if (std_contains (rq->cgi, i->name))
      request->which |= 1 << i->id;

  request->duration = stoi (std_get (rq->cgi, "duration"));
  if (request->duration < 1)
    request->duration = 1;
  else if (request->duration > 4)
    request->duration = 4;
}

/* Get a list of users in the teams requested. */
static void
get_users (struct request *rq, struct info_request *request)
{
  db_result *r;
  struct st s;
  int i;

  /* Make sure that the user really belongs to all the teams he wants
     info on. */
  db_query ("SELECT grp FROM user_group WHERE usr = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);
  for (i = 0; i < request->n_teams; i++)
    {
      size_t j;
      
      for (j = 0; j < dbr_nrows (r); j++)
	if (!strcmp (dbr_value (r, j, 0), request->teams[i]))
	  break;
      if (j >= dbr_nrows (r))
	fail ("user %s is cheating--not member of %s",
	      rq->cx.usr, request->teams[i]);
    }
  dbr_destroy (r);

  /* Get a list of users in the teams of interest. */
  st_create (&s, "(grp IN (");
  for (i = 0; i < request->n_teams; i++)
    {
      if (i)
	st_concat (&s, ", ");
      st_concat_quoted (&s, request->teams[i], '\'');
    }
  if (request->n_teams == 0)
    st_concat_quoted (&s, rq->cx.grp, '\'');
  st_concat (&s, "))");

  db_query ("SELECT DISTINCT usr FROM user_group WHERE %ss ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r, st_value (&s));

  /* Construct a condition string that will include only those users. */
  st_clear (&s);
  st_concat (&s, "(usr IN (");
  for (i = 0; i < (int) dbr_nrows (r); i++)
    {
      if (i)
	st_concat (&s, ", ");
      st_concat_quoted (&s, dbr_value (r, i, 0), '\'');
    }
  st_concat (&s, "))");
  request->cond = st_value (&s);

  /* Get username, full name, and last usage date/time for the users
     on the list, and record them. */
  db_query ("SELECT usr, name, last_usage FROM users WHERE %ss ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r, request->cond);

  request->n_users = dbr_nrows (r);
  request->users = xmalloc (sizeof *request->users * request->n_users);
  for (i = 0; i < request->n_users; i++)
    {
      struct user_data *u = &request->users[i];

      u->username = xstrdup (dbr_value (r, i, 0));

      u->full_name = u->last_login = NULL;
      if (request->which & (1 << IT_FULL_NAME))
	u->full_name = xstrdup (dbr_value (r, i, 1));
      if (request->which & (1 << IT_LAST_LOGIN))
	u->last_login = xstrdup (dbr_value (r, i, 2));

      u->days_used = u->ud_count = u->p_count = u->r_count = 0;
    }
  request->n_posted = 0;

  dbr_destroy (r);
}

/* Returns I advanced to user USERNAME.  REQUEST is the associated
   info_request record.  I may be a null pointer on entry.  Returns
   a null pointer if no such user exists. */
static struct user_data *
advance_to (const struct info_request *request, struct user_data *i,
	    const char *username)
{
  if (i == NULL)
    i = request->users;

  while (strcmp (i->username, username))
    {
      i++;
      
      if (i >= request->users + request->n_users)
	return NULL;
    }
  
  return i;
}

/* Count the number of days each user of interest used TeamSCOPE
   during the period of interest. */
static void
get_days_used (struct info_request *request)
{
  db_result *r;
  struct user_data *i;
  int row, count;

  db_query ("SELECT DISTINCT usr, date_part ('day', t) AS dow "
	    "FROM uselog "
	    "WHERE %ss "
	    "AND t > ('now'::datetime - '%d weeks'::timespan) "
	    "ORDER BY usr, dow",
	    DB_FORMAT | DB_TUPLES, &r, request->cond, request->duration);

  count = dbr_nrows (r);
  for (row = 0, i = NULL; row < count; row++)
    {
      i = advance_to (request, i, dbr_value (r, row, 0));

      if (i != NULL)
	i->days_used++;
    }
  dbr_destroy (r);
}

/* Count the number of file uploads and downloads by each user of
   interest during the period of interest. */
static void
get_ud_count (struct info_request *request)
{
  db_result *r;
  struct user_data *i;
  int row, count;

  db_query ("SELECT usr, count(*) "
	    "FROM activities "
	    "WHERE %ss "
	    "AND date > ('now'::datetime - '%d weeks'::timespan) "
	    "AND cat = %d AND (type >= %d AND type <= %d) "
	    "GROUP BY usr "
	    "ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r,
	    request->cond, request->duration,
	    AC_CAT_FILE, AC_FILE_DOWNLOAD, AC_FILE_UPLOAD);

  count = dbr_nrows (r);
  for (row = 0, i = NULL; row < count; row++)
    {
      i = advance_to (request, i, dbr_value (r, row, 0));
      if (i != NULL)
	i->ud_count += stoi (dbr_value (r, row, 1));
    }
  dbr_destroy (r);
}

/* Count the number of TeamSCOPE messages posted and read by each user
   of interest during the period of interest. */
static void
get_ru_count (struct info_request *request)
{
  db_result *r;
  struct user_data *i;
  int row, count;

  db_query ("SELECT usr, cat, type, count(*) "
	    "FROM activities "
	    "WHERE %ss "
	    "AND date > ('now'::datetime - '%d weeks'::timespan) "
	    "AND ((cat = %d AND type >= %d AND type <= %d) "
	    "OR (cat = %d AND type >= %d OR type <= %d)) "
	    "GROUP BY usr, cat, type "
	    "ORDER BY usr",
	    DB_FORMAT | DB_TUPLES, &r,
	    request->cond, request->duration,
	    AC_CAT_FILE, AC_FILE_POST, AC_FILE_POST_READ,
	    AC_CAT_MSG_BOARD, AC_MSG_POST, AC_MSG_READ);

  count = dbr_nrows (r);
  for (row = 0, i = NULL; row < count; row++)
    {
      int cat, type;
      int n;

      i = advance_to (request, i, dbr_value (r, row, 0));
      if (i == NULL)
	break;

      cat = stoi (dbr_value (r, row, 1));
      type = stoi (dbr_value (r, row, 2));

      n = stoi (dbr_value (r, row, 1));
      if ((cat == AC_CAT_FILE && type == AC_FILE_POST)
	  || (cat == AC_CAT_MSG_BOARD && type == AC_MSG_POST))
	{
	  i->p_count += n;
	  request->n_posted += n;
	}
      else if ((cat == AC_CAT_FILE && type == AC_FILE_DELETE)
	  || (cat == AC_CAT_MSG_BOARD && type == AC_MSG_DELETE))
	{
	  i->p_count -= n;
	  request->n_posted -= n;
	}
      else
	i->r_count += n;
    }
  dbr_destroy (r);
}

/* Display the results of the user's request. */
static void
display_request (struct info_request *request)
{
  struct user_data *i;
  int col;

  /* Table header. */
  html ("<TABLE BORDER=1>\n"
	"  <CAPTION>%s</CAPTION>\n"
	"  <TR>\n",
	_("User Info Summary"));

  for (col = 0; col < IT_N_TYPES; col++)
    if (request->which & (1 << col))
      {
	const char *title = info_types[col].title;
	if (title == NULL)
	  title = info_types[col].desc;

	htmls ("    <TH>");
	html (gettext (title), request->n_posted);
        htmls ("</TH>\n");
      }

  htmls ("  </TR>\n");

  /* Table body. */
  for (i = request->users; i < request->users + request->n_users; i++)
    {
      htmls ("  <TR ALIGN=CENTER>\n");
      
      for (col = 0; col < IT_N_TYPES; col++)
	if (request->which & (1 << col))
	  {
	    htmls ("    <TD>");

	    switch (col)
	      {
	      case IT_USERNAME:
		html_escape (i->username);
		break;

	      case IT_FULL_NAME:
		html_escape (i->full_name);
		break;

	      case IT_LAST_LOGIN:
		{
		  struct tm tm;
		  time_t t = datetime_parse (i->last_login, 0, 0, &tm);

		  htmls (t > 86400
			 ? datetime_format ("%b %d %Y %H:%M", &tm)
			 : _("never logged in") );
		}
		break;

	      case IT_USAGE_PATTERN:
		{
		  struct url url;
		  char *base;

		  base = xmalloc (strlen (i->username) + INT_DIGITS + 32);
		  sprintf (base, "infousagegif-%d-%s",
			   request->duration, i->username);
		  url_create (&url, base);
		  free (base);
		  
		  html ("<IMG WIDTH=%d HEIGHT=%d SRC=\"%s\">",
			I_WIDTH, I_HEIGHT, url_string (&url));
		  url_destroy (&url);
		}
		break;

	      case IT_DAYS_USED:
		html ("%d", i->days_used);
		break;

	      case IT_UD_COUNT:
		html ("%d", i->ud_count);
		break;

	      case IT_RU_COUNT:
		html ("%d/%d", i->r_count, i->p_count);
		break;

	      default:
		assert (0);
	      }
	    htmls ("</TD>\n");
	  }
      htmls ("  </TR>\n");
    }

  /* Table trailer and notes. */
  htmls ("</TABLE>\n");

  if (request->which & (1 << IT_RU_COUNT))
    {
      htmls ("<P><SUP>*</SUP>\n");
      html (_("Note: Number of posted messages is figured as messages posted\n"
	      "minus messages deleted during period of interest, so\n"
	      "negative values are possible.  Total number of messages\n"
	      "posted during period, minus number deleted, was %d.\n"),
	    request->n_posted);
      htmls ("</P>\n");
    }
}
