/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include "alloc.h"
#include "auth.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "identity.h"
#include "priv.h"
#include "scope.cgi.h"
#include "session.h"
#include "std.h"
#include "str.h"

/* Public functions. */

static void update_cache (struct request *);
static void retrieve_dir_info (struct request *, int update);

/* Sets the current real and effective uid and gid to USER, and clears
   all supplementary groups. */
void
priv_set_user (const char *user)
{
  struct passwd *passwd;
  
  assert (user != NULL);
  
  passwd = getpwnam (user);
  if (passwd == NULL)
    fail ("priv_set_user(): no such user \"%s\"", user);

  if (setgroups (0, NULL) == -1)
    fail ("priv_set_user(): setgroups failed for user %s", user);

  if (passwd->pw_gid < 100 || passwd->pw_uid < 100)
    fail ("priv_set_user(): passwd database corrupt: uid=%d, gid=%d "
	  "for user %s", passwd->pw_uid, passwd->pw_gid, user);

  if (setgid (passwd->pw_gid) == -1)
    fail ("priv_set_user(): setgid failed for user %s", user);

  if (setuid (passwd->pw_uid) == -1)
    fail ("priv_set_user(): setuid failed for user %s", user);
}

/* Makes sure that we have cached the proper information on how
   to set the current user. */
void
priv_cache (struct request *rq)
{
  int update;
  
  assert (rq->cx.usr[0] != '\0' && rq->cx.grp[0] != '\0');

  update = (strcmp (rq->cx.usr, std_get (rq->trusted, "user"))
	    || strcmp (rq->cx.grp, std_get (rq->trusted, "workgroup"))
	    || std_get_null (rq->trusted, "uid") == NULL);

  if (update)
    update_cache (rq);

  retrieve_dir_info (rq, update);
}

/* Set the current real and effective uid and gid and supplementary
   groups according to the cached identity data. */
void
priv_drop (struct request *rq)
{
  rq->cx.uid = stoi (std_get (rq->trusted, "uid"));
  rq->cx.gid = stoi (std_get (rq->trusted, "gid"));
  if (rq->cx.uid < 100 || rq->cx.gid < 100)
    fail ("priv_drop(): trusted data corrupt: uid=%d, gid=%d",
	  rq->cx.uid, rq->cx.gid);

  if (setgid (rq->cx.gid) == -1)
    fail ("priv_drop(): setgid(%d) failed for user %s",
	  rq->cx.gid, rq->cx.usr);

  /* Pick up all the supplementary GIDs. */
  {
    const char **tgids;
    int n_tgids;
    
    tgids = std_get_many (rq->trusted, "supp-gids", &n_tgids);
    if (n_tgids > 0)
      {
	int *vgids;
	int n_vgids;
	int i;

	vgids = xmalloc (sizeof *vgids * n_tgids);
	n_vgids = 0;
	for (i = 0; i < n_tgids; i++)
	  {
	    int vgid = stoi (tgids[i]);
	    
	    if (vgid >= 100)
	      vgids[n_vgids++] = vgid;
	  }

	if (n_vgids > 0 && setgroups (n_vgids, vgids) == -1)
	  fail ("priv_drop(): setgroups failed for user %s", rq->cx.usr);

	free (vgids);
      }
    }
  
  if (setuid (rq->cx.uid) == -1)
    fail ("priv_drop(): setuid(%d) failed for user %s",
	  rq->cx.uid, rq->cx.usr);
}

static void retrieve_groups (struct request *rq);
static void determine_group_admin (struct request *rq);
static void determine_system_admin (struct request *rq);
static char *xitoa (int);

/* Updates the identity cache that contains info on the session user's
   uid, gid, supplementary gids, and workgroup.  It's expensive to
   look through all the configuration files for this info, so we don't
   want to do it more than once per session. */
static void
update_cache (struct request *rq)
{
  struct passwd *passwd;
  struct group *group;
  
  assert (rq->cx.usr[0] != '\0');

  passwd = getpwnam (rq->cx.usr);
  if (passwd == NULL)
    fail ("no such user \"%s\"", rq->cx.usr);

  group = getgrnam (rq->cx.grp);
  if (group == NULL)
    fail ("no such group \"%s\"", rq->cx.grp);

  std_delete (rq->trusted, "user");
  std_delete (rq->trusted, "workgroup");
  std_delete (rq->trusted, "teamname");
  std_delete (rq->trusted, "uid");
  std_delete (rq->trusted, "gid");
  std_delete (rq->trusted, "supp-gids");
  std_delete (rq->trusted, "homedir");
  std_delete (rq->trusted, "shared-dir");
  std_delete (rq->trusted, "web-dir");
  std_delete (rq->trusted, "groups");
  std_delete (rq->trusted, "group-admin");
  std_delete (rq->trusted, "system-admin");

  std_put (rq->trusted, "gid", xitoa (group->gr_gid));
  std_put (rq->trusted, "teamname", rq->cx.fullgrp);
  std_put (rq->trusted, "homedir", passwd->pw_dir);

  /* Add the user's supplementary groups to "supp-gids".  
     Add the user's primary group in, too, if it's different from the
     one for his workgroup. */
  setgrent ();
  if (passwd->pw_gid != group->gr_gid)
    std_put (rq->trusted, "supp-gids", xitoa (passwd->pw_gid));
  for (;;)
    {
      struct group *supp_group;
      char **members;

      supp_group = getgrent ();
      if (supp_group == NULL)
	break;
      
      for (members = supp_group->gr_mem; *members; members++)
	if (!strcmp (*members, rq->cx.usr))
	  {
	    if (group->gr_gid != supp_group->gr_gid)
	      std_put (rq->trusted, "supp-gids", xitoa (supp_group->gr_gid));
	    break;
	  }
    }
  endgrent ();

  retrieve_groups (rq);
  
  determine_group_admin (rq);
  determine_system_admin (rq);

  /* Insert these last because they're what priv_cache() checks for.
     It wouldn't get saved anyway if there was an error, but what the
     hell, paranoia is good. */
  std_put (rq->trusted, "user", rq->cx.usr);
  std_put (rq->trusted, "workgroup", rq->cx.grp);
  std_put (rq->trusted, "uid", xitoa (passwd->pw_uid));
}

/* Sets the values of RQ->cx.shared, RQ->cx.home, and RQ->cx.web.  If
   UPDATE is nonzero then these are updated from the database and put
   into the cache, rather than read from the cache. */
static void
retrieve_dir_info (struct request *rq, int update)
{
  rq->cx.home = xstrdup (std_get (rq->trusted, "homedir"));
  assert (rq->cx.home[0] != '\0');

  if (update)
    {
      db_result *r;

      assert (rq->cx.grp[0] != '\0');
      db_query ("SELECT shareddir, webdir FROM groups WHERE grp = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, rq->cx.grp);

      rq->cx.shared = xstrdup (dbr_value (r, 0, 0));
      rq->cx.web = xstrdup (dbr_value (r, 0, 1));
      dbr_destroy (r);

      std_put (rq->trusted, "shared-dir", rq->cx.shared);
      std_put (rq->trusted, "web-dir", rq->cx.web);
    }
  else
    {
      rq->cx.shared = xstrdup (std_get (rq->trusted, "shared-dir"));
      rq->cx.web = xstrdup (std_get (rq->trusted, "web-dir"));
    }
  
  assert (rq->cx.shared != NULL && rq->cx.web != NULL);
}

/* Get a complete list of groups that the user belongs to. */
static void
retrieve_groups (struct request *rq)
{
  struct st s;
  db_result *r;
  size_t i;

  assert (rq->cx.usr[0] != '\0');
  db_query ("SELECT groups.grp, name FROM groups, user_group "
	    "WHERE groups.grp = user_group.grp AND usr = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);

  st_init (&s, 64);
  for (i = 0; i < dbr_nrows (r); i++)
    {
      st_clear (&s);
      st_printf (&s, "%s;%s", dbr_value (r, i, 0), dbr_value (r, i, 1));
      std_put (rq->trusted, "groups", st_value (&s));
    }
  st_destroy (&s);
}

/* Determine whether the user is privileged to act as a group admin in
   his group. */
static void
determine_group_admin (struct request *rq)
{
  db_result *r;
  
  db_query ("SELECT may_admin FROM user_group "
	    "WHERE grp = '%sq' and usr = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, rq->cx.grp, rq->cx.usr);
  if (dbr_value (r, 0, 0)[0] == 't')
    std_put (rq->trusted, "group-admin", "yes");
  dbr_destroy (r);
}

/* Determine whether the user is privileged to act as a system
   administrator. */
static void
determine_system_admin (struct request *rq)
{
  const char *filename = ADMINS_FILE;
  char *line = NULL;
  size_t line_size = 0;

  FILE *f;

  /* Open admins file. */
  f = fopen (filename, "r");
  if (f == NULL)
    {
      log ("opening %s: %s", filename, strerror (errno));
      return;
    }
  verify_file_owner (filename, f, 0);

  /* Process each line in the admins file. */
  for (;;)
    {
      /* Read a line. */
      long line_len = getline (&line, &line_size, f);
      if (-1 == line_len)
	{
	  if (ferror (f))
	    log ("reading %s: %s", filename, strerror (errno));
	  goto exit;
	}

      /* Remove any comment. */
      {
	char *comment = strchr (line, '#');
	if (comment != NULL)
	  *comment = '\0';
      }
      
      /* Check name read against user's name. */
      {
	char *bp = line + strspn (line, " \t\r\n");
	char *ep = bp + strcspn (bp, " \t\r\n");
	if (ep - bp > 0
	    && (size_t) (ep - bp) == strlen (rq->cx.usr)
	    && !memcmp (bp, rq->cx.usr, ep - bp))
	  {
	    std_put (rq->trusted, "system-admin", "yes");
	    goto exit;
	  }
      }
    }
 exit:
  free (line);
  fclose (f);
}

/* Converts X to a decimal string representation of its value in a
   static buffer. */
static char *
xitoa (int x)
{
  static char s[INT_DIGITS + 1];
  sprintf (s, "%d", x);
  return s;
}
