/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <stdlib.h>
#include "alloc.h"
#include "cgi.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "resource.h"
#include "scope.cgi.h"
#include "str.h"
#include "std.h"

struct avl_tree *resource_tree;

int n_resources;
struct resource *resource;

static void new_resource (void);
static void rename_resource (struct request *);
static void show_resource_list (void);

static void handle_new (struct request *);
static void handle_rename (struct request *);
static void handle_delete (struct request *);

static int compare_resource (const void *, const void *, void *);

void
scr_resources (struct request *rq)
{
  html ("<H2>%s</H2>\n", _("Shared Resources"));

  if (!strcmp (rq->op, "new"))
    {
      read_resource_list ();
      new_resource ();
    }
  else if (!strcmp (rq->op, "rename"))
    {
      read_resource_list ();
      rename_resource (rq);
    }
  else
    {
      handle_new (rq);
      handle_rename (rq);
      handle_delete (rq);
      read_resource_list ();
      show_resource_list ();
    }
}

const char *
resource_name (int seq)
{
  struct resource *r;
  
  assert (resource_tree != NULL);

  r = avl_find (resource_tree, &seq);
  if (r)
    return r->name;
  else
    fail ("bad resource sequence number %d", seq);
}


static void
show_resource_list (void)
{
  struct url rename, delete;
  int i;
  
  url_create (&rename, "resources");
  url_arg (&rename, "op", "rename");
  url_mark (&rename);

  url_create (&delete, "resources");
  url_mark (&delete);

  htmls ("<TABLE>\n");
  for (i = 0; i < n_resources; i++)
    {
      url_arg_int (&rename, "seq", resource[i].seq);
      url_arg_int (&delete, "delete", resource[i].seq);

      htmls ("<TR><TD>");
      html_escape (resource[i].name);
      html ("</TD>\n"
	    "<TD><A HREF=\"%s\">%s</A></TD>\n"
	    "<TD><A HREF=\"%s\">%s</A></TD>\n"
	    "</TR>\n",
	    url_string (&rename), _("Rename"),
	    url_string (&delete), _("Delete"));

      url_release (&rename);
      url_release (&delete);
    }
  htmls ("</TABLE>\n");

  url_arg (&delete, "op", "new");
  html ("<A HREF=\"%s\">%s</A>\n", url_string (&delete), _("New Resource"));

  url_destroy (&rename);
  url_destroy (&delete);
}

static void
rename_resource (struct request *rq)
{
  db_result *r;
  int seq;

  seq = stoi (std_get (rq->cgi, "seq"));
  
  html ("<P>%s</P>\n",
	_("Enter the new name for this resource, then click on\n"
	  "<STRONG>Submit</STRONG> to make the change take effect.\n"
	  "The name should specify both the type and location\n"
	  "of the resource; for instance, <STRONG>MSU/Case Center\n"
	  "V-Tel</STRONG>."));

  db_query ("SELECT name FROM resources WHERE seq = %d",
	    DB_FORMAT | DB_TUPLES, &r, seq);

  html_start_form ("resources", HT_POST, HT_URL_ENCODED);
  html_hidden ("rename", std_get (rq->cgi, "seq"));
  htmls ("<P>");
  html_text_field ("name", dbr_value (r, 0, 0), 30);
  htmls ("</P>\n"
	 "<P>\n");
  html_submit (_("Submit"));
  htmls ("</P>\n");
  html_end_form ();

  dbr_destroy (r);
}

static void
new_resource (void)
{
  html ("<P>%s</P>\n",
	_("Enter the name of the new resource, then click on\n"
	  "<STRONG>Submit</STRONG> to cause it to be created.\n"
	  "The name should specify both the type and location\n"
	  "of the resource; for instance, <STRONG>MSU/Case Center\n"
	  "V-Tel</STRONG>."));

  html_start_form ("resources", HT_POST, HT_URL_ENCODED);
  htmls ("<P>");
  html_text_field ("new", "", 30);
  htmls ("</P>\n"
	 "<P>");
  html_submit (_("Submit"));
  htmls ("</P>\n");
  html_end_form ();
}

static void
handle_new (struct request *rq)
{
  const char *new = std_get_null (rq->cgi, "new");
  if (new == NULL)
    return;

  db_query ("INSERT INTO resources VALUES ('%sq', nextval ('seq'))",
	    DB_FORMAT, NULL, new);
}

static void
handle_rename (struct request *rq)
{
  int seq = stoi (std_get (rq->cgi, "rename"));
  const char *name = std_get_null (rq->cgi, "name");

  if (seq == 0 || name == NULL)
    return;

  db_query ("UPDATE resources SET name = '%sq' WHERE seq = %d",
	    DB_FORMAT, NULL, name, seq);
}

static void
handle_delete (struct request *rq)
{
  int seq;
  db_result *r;
  int n;

  seq = stoi (std_get (rq->cgi, "delete"));
  if (seq == 0)
    return;

  db_begin ();
  db_query ("SELECT COUNT(*) FROM event_resource WHERE resource = %d",
	    DB_FORMAT | DB_TUPLES, &r, seq);

  n = stoi (dbr_value (r, 0, 0));
  if (n)
    {
      htmls ("<P>");
      html (_("Sorry, resource %s can't be deleted, since it is reserved\n"
	      "for use with %d calendar events.\n"),
	    resource_name (seq), n);
      htmls ("</P>\n");
    }
  else
    db_query ("DELETE FROM resources WHERE seq = %d", DB_FORMAT, NULL, seq);
  dbr_destroy (r);

  db_end ();
}

void
read_resource_list (void)
{
  db_result *r;

  if (resource != NULL)
    {
      int i;
      
      avl_destroy (resource_tree, NULL);
      for (i = 0; i < n_resources; i++)
	free ((char *) resource[i].name);
      free (resource);
    }

  db_query ("SELECT name, seq FROM resources ORDER BY name", DB_TUPLES, &r);

  resource_tree = avl_create (compare_resource, NULL);
  n_resources = dbr_nrows (r);
  resource = xmalloc (sizeof *resource * n_resources);
  
  {
    int i;

    for (i = 0; i < n_resources; i++)
      {
	resource[i].name = xstrdup (dbr_value (r, i, 0));
	resource[i].seq = stoi (dbr_value (r, i, 1));
	avl_force_insert (resource_tree, &resource[i]);
      }
  }

  dbr_destroy (r);
}

static int 
compare_resource (const void *_a, const void *_b, void *foo unused)
{
  const struct resource *a = _a;
  const struct resource *b = _b;

  if (a->seq > b->seq)
    return 1;
  else if (a->seq < b->seq)
    return -1;
  else
    return 0;
}
