/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include "activity.h"
#include "admin.h"
#include "alloc.h"
#include "auth.h"
#include "calendar.h"
#include "cgi.h"
#include "conf.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "files.h"
#include "help.h"
#include "html.h"
#include "identity.h"
#include "info.h"
#include "messages.h"
#include "options.h"
#include "priv.h"
#include "resource.h"
#include "scope.cgi.h"
#include "search.h"
#include "session.h"
#include "std.h"
#include "str.h"

/* Explained in scope.cgi.h. */
char *pgmname;
time_t cur_time;
struct tm cur_tm;

static void init_request (struct request *);
static void dispatch_screen (struct request *);
static void main_init (int, char **);
static void update_login_time (struct request *);
static void sigxcpu (int);
static void check_shut (struct request *);
static void scr_chat (struct request *);
static void scr_unknown (struct request *);
static void scr_null (struct request *);
static void connect_to_database (struct request *);

/* Main program. */
int
main (int argc, char **argv)
{
  struct request rq;
  
  /* This part of the program runs with root privileges.
     Security audit all code called from this section. */

  main_init (argc, argv);
  init_request (&rq);

  check_shut (&rq);
  rq.cgi = cgi_initialize ();

  /* Determine screen for display. */
  {
    const char *screen = std_get_null (rq.cgi, "screen");
  
    if (screen == NULL)
      {
	screen = getenv ("PATH_INFO");
	if (screen == NULL)
	  screen = "";
	else if (*screen == '/')
	  screen++;
      }

    rq.screen = xstrdup (screen);
  }

  session_initialize (&rq);
  connect_to_database (&rq);
  html_initialize ();

  /* For the team site, we leave out the HTML banner and trailer. */
  {
    const char *abbreviation = std_get (rq.cgi, "abbrev");

    if (!strcmp (abbreviation, "yes"))
      rq.abbrev = 1;
    else if (!strcmp (abbreviation, "partial"))
      rq.abbrev = 2;
  }
  
  auth_initialize (&rq);

  /* Check user authentication.  If failed, send one of the few
     acceptable screens. */
  if (rq.cx.usr[0] == '\0' || rq.cx.grp[0] == '\0'
      || (std_contains (rq.cgi, "needgrp")
	  && strcmp (rq.cx.grp, std_get (rq.cgi, "needgrp"))))
    {
      session_save_trusted (&rq);
      priv_set_user ("nobody");

      if (!has_prefix (rq.screen, "help")
	  && strcmp (rq.screen, "feedback")
	  && strcmp (rq.screen, "authfailed"))
	{
	  free (rq.screen);
	  rq.screen = xstrdup ("auth");
	}
	
      http_header ();
      html_banner (&rq);

      if (has_prefix (rq.screen, "help"))
	scr_help (&rq);
      else if (!strcmp (rq.screen, "feedback"))
	scr_feedback (&rq);
      else
	scr_auth (&rq);
      
      html_trailer (&rq);
      
      return EXIT_SUCCESS;
    }

  /* Update privilege cache and save trusted data if necessary. */
  priv_cache (&rq);
  session_save_trusted (&rq);

  /* If the user was directed to us because of lack of credentials,
     and he's now been supplied with credentials,
     then send him to the page he's interested in. */
  if (std_contains (rq.cgi, "redirect") && strcmp (rq.screen, "auth"))
    {
      const char *redirect = std_get (rq.cgi, "redirect");
      if (server_port == 80)
	printf ("Location: %s://%s%s\n\n",
		server_protocol, server_name, redirect);
      else
	printf ("Location: %s://%s:%d%s\n\n",
		server_protocol, server_name, server_port, redirect);

      return EXIT_SUCCESS;
    }

  /* If the user wants to go to his team site, send him there. */
  if (!strcmp (rq.screen, "teamsite"))
    {
      printf ("Location: %s://%s/teams/%s\n\n",
	      server_protocol, server_name, rq.cx.grp);

      return EXIT_SUCCESS;
    }

  /* Determine operation. */
  rq.op = std_get (rq.cgi, "op");
  
  /* Password changes and log-outs must be done as root. */
  if (!strcmp (rq.screen, "admin"))
    {
      scr_admin (&rq);
      if (getuid () < 100 || geteuid () < 100
	  || getgid () < 100 || getegid () < 100)
	abort ();
    }
  else
    {
      if (!strcmp (rq.screen, "password"))
	auth_password_change (&rq);
      else if (!strcmp (rq.screen, "logout"))
	session_logout (&rq);

      /* Drop privileges, read untrusted data, read options,
	 initialize timezone. */
      drop_privs_finish_init (&rq);
  
      /* Root privileges have been dropped.
	 Security audit the rest of the program anyway since it has
	 full access to the TeamSCOPE database. */

      /* Dispatch to one of the screen-specific handlers. */
      dispatch_screen (&rq);
    }
  
  session_save_untrusted (&rq);
  options_close (&rq);
  db_close ();

  return EXIT_SUCCESS;
}

void
drop_privs_finish_init (struct request *rq)
{
  /* Drop privileges. */
  priv_drop (rq);

  /* Root privileges have been dropped.
     Security audit the rest of the program anyway since it has full
     access to the TeamSCOPE database. */

  session_read_untrusted (rq);
  options_initialize (rq);
  datetime_init_timezone (&rq->cx);
}

/* Removes all but the variables we want from the environment.
   We can't be run by anyone untrusted anyhow, but paranoia pays. */
static void
sanitize_environment (void)
{
  static const char *okay[] =
    {
      "CONTENT_LENGTH",
      "CONTENT_TYPE",
      "PATH_INFO",
      "QUERY_STRING",
      "REDIRECT_SCOPE_SESSION",
      "REMOTE_USER",
      "REQUEST_METHOD",
      "REQUEST_URI",
      "SCOPE_SESSION",
      "SERVER_NAME",
      "TZ",
      NULL,
    };
  
  char **p;

  for (p = environ; *p != NULL; )
    {
      const char **q;
      
      for (q = okay; *q != NULL; q++)
	if (!strncmp (*p, *q, strlen (*q)))
	  break;

      if (*q == NULL)
	{
	  char **r;

	  for (r = p; *r != NULL; r++)
	    *r = r[1];
	}
      else
	p++;
    }
}

/* Miscellaneous startup tasks. */
static void
main_init (int argc unused, char **argv)
{
  /* Start the clock ticking. */
#if GLOBAL_DEBUGGING
  start_clock = clock ();
  gettimeofday (&start_timeval, NULL);
#endif

  /* Determine program name. */
  {
    char *p = strrchr (argv[0], '/');
    pgmname = p != NULL ? &p[1] : argv[0];
  }

  sanitize_environment ();

  /* Set a secure path for execution of subprograms. */
  if (putenv ("PATH=/bin:/sbin:/usr/bin:/usr/sbin") == -1)
    fail ("can't set PATH environment variable");

#if 0
  /* Log environment contents for debugging. */
  {
    char **envp;

    log ("environment:");
    for (envp = environ; *envp; envp++)
      log ("%s", *envp);
  }
#endif

  /* Ensure we're root. */
  if (geteuid () != 0)
    fail ("must run as root");

  /* Don't want anything screwing with file permissions. */
  umask (0);

  /* Read clock. */
  time (&cur_time);
  cur_tm = *localtime (&cur_time);

  /* For paranoia: Set CPU resource limit to avoid infinite loops. */
  {
    struct rlimit r;
    r.rlim_cur = 5;
    r.rlim_max = 600;
    
    signal (SIGXCPU, sigxcpu);

    if (-1 == setrlimit (RLIMIT_CPU, &r))
      fail ("setrlimit: %s", strerror (errno));
  }
  
#if 0
  /* Enable core dumps. */
  {
    struct rlimit r;
    r.rlim_cur = r.rlim_max = 1024*1024*1024; /* 1 GB. */
    setrlimit (RLIMIT_CORE, &r);

    chdir ("/");
  }
#endif
}

/* Check whether TeamSCOPE service is disabled. */
static void
check_shut (struct request *rq)
{
  FILE *f;

  f = fopen (CONF_DIR "/shut", "r");
  if (f == NULL)
    return;

  session_read_conf ();
  html_initialize ();
  http_header ();
  html_banner (rq);

  html ("<H1 ALIGN=CENTER>%s</H2>\n"
	"<P>%s:</P>\n"
	"<PRE>\n",
	_("TeamSCOPE Service Disabled"),
	_("TeamSCOPE service is temporarily disabled\n"
	  "for maintenance or upgrade.\n"
	  "Additional information may be available below"));

  for (;;)
    {
      char buf[1024];

      if (fwrite (buf, 1, fread (buf, 1, sizeof buf, f), stdout) != sizeof buf)
	break;
    }
  
  htmls ("</PRE>\n"
	 "<P>");
  html (_("Your patience is appreciated while service is restored.\n"
	  "Send questions or comments to the TeamSCOPE administrator,\n"
	  "%s.\n"),
	admin_email);
  htmls ("</P>\n");

  html_trailer (rq);

  exit (EXIT_SUCCESS);
}

static void
init_request (struct request *rq)
{
  init_context (&rq->cx);
  rq->cgi = NULL;
  rq->session_id[0] = '\0';
  rq->trusted = NULL;
  rq->untrusted = NULL;
  rq->screen = NULL;
  rq->op = NULL;
  rq->abbrev = 0;
}

/* Call the proper function for SCREEN. */
static void
dispatch_screen (struct request *rq)
{
  /* Defines a screen. */
  struct screen
    {
      char *name;				/* Screen name. */
      int flags;				/* Flags. */
      void (*dispatch) (struct request *);	/* Dispatch function. */
    };

  /* Screen flags. */
  enum
    {
      SCR_NO_HEADERS = 001,	/* Don't print HTTP & HTML header/trailer. */
      SCR_NO_UPDATE_LOGIN = 002	/* Don't update user login time. */
    };

  /* List of screens. */
  static const struct screen screens[] =
    {
      {"admin",			0,	scr_admin},

      {"chat",			003,	scr_chat},

      {"activity",		0,	scr_activity},

      {"files",			001,	scr_files},

      {"messages",		0,	scr_messages},

      {"calendar",		0,	scr_calendar},
      {"calendargif",		003,	scr_calendargif},
      {"ganttgif",		003,	scr_ganttgif},

      {"resources",		0,	scr_resources},

      {"options",		0,	scr_options},
      {"webfilter",		0,	scr_webfilter},
      {"emailfilter",		0,	scr_emailfilter},
      {"password",		001,	scr_password},

      {"info",			0,	scr_info},
      {"infousagegif",		003,	scr_infousagegif},

      {"search",		0,	scr_search},
      {"activitysearch",	0,	scr_activitysearch},
      {"filesearch",		0,	scr_filesearch},

      {"help",			0,	scr_help},
      {"feedback",		0,	scr_feedback},

      {"welcome",		0,	scr_welcome},

      {"auth",			0,	scr_auth},
      {"",			0,	scr_null},
      {NULL,			0,	scr_unknown},
    };

  char *screen_prefix;
  const struct screen *s;
  int len;
  
  len = strcspn (rq->screen, "-");
  screen_prefix = local_alloc (len + 1);
  memcpy (screen_prefix, rq->screen, len);
  screen_prefix[len] = '\0';

  for (s = screens; s->name; s++)
    if (!strcmp (screen_prefix, s->name))
      break;

  local_free (screen_prefix);

#if DEBUGGING
  if (rq->screen[0] != '\0' && s->name == NULL)
    log ("no such screen %s", rq->screen);
#endif
  
  if (!(s->flags & SCR_NO_HEADERS))
    {
      http_header ();
      html_banner (rq);
    }
  
  s->dispatch (rq);

  if (!(s->flags & SCR_NO_HEADERS))
    html_trailer (rq);
  if (!(s->flags & SCR_NO_UPDATE_LOGIN))
    update_login_time (rq);

#if __CHECKER__ && 0
  {
    extern void __chkr_garbage_detector ();

    fflush (stderr);
    __chkr_garbage_detector ();
  }
#endif
}

#if __CHECKER__
/* This prints an "Assertion failed" message and aborts.  */
void __assert_fail __P ((__const char *__assertion,
			 __const char *__file,
			 unsigned int __line,
			 __const char *__function))
{
  fprintf (stderr, "%s:%u: %s: Assertion `%s' failed, inducing segfault\n",
	   __file, __line, __function, __assertion);
  *(char *) 0 = 0;
  abort ();
}
#endif /* __CHECKER__ */

/* Update the user last-login time.  This could be done at the
   beginning of the program, but if it were then we'd hold the lock on
   the `users' database table for this particular user for quite a
   while longer. */
static void
update_login_time (struct request *rq)
{
  char *query_string = getenv ("QUERY_STRING");
  if (query_string == NULL)
    query_string = "";
  
  db_query ("INSERT INTO uselog "
	    "VALUES ('%sq', datetime (timenow ()), '%sq', '%sq', '%sq')",
	    DB_FORMAT, NULL, rq->cx.usr, rq->screen, rq->cx.grp, query_string);

  db_query ("UPDATE users "
	    "SET last_usage = datetime (timenow ()) "
	    "WHERE usr = '%sq'",
	    DB_FORMAT, NULL, rq->cx.usr);
}

/* Handle a SIGXCPU signal, which indicates that we've used up all the
   CPU time we expected to need.  This probably indicates that we've
   entered an infinite loop.  Abort the program and log an error message.

   Note: stderr might theoretically be in use, but it's unlikely.  At
   any rate we've dropped root by now, so at worst we munge the user's
   files.  *shrug* */
static void
sigxcpu (int signum unused)
{
  if (!geteuid ())
    abort ();
  
  fail ("infinite loop");
}

/* Tell the user we don't know about that screen. */
static void
scr_unknown (struct request *rq)
{
  htmls ("<P ALIGN=CENTER>\n");
  html (_("The function you've requested is either\n"
	  "unavailable or unimplemented in this TeamSCOPE version.\n"
	  "If you think that this is a bug, please report it to\n"
	  "the server administrator, %s.\n"),
	admin_email);
  htmls ("</P>\n");

  html_powered_by (rq);
}

/* Display the TeamSCOPE Java chat applet in a window by itself as
   requested by the user. */
static void
scr_chat (struct request *rq)
{
  html ("Content-Type: text/html\n\n");
  html_chat_applet (rq);
}

/* No particular screen requested.
   Pick either the welcome screen or the activity screen. */
static void
scr_null (struct request *rq)
{
  if (show_welcome (&rq->cx))
    scr_welcome (rq);
  else
    scr_activity (rq);
}

/* Opens a connection to the database.  This is a wrapper around
   db_open() that sets the effective uid as appropriate. */
static void
connect_to_database (struct request *rq)
{
  uid_t scope_uid;

  /* Obtain the `scope' user's uid.  It might be cached in
     trusted_data. */
  if (std_get_null (rq->trusted, "scope-uid") != NULL)
    scope_uid = strtol (std_get (rq->trusted, "scope-uid"), NULL, 10);
  else
    {
      char buf[INT_DIGITS + 1];
      struct passwd *passwd;
	
      passwd = getpwnam ("scope");
      if (passwd == NULL)
	fail ("can't find `scope' user in /etc/passwd");
      scope_uid = passwd->pw_uid;

      sprintf (buf, "%ld", (long) scope_uid);
      std_put (rq->trusted, "scope-uid", buf);
    }

  /* Set effective uid to `scope'. */
  if (scope_uid == 0)
    fail ("scope user can't have uid 0");
  if (setreuid (0, scope_uid) == -1)
    fail ("can't seteuid(%d) to scope", scope_uid);

  db_open (NULL);

  /* Restore our original euid. */
  if (setreuid (0, 0) == -1)
    fail ("can't setreuid(0,0)");
}

/* Print a helpful log message prefix to stderr of the form `May 13
   19:31:30 1999 scope[31230]'.  Helper function for error.c. */
void
msg_prefix (FILE *file, const char *function_name)
{
  char *t = asctime (&cur_tm);
  *strchr (t, '\n') = '\0';

  fprintf (file, "%s %s[%ld] ", &t[4], pgmname, (long) getpid ());
  if (function_name != NULL)
    fprintf (file, "(%s) ", function_name);
}
