/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include "alloc.h"
#include "avl.h"
#include "error.h"
#include "identity.h"
#include "str.h"
#include "std.h"

/* A key=>values mapping.  This is the type stored in the AVL tree. */
struct st_entry
  {
    /* This must be the first member in this struct. */
    char *key;

    int n;
    char *value[2];
  };

/* The entire string dictionary. */
struct std
  {
    struct avl_tree *tree;
    int allow_dup;
    int modified;
  };

/* Public routines. */

static int compare_st_entry (const void *, const void *, void *);
static void free_st_entry (void *_a, void *param unused);
static void add_line (const char *fn, struct std *dict, const char *line,
		      struct st *s, struct st *t, int subst);
static void escape (struct st *buffer, const char *string);
static const char *unescape (const char *fn, const struct std *dict,
			     const char *string, struct st *buffer,
			     int flags);

/* Create and return a new string dictionary.  If ALLOW_DUP is nonzero
   then entries with duplicate keys are allowed in the dictionary;
   otherwise, they are not allowed. */
struct std *
std_create (int allow_dup)
{
  struct std *dict = xmalloc (sizeof *dict);

  dict->tree = avl_create (compare_st_entry, NULL);
  dict->allow_dup = allow_dup;
  dict->modified = 1;

  return dict;
}

/* Destroys string dictionary DICT. */
void
std_destroy (struct std *dict)
{
  assert (dict != NULL);
  
  avl_destroy (dict->tree, free_st_entry);
  free (dict);
}

/* Returns the number of keys in DICT. */
int
std_count (struct std *dict)
{
  assert (dict != NULL);

  return avl_count (dict->tree);
}

/* Insert KEY and VALUE into DICT.  Returns nonzero if the pair could
   be inserted, zero otherwise.  KEY and VALUE remain owned by the
   caller.  The only reason that this function will return zero is a
   duplicate entry, if this dictionary does not accept duplicates.  A
   null VALUE will result in KEY being inserted in DICT without any
   values attached to it. */
int
std_put (struct std *dict, const char *key, const char *value)
{
  void **item;

  assert (dict != NULL && key != NULL);
  item = avl_probe (dict->tree, &key);
  if (*item != &key)
    {
      if (value == NULL)
	return 1;
	
      if (dict->allow_dup)
	{
	  struct st_entry *e = *item;
	  e = xrealloc (e, sizeof *e + sizeof (char *) * e->n);
	  e->value[e->n++] = xstrdup (value);
	  e->value[e->n] = NULL;
	  *item = e;
	  dict->modified = 1;
	}
      else
	return 0;
    }
  else
    {
      struct st_entry *new = xmalloc (sizeof *new);
      new->key = xstrdup (key);
      if (value)
	{
	  new->n = 1;
	  new->value[0] = xstrdup (value);
	  new->value[1] = NULL;
	}
      else
	{
	  new->n = 0;
	  new->value[0] = NULL;
	}
      *item = new;
      dict->modified = 1;
    }

  return 1;
}

/* Insert KEY and VALUE into DICT.  If KEY already existed in DICT,
   its values are deleted before insertion. */
void
std_replace (struct std *dict, const char *key, const char *value)
{
  assert (dict != NULL && key != NULL);
  
  std_delete (dict, key);
  std_put (dict, key, value);
}

/* Delete KEY from DICT, if it exists. */
void
std_delete (struct std *dict, const char *key)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_delete (dict->tree, &key);
  if (e)
    {
      free_st_entry (e, NULL);
      dict->modified = 1;
    }
}

/* Returns nonzero only if KEY exists in DICT.
   NOTE: KEY might not have a value, so don't depend on std_get_null()
   to return non-NULL just because std_contains() returns nonzero. */
int
std_contains (const struct std *dict, const char *key)
{
  assert (dict && key);
  return avl_find (dict->tree, &key) != NULL;
}
  
/* Returns nonzero only if KEY exists in DICT and one of its values
   is VALUE. */
int
std_value_contains (const struct std *dict, const char *key, const char *value)
{
  struct st_entry *e;
  int i;
  
  assert (dict && key && value);
  e = avl_find (dict->tree, &key);
  if (e == NULL)
    return 0;

  for (i = 0; i < e->n; i++)
    if (!strcmp (e->value[i], value))
      return 1;
  
  return 0;
}
  
/* Look up KEY in DICT and return the corresponding value if any or a
   null pointer if none.  If there are multiple values then the first
   is returned.

   Note that this function will return a null pointer if KEY does not
   exist in DICT *or* if KEY does exist but contains no values. */
const char *
std_get_null (const struct std *dict, const char *key)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_find (dict->tree, &key);
  return e != NULL ? e->value[0] : NULL;
}

/* Look up KEY in DICT and return the corresponding value if it
   exists, has at least one value, and that value is not a string
   containing only spaces.  Otherwise, returns a null pointer. */
const char *
std_get_nonempty (const struct std *dict, const char *key)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_find (dict->tree, &key);
  return (e != NULL && e->value[0] != NULL && !empty (e->value[0])
	  ? e->value[0]
	  : NULL);
}

/* Look up KEY in DICT and return the corresponding value if any or an
   empty string if none.  If there are multiple values then the first
   is returned.  If KEY exists but has no values, also returns an
   empty string. */
const char *
std_get (const struct std *dict, const char *key)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_find (dict->tree, &key);
  return e != NULL && e->value[0] != NULL ? e->value[0] : "";
}

/* Look up KEY in DICT and return the corresponding value if any, or
   fail if none. */
const char *
std_get_assert (const struct std *dict, const char *key)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_find (dict->tree, &key);

  if (e == NULL || e->value[0] == NULL)
    fail ("dictionary missing required key %s", key);
      
  return e->value[0];
}

/* Look up KEY in DICT.  If found, returns a null-terminated list.  If
   N is non-null, sets *N to the length of list (excluding the null
   terminator).  If not found, returns a null pointer. */
const char **
std_get_many (const struct std *dict, const char *key, int *n)
{
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  e = avl_find (dict->tree, &key);
  if (e != NULL)
    {
      if (n != NULL)
	*n = e->n;
      return (const char **) &e->value;
    }
  else
    {
      if (n != NULL)
	*n = 0;
      return NULL;
    }
}

/* Retrieve a subdictionary stored as KEY from DICT. */
struct std *
std_get_subdict (struct std *dict, const char *key)
{
  struct std *new;
  struct st_entry *e;

  assert (dict != NULL && key != NULL);
  new = std_create (1);
  e = avl_find (dict->tree, &key);
  if (e != NULL)
    {
      struct st s, t;
      int i;

      st_init (&s, 64);
      st_init (&t, 64);
      for (i = 0; i < e->n; i++)
	add_line ("subdict", new, e->value[i], &s, &t, 0);
      st_destroy (&s);
      st_destroy (&t);
    }
  new->modified = 0;
  return new;
}

/* Store subdictionary INNER_DICT as key OUTER_KEY in dictionary
   OUTER_DICT.  The subdictionary may be retrieved later using
   std_get_subdict(), and it may be stored on disk and read
   back later using the usual functions. */
void
std_put_subdict (struct std *outer_dict, const char *outer_key,
		 const struct std *inner_dict)
{
  struct st s;
  
  assert (outer_dict != NULL && outer_dict->allow_dup
	  && outer_key != NULL && inner_dict != NULL);

  st_init (&s, 64);
  std_delete (outer_dict, outer_key);

  {
    struct std_iterator iter;
    const char *inner_key;
    char *inner_value;

    std_init_iterator (inner_dict, &iter);
    while (std_iterate (&iter, &inner_key, &inner_value))
      {
	st_clear (&s);
	escape (&s, inner_key);

	if (inner_value != NULL)
	  {
	    st_putchar (&s, '=');
	    escape (&s, inner_value);
	  }

	std_put (outer_dict, outer_key, st_value (&s));
      }
  }

  st_destroy (&s);
}

/* Set up ITER as an iterator for DICT. */
void
std_init_iterator (const struct std *dict, struct std_iterator *iter)
{
  assert (dict != NULL && iter != NULL);
  iter->tree = dict->tree;
  avl_init_traverser (&iter->trav);
  iter->entry = NULL;
}

/* Sets KEY and VALUE to the next key/value pair in the dictionary
   being iterated with iterator ITER.  Keys are returned in
   lexicographic order; values with the same key are returned in order
   of insertion.  Returns zero if the function failed due to the last
   element having been already seen, nonzero otherwise. */
int
std_iterate (struct std_iterator *iter, const char **key, char **value)
{
  assert (iter != NULL && key != NULL && value != NULL);
  if (iter->entry == NULL)
    {
      iter->entry = avl_traverse (iter->tree, &iter->trav);
      if (iter->entry == NULL)
	return 0;

      iter->index = 0;
    }
  
  *key = iter->entry->key;
  *value = iter->entry->value[iter->index];

  if (++iter->index >= iter->entry->n)
    iter->entry = NULL;

  return 1;
}

/* Attempts to lock file descriptor FD with lock type TYPE and returns
   nonzero only if successful.  Note: this will block until the lock
   becomes available, so failure is really failure, not merely a
   temporary condition. */
int 
lock (int type, int fd)
{
  /* See _Advanced Programming in the UNIX Environment_,
     p. 368-370. */
  struct flock lock;

  lock.l_type = type;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  do
    {
      if (fcntl (fd, F_SETLKW, &lock) == 0)
	return 1;
    }
  while (errno == EINTR);

  return 0;
}

/* Reads a string dictionary from file FILENAME into DICT and returns
   nonzero only if successful.  The dictionary may be pre-initialized
   to provide defaults.
   If SUBST is nonzero, then $() and ${} variable substitutions are
   performed on the variables as they are read.
   The file being read must belong to the user with uid UID. */
int
std_read (struct std *dict, const char *filename, int flags, uid_t uid)
{
  struct st s, t;
  
  char *line = NULL;
  size_t line_size;

  FILE *f;

  f = fopen (filename, "r");
  if (f == NULL)
    {
      log ("opening %s: %s", filename, strerror (errno));
      return 0;
    }

  /* Verify the file's owner.  This prevents users from causing
     problems by setting a symlink from ~/.scope/options to
     /etc/passwd, etc.

     Note that in normal operation this extra paranoia isn't really
     justified, since std_read() is normally called on
     ~/.scope/options when running under the user's uid, not root.
     However, it *is* justified when called from i.e.,
     get_user_context(), which typically runs from scope-db as
     root. */
  verify_file_owner (filename, f, uid);

  if (!lock (F_RDLCK, fileno (f)))
    return 0;

  st_init (&s, 64);
  st_init (&t, 64);
  for (;;)
    {
      long line_len;

      line_len = getline (&line, &line_size, f); 
      if (line_len == -1)
	{
	  dict->modified = 0;
	  
	  if (ferror (f))
	    {
	      log ("error reading %s: %s", filename, strerror (errno));
	      std_destroy (dict);
	      dict = NULL;
	    }

	  free (line);
	  fclose (f);

	  st_destroy (&s);
	  st_destroy (&t);
	  return 1;
	}

      assert (line_len > 0 && line[line_len - 1] == '\n');
      line[--line_len] = '\0';

      add_line (filename, dict, line, &s, &t, flags);
    }
}

/* Writes string dictionary DICT to file FILENAME.  Returns nonzero
   only if successful. */
int
std_write (const struct std *dict, const char *filename, mode_t mode)
{
  FILE *f;
  
  {
    int fd;

    /* Open file. */
    fd = open (filename, O_WRONLY | O_CREAT, mode);
    if (fd == -1)
      {
	log ("error opening %s: %s", filename, strerror (errno));
	return 0;
      }

    /* Try to lock, then truncate if successful. */
    if (!lock (F_WRLCK, fd) || ftruncate (fd, 0) == -1)
      {
	log ("error writing %s: %s", filename, strerror (errno));
	return 0;
      }

    f = fdopen (fd, "w");
    if (f == NULL)
      {
	log ("error opening %s as stream: %s", filename, strerror (errno));
	return 0;
      }
  }
  
  {
    struct std_iterator iter;
    const char *key;
    char *value;
    struct st s;
    
    st_init (&s, 64);
    std_init_iterator (dict, &iter);
    while (std_iterate (&iter, &key, &value))
      {
	st_clear (&s);
	escape (&s, key);
	st_putchar (&s, '=');
	escape (&s, value);
	st_putchar (&s, '\n');
	fwrite (st_value (&s), st_length (&s), 1, f);
      }
  }

  if (ferror (f) || fclose (f) == EOF)
    {
      log ("error writing %s: %s", filename, strerror (errno));
      return 0;
    }

  return 1;
}

/* Returns nonzero only if DICT was modified since it was created or
   read from a file. */
int
std_modified (const struct std *dict)
{
  return dict->modified;
}

#if GLOBAL_DEBUGGING
/* Writes contents of DICT to stderr. */
void
std_print (const struct std *dict)
{
  struct avl_traverser trav;

  avl_init_traverser (&trav);
  for (;;)
    {
      struct st_entry *e = avl_traverse (dict->tree, &trav);
      if (e == NULL)
	break;

      if (e->n)
	{
	  int i;

	  for (i = 0; i < e->n; i++)
	    fprintf (stderr, "\t\"%s\" = \"%s\"\n", e->key, e->value[i]);
	}
      else
	fprintf (stderr, "\t\"%s\"\n", e->key);
    }
}
#endif /* GLOBAL_DEBUGGING */

/* Private routines. */

/* Compare two struct st_entry's lexicographically and return
   a strcmp()-type result. */
static int
compare_st_entry (const void *pa, const void *pb, void *param unused)
{
  const struct st_entry *a = pa;
  const struct st_entry *b = pb;

  assert (a != NULL && b != NULL);
  return strcmp (a->key, b->key);
}

/* Free the specified struct st_entry and all its associated
   strings. */
static void
free_st_entry (void *pa, void *param unused)
{
  struct st_entry *a = pa;
  int i;

  assert (a != NULL);
  
  free (a->key);
  for (i = 0; i < a->n; i++)
    free (a->value[i]);
  
  free (a);
}

/* Return the value of uppercase hex digit CH,
   or -1 if it's not an uppercase hex digit. */
static int
Xdigit_value (int ch)
{
  static const char letters[] = "0123456789ABCDEF";
  const char *p = strchr (letters, ch);
  return p != NULL ? p - letters : -1;
}

/* Add LINE, which consists of a `key=value' formatted string, to
   DICT.  S and T are used as temporary buffers. */
static void
add_line (const char *fn, struct std *dict, const char *line,
	  struct st *s, struct st *t, int flags)
{
  const char *stop;
  const char *value;

  st_clear (s);
  stop = unescape (fn, dict, line, s, flags);
  if (st_length (s) == 0)
    return;
  
  if (*stop == '=')
    {
      st_clear (t);
      unescape (fn, dict, stop + 1, t, flags | STDF_LIBERAL);
      value = st_value (t);
    }
  else
    value = NULL;

  std_put (dict, st_value (s), value);
}

/* Copies characters from STRING to BUFFER, escaping characters with %
   as necessary. */
static void
escape (struct st *buffer, const char *string)
{
  for (;;)
    {
      switch (*string)
	{
	case '=':
	case '#':
	case '%':
	case '$':
	case '\n':
	  st_printf (buffer, "%%%02X", (unsigned char) *string);
	  break;

	case '\0':
	  return;

	default:
	  st_putchar (buffer, (unsigned char) *string);
	}

      string++;
    }
}

/* Copies characters from SRC into DST, interpreting %-encoded
   escapes.  Stops at a null or # and returns a pointer to the stop
   character.  If STDF_LIBERAL is not set in FLAGS then stops at = as
   well.

   If STDF_SUBST is set in FLAGS then variable substitutions are
   performed on the value string. If STDF_TRIM_SPACES is set then
   leading and trailing spaces are ignored. */
static const char *
unescape (const char *fn, const struct std *dict,
	  const char *src, struct st *dst, int flags)
{
  if (flags & STDF_TRIM_SPACES)
    while (isspace ((unsigned char) *src))
      src++;
  
  for (;; src++)
    switch (*src)
      {
      case '=':
	if (flags & STDF_LIBERAL)
	  {
	    st_putchar (dst, *src);
	    break;
	  }
	/* fall through */

      case '#':
      case '\0':
	if (flags & STDF_TRIM_SPACES)
	  {
	    char *s = st_value (dst);
	    size_t len = st_length (dst);
	    while (len > 0 && isspace ((unsigned char) s[len - 1]))
	      len--;
	    st_truncate (dst, len);
	  }
	  
	return src;

      case '$':
	if ((flags & STDF_SUBST) && (src[1] == '(' || src[1] == '{'))
	  {
	    char *name;
	    const char *value;
	    size_t len;

	    len = strcspn (src + 2, src[1] == '(' ? ")" : "}");
	    name = local_alloc (len + 1);
	    memcpy (name, src + 2, len);
	    name[len] = '\0';

	    value = std_get_null (dict, name);
	    if (value != NULL)
	      st_concat (dst, value);
	    else
	      log ("%s: unknown variable `%s'", fn, name);
	    src += 2 + len;

	    local_free (name);
	  }
	else
	  st_putchar (dst, *src);
	break;

      case '%':
	{
	  int a = Xdigit_value (src[1]);
	  if (a != -1)
	    {
	      int b = Xdigit_value (src[2]);
	    
	      if (b != -1)
		{
		  st_putchar (dst, a * 16 + b);
		  src += 2;
		  break;
		}
	    }
	}
	/* fall through */

      default:
	st_putchar (dst, *src);
      }
}
